#include <stdio.h>
#include <apollo/base.h>
#include <apollo/gpr.h>
#include <tml/tml_c_types.h>
#include <tml/tml_dp_graph.hxx>
#include <tml/tml_os_depends.h>

extern long create_bitmap();
extern void enable_all_keys();

/**************************************************************/
tml_c_model example_read(char *file_name, int f_len)
{
    tml_c_model      return_model; /* the model created */
    FILE            *file_id;
    char            *open_type = "r";
    tml_c_node       current_node;
    tml_c_gnode      current_gnode,
                     last_gnode;
    tml_c_property   font_prop,
                     margin_prop,
                     wrap_prop;
    char            *char_buf;
    int              buf_size,
                     chars_attached,
                     chars_back;
    int              add_font_prop;


    /* open file */
    file_id = fopen(file_name, open_type);
    if (!file_id) {
        fprintf(stderr, "unable to open file\n");
        return 0;
        }

    /* allocate model */
    return_model =  model_class_alloc();
    if (!return_model) {
        fprintf(stderr, "unable to create model\n");
        return 0;
        }

    /* make root mode for model */
    current_node = model_make_root(return_model);
    if (!current_node) {
        fprintf(stderr, "unable to create root node\n");
        return 0;
        }

    /* make a character wrap property */
    /* associate it with the root node */
    /* enables character wrap for entire document */
    /*   except where overridden, for a subtree in */
    /*   the model, by a different wrap property */
    /*   attached to some other node */
    wrap_prop = tml_c_make_wrap_prop(char_wrap, return_model);
    if (!wrap_prop) {
        fprintf(stderr, "unable to make wrap prop\n");
        return 0;
        }
    /* connect wrap property to root node */
    tml_c_connect_prop(wrap_prop, current_node);

    /* make a margin property */
    /* use 40-pixel left and right margin */
    /* no top or bottom margin */
    margin_prop = tml_c_make_margin_prop(40, 40, 0, 0, return_model);
    if (!margin_prop) {
        fprintf(stderr, "unable to make margin property \n");
        return 0;
        }
    /* connect margin property to root node */
    tml_c_connect_prop(margin_prop, current_node);

    /* make a terminal node */
    current_gnode = gnode_class_alloc(return_model);
    if (!current_gnode) {
        fprintf(stderr, "unable to create node\n");
        return 0;
        }
    /* connect it as child of root */
    node_connect_node(current_node, current_gnode, n_down);

    /* make a font property */
    font_prop =
      tml_c_make_font_prop("/sys/dm/fonts/helvetica12", 25, return_model);
    if (!font_prop) {
        fprintf(stderr, "failed in making the font properly\n");
        exit(0);
        }
    /* connect font property to first terminal node */
    /* overrides, for this node only (because it is a */
    /*   terminal node) the default font property value */
    /*   associated with each model */
    tml_c_connect_prop(font_prop, current_gnode);

    /* allocate space for reading characters */
    buf_size = 256;
    char_buf = (char*) tml_c_tml_os_alloc(
                         (unsigned long) buf_size,
                         (unsigned long) sizeof(char));
    if (!char_buf) {
        fprintf(stderr, "unable to allocate character buffer\n");
        return 0;
        }

    /* read 256 bytes from file */
    chars_back = fread(char_buf, 1, buf_size, file_id);
    /* associate them with the newly-created node */
    chars_attached = node_attach_data(current_gnode, char_buf, chars_back);
    /* check for read error or error in attaching data */
    if (chars_attached != chars_back) {
        fprintf(stderr, "unable to attach all characters to a node");
        model_free(return_model);
        return 0;
        }

    /* remember node just created */
    last_gnode = current_gnode;
    /* check for more characters in file */
    if (chars_back < buf_size)
        return return_model;
    /* file-reading loop */
    /* read characters from file */
    /* create a new terminal node */
    /* associate characters with newly-created node */
    /* connect new node to last node created */
    /* remember newly-created node as last created */
    /* change font on alternate nodes */
    add_font_prop = 0;
    for (chars_back = fread(char_buf, 1, buf_size, file_id);
      chars_back;
      chars_back = fread(char_buf, 1, buf_size, file_id)) {

        /* create a new terminal node */
        current_gnode = gnode_class_alloc(return_model);

        /* quit if allocation fails */
        if (!current_gnode) {
            fprintf(stderr, "unable to alloc node\n");
            return return_model;
            }

        /* associate characters read with newly-created node */
        chars_attached = node_attach_data(current_gnode, char_buf, chars_back);
        if ( chars_attached != chars_back ) {
            fprintf(stderr, "unable to attach all characters to a node\n");
            model_free(return_model);
            return 0;
            }

        /* attach font property to alternate nodes */
        if (add_font_prop) {
            tml_c_connect_prop(font_prop, current_gnode);
            add_font_prop = 0;
            }
        else
            add_font_prop = 1;

        /* connect new node to last node created */
        node_connect_node(last_gnode, current_gnode, n_right);
        /* remember newly-created node as last created */
        last_gnode = current_gnode;
        }

    /* close file */
    chars_back = fclose(file_id);

    /* release dynamically allocated storage */
    tml_c_tml_os_free(char_buf);

    /* return model */
    return return_model;
}
/**************************************************************/

main()
{
    tml_c_model      display_model;
    tml_c_view_port  display_viewport;
    tml_c_format     display_format;
    C_MODEL_POS      start_display;
    long             bitmap_desc;
    char             unob;
    gpr_$event_t     ev_type;
    char             ev_value = '\0';
    gpr_$position_t  ret_coord;
    status_$t        status;
    long             dx, dy;    /* size of viewport */
    C_SCREEN_XY      inq_pos;   /* any (x,y) coordinate on the screen */
    tml_c_phead      last_page; /* used for going to bottom of file */
    tml_c_cline      last_line; /* used for going to bottom of file */

    /* initialize TMLib */
    tml_init();

    /* associate all viewport objects with the GPR graphic system */
    tml_c_tml_dp_set_dflt_vpftbl(tml_c_tml_dp_get_gpr_vpftbl());

/**************************************************************/
    /* create a model from the ASCII file test_file */
    display_model = example_read("test_file", 9);
    if (!display_model)
        exit (1);
/**************************************************************/

    /* allocate a viewport */
    display_viewport = view_port_class_alloc();
    /* create graphics display area */
    bitmap_desc = create_bitmap();

    /* assign the display area to the viewport */
    view_port_set_handle(display_viewport, &bitmap_desc);

    /* create a format */
    /* attach it to the viewport */
    view_port_attach_format(display_viewport, 0, 0, 0);
    /* get a handle for the format created by the last call */
    display_format = view_port_get_format(display_viewport);
    /* attach the model to the format that is associated with the viewport */
    format_attach(display_format, display_model);

    start_display.in_node = model_get_root(display_model);
    /* quit if nothing attached to root */
    if (!start_display.in_node)
        exit(1);

    /* fill viewport with text; first character placed at pixel (0,0) */
    tml_dp_damage_repair(view_port_get_handle(display_viewport));

    /* enable keyboard */
    enable_all_keys();

    while (ev_value != 'e') { /* typing 'e' exits the program */

        unob = gpr_$event_wait(&ev_type, (unsigned char*) &ev_value,
                               &ret_coord, &status);

        switch (ev_value) {

            case('u'):
                /* move viewport up the conceptual page */
                /*   half the viewport height */
                /* get height of viewport -- it may have changed */
                dy = view_port_get_hi(display_viewport);
                /* find character at top of viewport */
                view_port_xy_2_mpos(display_viewport,
                                    &zero_zero, &start_display);
                /* if a character was found */
                /*   display it halfway down the screen */
                if (start_display.in_node) {
                    /* get true character location */
                    view_port_mpos_2_xy(display_viewport,
                                        &start_display, &inq_pos, 0);
                    inq_pos.at_y = dy/2;
                    view_port_draw_at(display_viewport, &start_display,
                                      &inq_pos);
                    }
                break;

            case('d'):
                /* move the viewport down the conceptual page */
                /*   half the viewport height */
                /* get height of viewport -- it may have changed */
                dy = view_port_get_hi(display_viewport);
                /* find the character at pixel (0,dy/2) */
                inq_pos.at_x = 0;
                inq_pos.at_y = dy/2;
                view_port_xy_2_mpos(display_viewport, &inq_pos,
                                    &start_display);
                /* if a character was found, display it */
                /*   at top of screen */
                if (start_display.in_node)
                    /* get true character location */
                    view_port_mpos_2_xy(display_viewport,
                                        &start_display, &inq_pos, 0);
                    inq_pos.at_y = 0;
                    view_port_draw_at(display_viewport, &start_display,
                                      &inq_pos);
                break;


            case ('l'):
                /* move left on the conceptual page half the viewport width */
                /* get width of viewport */
                dx = view_port_get_width(display_viewport);
                /* find the character if any at pixel (0,0) */
                view_port_xy_2_mpos(display_viewport, &zero_zero,
                                    &start_display);
                /* if a character was found, redisplay */
                if (start_display.in_node) {
                     /* get true character location */
                     if (0 == view_port_mpos_2_xy(display_viewport,
                                                  &start_display, &inq_pos, 0))
                         break;
                     /* move half a viewport left */
                     inq_pos.at_x += (dx/2);
                     view_port_draw_at(display_viewport, &start_display,
                                       &inq_pos);
                     }
                break;

            case ('r'):
                /* move right on conceptual page half the viewport width */
                /* get width of viewport */
                dx = view_port_get_width(display_viewport);
                /* find the character if any at pixel (0,0) */
                view_port_xy_2_mpos(display_viewport, &zero_zero,
                                    &start_display);
                /* if a character was found, redisplay */
                if (start_display.in_node) {
                     /* get true character location */
                     if (0 == view_port_mpos_2_xy(display_viewport,
                                                  &start_display, &inq_pos, 0))
                         break;
                     /* move half a viewport right */
                     inq_pos.at_x -= (dx/2);
                     view_port_draw_at(display_viewport, &start_display,
                                       &inq_pos);
                     }
                break;

            case ('t'):
                /* move to display top of file */
                /* get root node of model */
                /*   and move to first model position */
                start_display.in_node = model_get_root(display_model);
                start_display.char_index = 0;
                model_pos_get_next(&start_display, &start_display);
                /* if a character was found, display it at pixel (0,0) */
                if (start_display.in_node)
                    view_port_draw_at(display_viewport,
                                      &start_display, &zero_zero);
                break;

            case ('b'):
                /* move to display end of file */
                /* get root node of model */
                /* get last node dominated by root */
                start_display.in_node =
                  node_get_last( model_get_root(display_model) );
                /* quit if no character found */
                if (!start_display.in_node) break;
                /* get index of last character for that node */
                start_display.char_index =
                  node_get_max_index(start_display.in_node);
                /* find the conceptual page containing the model position */
                last_page = format_mpos_2_phead(display_format, &start_display);
                /* quit if no page found */
                if (!last_page) break;
                /* find last cline on last page */
                last_line = phead_mpos_2_cline(last_page, &start_display);
                /* quit if no line found */
                if (!last_line) break;
                /* set up a screen_xy for the bottom of the page */
                /*   at the height of the last line */
                inq_pos.at_x = 0;
                inq_pos.at_y =
                  view_port_get_hi(display_viewport)
                  - cline_get_hi(last_line);
                view_port_draw(display_viewport, last_line, &inq_pos);
                break;

            default:
                break;
        }
    }

    /* release dynamically allocated storage */
    view_port_free(display_viewport);
    model_free(display_model);
}

