#include <stdio.h>
#include <apollo/base.h>
#include <apollo/gpr.h>
#include <tml/tml_c_types.h>
#include <tml/tml_dp_graph.hxx>

extern long create_bitmap();
extern void enable_all_keys();

main()
{
    tml_c_model      display_model;
    tml_c_view_port  display_viewport;
    C_MODEL_POS      start_display;
    long             bitmap_desc;
    char             unob;
    gpr_$event_t     ev_type;
    char             ev_value = '\0';
    gpr_$position_t  ret_coord;
    status_$t        status;
    long             dx, dy;    /* size of viewport */
    C_SCREEN_XY      inq_pos;   /* any (x,y) position on the screen */
    tml_c_phead      last_page; /* used for going to bottom of file */
    tml_c_cline      last_line; /* used for going to bottom of file */

    /*******************************************/
    tml_c_format     use_format;
    tml_c_property   wrap_property,
                     margin_property;
    /*******************************************/

    /* initialize TMLib */
    tml_init();

    /* associate all viewport objects with the GPR graphic system */
    tml_c_tml_dp_set_dflt_vpftbl(tml_c_tml_dp_get_gpr_vpftbl());

    /* create a model from the ASCII file test_file */
    display_model = tml_c_read_uasc_path("test_file", 9);
    if (!display_model)
        exit (1);

    /* allocate a viewport */
    display_viewport = view_port_class_alloc();
    /* create a graphics display area */
    /* use a routine that depends on the graphics system */
    bitmap_desc = create_bitmap();

    /* assign the display area to the viewport */
    view_port_set_handle(display_viewport, &bitmap_desc);

    /********************************************/
    /* create a format */
    use_format = format_class_alloc();
    if (!use_format) {
        fprintf(stderr, "unable to create new format\n");
        exit(1);
        }
    /* attach format to viewport, setting format page size */
    /*   to 4 inches wide by 5 inches high for displays with */
    /*   resolution of 72 pixels per inch */
    view_port_attach_format(display_viewport, use_format, (4 * 72), (5 * 72));
    /* attach model to format that is associated with viewport */
    format_attach(use_format, display_model);
    /*******************************************/

    /* get root node of model */
    start_display.in_node = model_get_root(display_model);
    /* quit on failure */
    if (!start_display.in_node)
        exit(1);

    /*******************************************/
    /* make word wrap property */
    wrap_property = tml_c_make_wrap_prop(word_wrap, display_model);
    if (!wrap_property) {
        fprintf(stderr, "unable to make wrap property\n");
        exit(1);
        }
    /* attach wrap property to root node of model */
    tml_c_connect_prop(wrap_property, start_display.in_node);

    /* make margin property */
    /* set top and bottom margins to 10 pixels */
    /*   left and right margins to 20 pixels */
    margin_property = tml_c_make_margin_prop(20, 20, 10, 10, display_model);
    if (!margin_property) {
        fprintf(stderr, "unable to make margin property\n");
        exit(1);
        }
    /* attach margin property to root node of model */
    tml_c_connect_prop(margin_property, start_display.in_node);
    /******************************************/

    start_display.char_index = 0;
    /* get first character in model */
    model_pos_get_next(&start_display, &start_display);

    /* display top of file */
    tml_dp_damage_repair((void*) (view_port_get_handle(display_viewport)));

    /* enable keyboard */
    enable_all_keys();

    while (ev_value != 'e') { /* typing 'e' exits the program */

        unob = gpr_$event_wait(&ev_type, (unsigned char*) &ev_value,
                               &ret_coord, &status);

        switch (ev_value) {

            case ('u'):
                /* move viewport up the conceptual page */
                /*   half the viewport height */
                /* get height of viewport -- it may have changed */
                dy = phead_get_hi(display_viewport);
                /* find character at top of viewport */
                view_port_xy_2_mpos(display_viewport,
                                    &zero_zero, &start_display);
                /* if a character was found */
                /*   display it halfway down the screen */
                if (start_display.in_node) {
                    /* get true character location */
                    view_port_mpos_2_xy(display_viewport,
                                        &start_display, &inq_pos, 0);
                    inq_pos.at_y = dy/2;
                    view_port_draw_at(display_viewport,
                                      &start_display, &inq_pos);
                    }
                break;

            case ('d'):
                /* move the viewport down the conceptual page */
                /*   half the viewport height */
                /* get height of viewport -- it may have changed */
                dy = phead_get_hi(display_viewport);
                /* find the character at pixel (0,dy/2) */
                inq_pos.at_x = 0;
                inq_pos.at_y = dy/2;
                view_port_xy_2_mpos(display_viewport, &inq_pos, &start_display);
                /* if a character was found, display it */
                /*   at top of screen */
                if (start_display.in_node) {
                    view_port_mpos_2_xy(display_viewport, &start_display,
                                        &inq_pos, 0);
                    inq_pos.at_y = 0;
                    view_port_draw_at(display_viewport,
                                      &start_display, &inq_pos);
                    }
                break;


            case ('l'):
                /* move left on the conceptual page half the viewport width */
                /* get width of viewport */
                dx = view_port_get_width(display_viewport);
                /* find the character if any at pixel (0,0) */
                view_port_xy_2_mpos(display_viewport,
                                    &zero_zero, &start_display);
                /* if a character was found, redisplay */
                if (start_display.in_node) {
                     /* get true character location */
                     if (0 == view_port_mpos_2_xy(display_viewport,
                                                  &start_display, &inq_pos, 0))
                         break;
                     /* move half a viewport left */
                     inq_pos.at_x += (dx/2);
                     view_port_draw_at(display_viewport,
                                       &start_display, &inq_pos);
                     }
                break;

            case ('r'):
                /* move right on conceptual page half the viewport width */
                /* get width of viewport */
                dx = view_port_get_width(display_viewport);
                /* find the character if any at pixel (0,0) */
                view_port_xy_2_mpos(display_viewport,
                                    &zero_zero, &start_display);
                /* if a character was found, redisplay */
                if (start_display.in_node) {
                     /* get true character location */
                     if (0 == view_port_mpos_2_xy(display_viewport,
                                                  &start_display, &inq_pos, 0))
                         break;
                     /* move half a viewport right */
                     inq_pos.at_x -= (dx/2);
                     view_port_draw_at(display_viewport,
                                       &start_display, &inq_pos);
                     }
                break;

            case ('t'):
                /* move to display top of file */
                /* get root node of model */
                /*   and move to first model position */
                start_display.in_node = model_get_root(display_model);
                start_display.char_index = 0;
                model_pos_get_next(&start_display, &start_display);
                /* if a character was found, display it at pixel (0,0) */
                if (start_display.in_node)
                    view_port_draw_at(display_viewport,
                                      &start_display, &zero_zero);
                break;

            case ('b'):
                /* move to display end of file */
                /* get root node of model */
                /* get last node dominated by root */
                start_display.in_node =
                  node_get_last( model_get_root(display_model) );
                /* quit if no character found */
                if (!start_display.in_node) break;
                /* get index of last character for that node */
                start_display.char_index =
                  node_get_max_index(start_display.in_node);
                /* find the conceptual page containing the model position */
                last_page = format_mpos_2_phead(use_format, &start_display);
                /* quit if no page found */
                if (!last_page) break;
                /* find last cline on last page */
                last_line = phead_mpos_2_cline(last_page, &start_display);
                /* quit if no line found */
                if (!last_line) break;
                /* set up a screen_xy for the bottom of the page */
                /*   at the height of the last line */
                inq_pos.at_x = 0;
                inq_pos.at_y =
                  view_port_get_hi(display_viewport)
                  - cline_get_hi(last_line);
                view_port_draw(display_viewport, last_line, &inq_pos);
                break;

            default:
                break;
        }
    }
    /* release dynamically allocated storage */
    view_port_free(display_viewport);
    format_free(use_format);
    model_free(display_model);
}
