#include <stdio.h>
#include <apollo/base.h>
#include <apollo/gpr.h>
#include <tml/tml_objects.hxx>
#include <tml/tml_functions.hxx>
#include <tml/tml_dp_graph.hxx>

#ifdef __cplusplus
extern "C" long create_bitmap();
extern "C" void enable_all_keys();
#endif

extern long create_bitmap();
extern void enable_all_keys();

main()
{
    model           *display_model;
    view_port       *display_viewport;
    format          *display_format;
    model_pos        start_display;
    long             bitmap_desc;
    char             unob;
    gpr_$event_t     ev_type;
    char             ev_value;
    gpr_$position_t  ret_coord;
    status_$t        status;

    /*****************************************/
    long       dx, dy;    /* size of viewport */
    screen_xy  inq_pos;   /* any (x,y) coordinate on the screen */
    phead     *last_page; /* used for going to bottom of file */
    cline     *last_line; /* used for going to bottom of file */
    /*****************************************/

    /* associate all viewport objects with the GPR graphic system */
    tml_dp_set_dflt_vpftbl(tml_dp_get_gpr_vpftbl());

    /* create a model from the ASCII file test_file */
    display_model = read_uasc("test_file", 9);

    /* allocate a viewport */
    display_viewport = new view_port;
    /* create a graphics display area, */
    /*   using a routine that depends on the graphics system */
    bitmap_desc = create_bitmap();

    /* assign the display area to the viewport */
    display_viewport->set_handle(&bitmap_desc);

    /* create a format and attach it to the viewport */
    display_viewport->attach((format*) 0);
    /* get a handle to the format created by the last call */
    display_format = display_viewport->get_format();
    /* attach the model to the format that is associated with the viewport */
    display_format->attach(display_model);

    /* fill viewport with text; first character placed at pixel (0,0) */
    tml_dp_damage_repair(display_viewport->get_handle());

    /* enable keyboard */
    enable_all_keys();

    /*****************************************/
    while (ev_value != 'e') { /* typing 'e' exits the program */

        unob = gpr_$event_wait(&ev_type, (unsigned char*) &ev_value,
                               &ret_coord, &status);

        switch (ev_value) {

            case ('u'):
                /* move the viewport up the conceptual page */
                /*   half the viewport height */
                /* get height of viewport -- it may have changed */
                dy = display_viewport->get_hi();
                /* find character at top of viewport */
                start_display = display_viewport->xy_2_mpos(zero_zero);
                /* if character found, display it at halfway down the screen */
                if (start_display.in_node) {
                    inq_pos.at_x = 0;
                    inq_pos.at_y = dy/2;
                    display_viewport->draw(start_display, inq_pos);
                    }
                break;

            case ('d'):
                /* move the viewport down the conceptual page */
                /*   half the viewport height */
                /* get height of viewport -- it may have changed */
                dy = display_viewport->get_hi();
                /* find the character at pixel (0,dy/2) */
                inq_pos.at_x = 0;
                inq_pos.at_y = dy/2;
                start_display = display_viewport->xy_2_mpos(inq_pos);
                /* if character found, display it at pixel (0,0) */
                if (start_display.in_node)
                    display_viewport->draw(start_display, zero_zero);
                break;

            case ('l'):
                /* move left on the conceptual page */
                /*   half the viewport width */
                /* get width of viewport */
                dx = display_viewport->get_width();
                /* find the character at pixel (0,0) */
                start_display = display_viewport->xy_2_mpos(zero_zero);
                /* if character found, display it at pixel (dx/2,0) */
                if (start_display.in_node) {
                    /* get true character location */
                    inq_pos = display_viewport->mpos_2_xy(start_display);
                    /* move half a viewport left */
                    inq_pos.at_x += (dx/2);
                    display_viewport->draw(start_display, inq_pos);
                    }
                break;

            case ('r'):
                /* move right on the conceptual page */
                /*   half the viewport width */
                /* get width of viewport */
                dx = display_viewport->get_width();
                /* find character at pixel (0,0) */
                start_display = display_viewport->xy_2_mpos(zero_zero);
                /* if character found, display it at pixel (-dx/2,0) */
                if (start_display.in_node) {
                    /* get true character location */
                    inq_pos = display_viewport->mpos_2_xy(start_display);
                    /* move half a viewport right */
                    inq_pos.at_x -= (dx/2);
                    display_viewport->draw(start_display, inq_pos);
                    }
                break;

            case ('t'):
                /* move to display top of file */
                /* get root node of model */
                /* get first model position */
                start_display.in_node = display_model->get_root();
                start_display.char_index = 0;
                start_display = start_display.get_next();
                /* if character found, display it at pixel (0,0) */
                if (start_display.in_node)
                    display_viewport->draw(start_display, zero_zero);
                break;

            case ('b'):
                /* move to display bottom of file */
                /* get root node of model */
                /* get last node dominated by root */
                /* get the root node, and last node dominated by the root */
                start_display.in_node = display_model->get_root()->get_last();
                /* quit if no node found */
                if (!start_display.in_node) break;
                /* get index of last character in last node */
                start_display.char_index =
                  start_display.in_node->get_max_index();
                /* find conceptual page containiong that model position */
                last_page = display_format->mpos_2_phead(start_display);
                /* quit if no page found */
                if (!last_page) break;
                /* find last cline on last page */
                last_line = last_page->mpos_2_hcline(start_display);
                /* quit if no line found */
                if (!last_line) break;
                /* set up a screen_xy for bottom of page */
                /*   at height of last line */
                inq_pos.at_x = 0;
                inq_pos.at_y = display_viewport->get_hi() - last_line->get_hi();
                display_viewport->draw(last_line, inq_pos);
                break;

            default:
                break;
        }
    }
    /* release dynamically allocated storage */
    delete display_viewport;
    /*****************************************/
}
