
#include <apollo/base.h>
#include <apollo/gpr.h>

long create_bitmap()
{
    gpr_$disp_char_t   disp;
    short              disp_len_returned;
    gpr_$offset_t      bitmap_size;
    status_$t          status;
    gpr_$bitmap_desc_t return_bitmap;

    bitmap_size.x_size = 1024;
    bitmap_size.y_size = 1024;

    /* brook no interference from other gpr_$direct mode programs */
    gpr_$acquire_display(&status);

    /* inquire number of planes in display */
    /* initializing gpr with this number of planes allows correct behavior */
    /*   for both monochrome and color nodes */
    gpr_$inq_disp_characteristics(gpr_$direct,
                                  stream_$stdout,
                                  (short) sizeof(gpr_$disp_char_t),
                                  &disp,
                                  &disp_len_returned,
                                  &status);

    /* initialize gpr: */
    gpr_$init(gpr_$direct,                      /* using direct mode ... */
              stream_$stdout,                   /* ... in output pad of DM window */
              bitmap_size,                      /* ... using 1024 x 1024 pixel "subwindow" */
              (gpr_$rgb_plane_t) disp.n_planes, /* ... with interest in all display planes */
              &return_bitmap,                   /* create this initial bitmap */
              &status);

    /* returned bitmap has default foreground and background settings */
    /*  -- reset to foreground and background values of DM window used */
    /* NOTE: the inquire calls work only in gpr direct (as here) and */
    /* in gpr frame mode. This code is not necessary in other gpr modes. */
    gpr_$set_text_value(gpr_$inq_foreground(&status), &status);
    gpr_$set_text_background_value(gpr_$inq_background(&status), &status);

    /* release display */
    gpr_$release_display(&status);

    /* return initial bitmap */
    return return_bitmap;
}

