{ gpr_example_2.pas }
PROGRAM gpr_example_2;

%include '/sys/ins/base.ins.pas';
%include '/sys/ins/gpr.ins.pas';
%include '/sys/ins/error.ins.pas';
%include '/usr/include/tml/tml.ins.pas';

PROCEDURE create_bitmap(OUT return_bitmap : integer32); EXTERN;
PROCEDURE enable_all_keys; EXTERN;
PROCEDURE tml_dp_damage_repair(viewport_handle : tml_c_bitmap_id); EXTERN;

TYPE
   FILE_NAME_TYPE     = ARRAY[0..26] of CHAR;
   PTR_TO_TML_C_MODEL = ^TML_C_MODEL;

VAR
    filename          :array[0..8] of char;
    filename_length   :integer32;
    display_model     :TML_C_MODEL;
    display_viewport  :TML_C_VIEW_PORT;
    display_format    :TML_C_FORMAT;
    start_display     :C_MODEL_POS;
    p_start_display   :PTR_TO_C_MODEL_POS   := addr(start_display);
    bitmap            :INTEGER32;
    dy                :INTEGER32;           { viewport height }
    dx                :INTEGER32;           { viewport width }
    inq_pos           :C_SCREEN_XY;         { any x,y coordinate on the screen }
    p_inq_pos         :PTR_TO_C_SCREEN_XY   := addr(inq_pos);
    last_page         :TML_C_PHEAD;         { used for going to bottom of file }
    last_line         :TML_C_CLINE;         { used for going to bottom of file }


{************************************************************}
    use_format        :TML_C_FORMAT;
    wrap_property,
    margin_property   :TML_C_PROPERTY;
{************************************************************}

    dummy_integer     :integer32;
    dummy_format      :TML_C_FORMAT;
    ten               :INTEGER32            := 10;
    twenty            :INTEGER32            := 20;
    unob              :boolean;
    ev_type           :gpr_$event_t;
    ev_value          :char                 := '0';
    ret_coord         :gpr_$position_t;
    status            :status_$t;

BEGIN

{ initialize TMLib }
tml_init;

{ associate all viewport objects with the GPR graphic system }
tml_c_tml_dp_set_dflt_vpftbl(tml_c_tml_dp_get_gpr_vpftbl());

{ create a model from the ASCII file test_file }
filename := 'test_file';
filename_length := SIZEOF(filename);
display_model := tml_c_read_uasc_path(filename[0], filename_length);

IF (display_model <> NIL) THEN
   BEGIN
   { allocate a viewport }
   display_viewport := view_port_class_alloc;

   { create a graphics display area }
   create_bitmap(bitmap);

   { assign the display area to the viewport }
   dummy_integer := view_port_set_handle(display_viewport,
                                         tml_c_bitmap_id(addr(bitmap)));

{*****************************************************************}
   { create a format }
   use_format := format_class_alloc;

   IF use_format = NIL THEN
     writeln('unable to create new format')
   ELSE BEGIN
     { attach format to viewport, setting format page size }
     {  to 4 inches wide by 5 inches high for displays with }
     {  resolution of 72 pixels per inch }
     dummy_integer :=
       view_port_attach_format(display_viewport, use_format,
                               integer32(4*72), integer32(5*72));

     { attach model to format that is associated with viewport }
     format_attach(use_format, display_model);
{*****************************************************************}

     { get root node of model }
     start_display.in_node := model_get_root(display_model);

     { quit on failure }
     IF (start_display.in_node <> NIL) THEN
        BEGIN
{***************************************************************}
        { make word wrap property }
        wrap_property := tml_c_make_wrap_prop(word_wrap, display_model);
        IF wrap_property = NIL THEN
           writeln('unable to make wrap_property')
        ELSE BEGIN
           { attach word wrap property to root node of model }
           tml_c_connect_prop(wrap_property, start_display.in_node);

           { make margin property }
           { set top and bottom margins to 10 pixels }
           {  left and right margins to 20 pixels }
           margin_property := tml_c_make_margin_prop(twenty, twenty,
                                                     ten,    ten,
                                                     display_model);

           IF margin_property = NIL THEN
              writeln('unable to make margin property')
           ELSE BEGIN
              { attach margin property to root node of model }
              tml_c_connect_prop(margin_property, start_display.in_node);
{***************************************************************}

              start_display.char_index := 0;

              { get first character in file }
              p_start_display := model_pos_get_next(start_display,
                                                    start_display);

              { display top of file }
              tml_dp_damage_repair(view_port_get_handle(display_viewport));

              { enable keyboard }
              enable_all_keys();

              WHILE (ev_value <> 'e') DO
              { typing 'e' exits the program }

                 BEGIN

                 unob := gpr_$event_wait(ev_type, ev_value, ret_coord, status);

                 CASE ev_value OF

                 'u' :
                    BEGIN
                    { move viewport up the conceptual page }
                    {   half the viewport height }
                    { get height of viewport -- it may have changed }
                    dy := phead_get_hi(display_viewport);
                    { find character at top of viewport }
                    p_start_display :=
                      view_port_xy_2_mpos(display_viewport,
                                          zero_zero, start_display);
                    { if a character was found }
                    {   display it halfway down the screen }
                    IF (start_display.in_node <> NIL) THEN
                       BEGIN
                       { get true character location }
                       p_inq_pos :=
                         view_port_mpos_2_xy(display_viewport,
                                             start_display, inq_pos, NIL);
                       inq_pos.at_y := dy div 2;
                       view_port_draw_at(display_viewport, start_display,
                                         inq_pos);
                       END;
                    END;

                 'd' :
                    BEGIN
                    { move the viewport down the conceptual page }
                    {   half the viewport height }
                    { get height of viewport -- it may have changed }
                    dy := phead_get_hi(display_viewport);
                    { find the character at pixel (0,dy/2) }
                    inq_pos.at_x := 0;
                    inq_pos.at_y := dy div 2;
                    p_start_display :=
                      view_port_xy_2_mpos(display_viewport,
                                          inq_pos, start_display);
                    { if a character was found, display it }
                    {   at top of screen }
                    IF (start_display.in_node <> NIL) THEN
                       BEGIN
                       { get true character location }
                       p_inq_pos :=
                         view_port_mpos_2_xy(display_viewport,
                                             start_display, inq_pos, NIL);
                       inq_pos.at_y := 0;
                       view_port_draw_at(display_viewport, start_display,
                                         inq_pos);
                       END;
                    END;

                 'l' :
                    BEGIN
                    { move left on the conceptual page half the viewport width }
                    { get width of viewport }
                    dx := view_port_get_width(display_viewport);
                    { find the character if any at pixel (0,0) }
                    p_start_display :=
                      view_port_xy_2_mpos(display_viewport,
                                          zero_zero, start_display);
                    { if a character was found, redisplay }
                    IF (start_display.in_node <> NIL) THEN
                       BEGIN
                       { get true character location }
                       p_inq_pos :=
                         view_port_mpos_2_xy(display_viewport,
                                             start_display, inq_pos, NIL);
                       { move half a viewport left }
                       inq_pos.at_x := inq_pos.at_x + (dx div 2);
                       view_port_draw_at(display_viewport,
                                         start_display, inq_pos);
                       END;
                    END;

                 'r' :
                    BEGIN
                    { move right on conceptual page half the viewport width }
                    { get width of viewport }
                    dx := view_port_get_width(display_viewport);
                    { find the character if any at pixel (0,0) }
                    p_start_display :=
                      view_port_xy_2_mpos(display_viewport,
                                          zero_zero, start_display);
                    { if a character was found, redisplay }
                    IF (start_display.in_node <> NIL) THEN
                       BEGIN
                       { get true character location }
                       p_inq_pos :=
                         view_port_mpos_2_xy(display_viewport,
                                             start_display, inq_pos, NIL);
                       { move half a viewport right }
                       inq_pos.at_x := inq_pos.at_x - (dx div 2);
                       view_port_draw_at(display_viewport,
                                         start_display, inq_pos);
                       END;
                    END;

                 't' :
                    BEGIN
                    { move to display top of file }
                    { get root node of model }
                    {  and move to first model position }
                    start_display.in_node :=
                      model_get_root(display_model);
                    start_display.char_index := 0;
                    p_start_display :=
                      model_pos_get_next(start_display, start_display);
                    { if a character was found, display it at pixel (0,0) }
                    IF (start_display.in_node <> NIL) THEN
                        view_port_draw_at(display_viewport, start_display,
                                          zero_zero);
                    END;

                 'b' :
                    BEGIN
                    { move to display end of file }
                    { get root node of model }
                    { get last node dominated by root }
                    start_display.in_node :=
                      node_get_last(model_get_root(display_model));
                    { quit if no character found }
                    IF (start_display.in_node <> NIL) THEN
                       BEGIN
                       { get index of last character for that node }
                       start_display.char_index :=
                          node_get_max_index(start_display.in_node);
                       { find conceptual page containing the model position }
                       last_page :=
                         format_mpos_2_phead(use_format, start_display);
                       { quit if no page found }
                       IF (last_page <> NIL) THEN
                          BEGIN
                          { find last cline on last page }
                          last_line := phead_mpos_2_cline(last_page,
                                                      start_display);
                          { quit if no line found }
                          IF (last_line <> NIL) THEN
                             BEGIN
                             { set up a screen_xy for bottom of page }
                             {  at height of last line }
                             inq_pos.at_x := 0;
                             inq_pos.at_y :=
                               view_port_get_hi(display_viewport)
                               - cline_get_hi(last_line);
                             view_port_draw(display_viewport, last_line,
                                            inq_pos);
                             END;

                          END;

                       END;

                    END;

                 END;

                 END;

{***************************************************************}
              END;

           END;
{***************************************************************}

        END;

      { release dynamically allocated storage }
      view_port_free(display_viewport);
      format_free(display_format);
      model_free(display_model);
      END;

   END;

END.

