#include <stdio.h>
#include <tml/tml_c_types.h>
#include <tml/tml_os_depends.h>
#include <tml_exam/tml_x_exam_kbd.h>

    /*****************************************/
#include <dlg/Xlib.h>
#include <dlg/Xutil.h>
#include <tml/tml_dp_x_graph.h>
#include <tml/tml_dp_graph.hxx>

#ifdef not_vanilla_C
extern struct window_info* make_window(long mask_selection);
#else
extern struct window_info* make_window();
#endif

main()
{
    tml_c_model      display_model;
    tml_c_view_port  display_viewport;
    tml_c_format     display_format;
    C_MODEL_POS      start_display;

    /***************************************/
    struct window_info    *window_desc;
           XEvent          return_event;  /* records events */
           KeySym          keysym_return;
           XComposeStatus  status_return;
           char            return_char;
    /***************************************/
    long             mask_selection =
                       KeyPressMask |       /* Ask XWindows for notice */
                       ButtonPressMask |    /* of key, mouse, expose, */
                       ExposureMask |       /* and window size and */
                       StructureNotifyMask; /* location change events */

    int              c;
    long             dx, dy;  /* size of viewport */
    C_SCREEN_XY      inq_pos; /* any (x,y) coordinate on the screen */

    tml_c_v_cursor  the_cursor;
    C_SCREEN_XY     curs_loc;
    C_MODEL_POS     curs_pos;
    C_MODEL_POS     hold_pos;
    tml_c_cline     old_top_line;
    tml_c_cline     new_top_line;
    tml_c_cline     first_line;
    tml_c_phead     first_page;
    C_XY_COORD      at_coord;
    char            text[10];
    int             textlen;
    tml_c_phead     on_phead;
    tml_c_cline     in_cline;
    csm_c_char_set  in_char_set;
    tml_c_font_tbl  use_font;
    long            char_val;

    /* initialize TMLib */
    tml_init();

    /***************************************/
    /* set up XWindows */
    /* get information about display, etc. */
    window_desc = make_window(mask_selection);
    if (!window_desc) {
        fprintf(stderr, "unable to initialize X\n");
        exit(1);
        }
    /* set font display */
    /* MUST BE DONE BEFORE CREATING A MODEL */
    tml_dp_set_font_display(window_desc->display);
    /***************************************/

    /* create a model from the ASCII file test_file */
    display_model = tml_c_read_uasc_path("test_file", 9);
    if (!display_model)
        exit (1);

    /* allocate a viewport */
    display_viewport = view_port_class_alloc();

    /* assign the display area to the viewport */
    view_port_set_handle(display_viewport, window_desc);

    /* create a format */
    /* attach it to the viewport */
    view_port_attach_format(display_viewport, 0, 0, 0);
    /* get a handle for the format created by the last call */
    display_format = view_port_get_format(display_viewport);
    /* attach model to format that is associated with viewport */
    format_attach(display_format, display_model);

    /* allocate cursor */
    /* attach cursor to viewport at first character */
    the_cursor = v_cursor_class_alloc();
    if (!the_cursor) {
        fprintf(stderr, "unable to allocate a cursor\n");
        exit(1);
        }
    view_port_attach_cursor(display_viewport, the_cursor, &zero_zero);
    v_cursor_enable(the_cursor, 0);

    while (1) {
        XNextEvent(window_desc->display, &return_event);
        switch (return_event.type) {

            case ConfigureNotify:
                /* resize event */
                tml_dp_record_new_size(
                  (void*) view_port_get_handle(display_viewport),
                  return_event.xconfigure.width,
                  return_event.xconfigure.height);
                break;

            case Expose:
                /* refresh event */
                tml_dp_add_clip_rect(view_port_get_handle(display_viewport),
                                     return_event.xexpose.x,
                                     return_event.xexpose.y,
                                     return_event.xexpose.width,
                                     return_event.xexpose.height);
                if (return_event.xexpose.count == 0)
                    tml_dp_damage_repair(
                      (void*) view_port_get_handle(display_viewport));
                break;

            case ButtonPress:
                /* mouse event */
                /* ignore all except ButtonPress on Button1 */
                if (Button1 == return_event.xbutton.button) {
                    /* move TMLib cursor to where pointing cursor was at click */
                    inq_pos.at_x = return_event.xbutton.x;
                    inq_pos.at_y = return_event.xbutton.y;
                    v_cursor_xy_set(the_cursor, &inq_pos);
                    }
                break;

            case KeyPress:
                textlen = XLookupString((XKeyEvent*) &return_event,
                                        text, 10, &keysym_return, 0);
                text[textlen] = (char)0;
                switch (keysym_return) {

                    /* The constant values for keyboard return values */
                    /*   in the case statements come from tml_x_exam_kbd.h. */
                    /* Keysym return values are sensitive to the keyboard map */
                    /*   that is in effect at runtime. */
                    /* To use different values, edit tml_x_exam_kbd.h. */
                    /* NOTE: On Apollo systems running Xapollo with */
                    /*   the -k option, these keys are reserved for the */
                    /*   Display Manager and are never passed to the program. */

                    case tmlScrollUp:
                        /* up-box-arrow key: scroll up half a viewport */
                        /* get (x,y) position of cursor in viewport */
                        v_cursor_get_raw_location(the_cursor, &curs_loc);
                        /* get height of viewport -- it may have changed */
                        dy = view_port_get_hi(display_viewport);
                        /* get top line of viewport */
                        old_top_line =
                          view_pos_get_cline(view_port_get_top_line(
                                             display_viewport));
                        /* find the character if any at top of viewport */
                        view_port_xy_2_mpos(display_viewport, 
                                            &zero_zero, &start_display);
                        /* if a character was found */
                        /*   display it halfway down the screen */
                        if (start_display.in_node) {
                            view_port_mpos_2_xy(display_viewport,
                                                &start_display, &inq_pos, 0);
                            inq_pos.at_y = dy/2;
                            view_port_draw_at(display_viewport,
                                              &start_display, &inq_pos);
                            /* if cursor character not on screen, */
                            /*   set cursor at position it had before scroll */
                            v_cursor_get_model_pos(the_cursor, &curs_pos);
                            if (0 == view_port_mpos_2_vpos(display_viewport,
                                                           &curs_pos))
                                v_cursor_xy_set(the_cursor, &curs_loc);
                            }
                        /* if no movement and line above exists, */
                        /*   go to top of file */
                        /* get new top line of viewport */
                        new_top_line =
                          view_pos_get_cline(view_port_get_top_line(
                                             display_viewport));
                        if (new_top_line &&
                          (new_top_line == old_top_line) &&
                          (cline_get_prior(new_top_line))) {
                            /* no movement and prior line exists */
                            /* get first position in model */
                            start_display.in_node =
                              model_get_root(format_get_model(display_format));
                            start_display.char_index = 0;
                            model_pos_get_next(&start_display, &start_display);
                            /* get and correct position of first position */
                            /*   on its conceptual page */
                            first_page =
                              format_mpos_2_phead(display_format,
                                                  &start_display);
                            first_line =
                              phead_mpos_2_cline(first_page, &start_display);
                            cline_mpos_2_xy(first_line,
                                            &start_display, &at_coord);
                            inq_pos.at_x = at_coord.x_coord
                              + view_port_get_hoffset(display_viewport);
                            inq_pos.at_y = 0;
                            view_port_draw_at(display_viewport,
                                              &start_display, &inq_pos);
                            /* if cursor character not on screen, */
                            /*   set cursor at position it had before scroll */
                            v_cursor_get_model_pos(the_cursor, &curs_pos);
                            if (0 == view_port_mpos_2_vpos(display_viewport,
                                                           &curs_pos))
                                v_cursor_xy_set(the_cursor, &curs_loc);
                            }
                        break;

                    case tmlScrollDown:
                        /* down-box arrow key: scroll down half a viewport */
                        /* get (x,y) position of cursor in viewport */
                        v_cursor_get_raw_location(the_cursor, &curs_loc);
                        /* get height of viewport -- it may have changed */
                        dy = view_port_get_hi(display_viewport);
                        /* find character halfway down and furthest left */
                        inq_pos.at_x = 0;
                        inq_pos.at_y = dy/2;
                        view_port_xy_2_mpos(display_viewport,
                                            &inq_pos, &start_display);
                        /* if character found, display at top of viewport */
                        if (start_display.in_node) {
                            /* get true location of character */
                            view_port_mpos_2_xy(display_viewport,
                                                &start_display, &inq_pos, 0);
                            inq_pos.at_y = 0;
                            /* display at new coordinates */
                            view_port_draw_at(display_viewport,
                                              &start_display, &inq_pos);
                            /* if cursor character not on screen, */
                            /*   set cursor at position it had before scroll */
                            v_cursor_get_model_pos(the_cursor, &curs_pos);
                            if (!view_port_mpos_2_vpos(display_viewport,
                                                       &curs_pos))
                                v_cursor_xy_set(the_cursor, &curs_loc);
                            }
                        break;

                    case tmlScrollLeft:
                        /* left-box arrow key: */
                        /*   scroll left 1/4 viewport width */
                        /* get (x,y) position of cursor in viewport */
                        v_cursor_get_raw_location(the_cursor, &curs_loc);
                        /* get width of viewport -- it may have changed */
                        dx = view_port_get_width(display_viewport);
                        /* find character at pixel (0,0) */
                        view_port_xy_2_mpos(display_viewport,
                                            &zero_zero, &start_display);
                        if (start_display.in_node) {
                            /* get (x,y) of topmost character */
                            view_port_mpos_2_xy(display_viewport,
                                                &start_display, &inq_pos, 0);
                            /* increment x coordinate by 1/4 viewport width */
                            inq_pos.at_x += dx/4;
                            /* display at new coordinates */
                            view_port_draw_at(display_viewport,
                                              &start_display, &inq_pos);
                            /* if cursor character not on screen, */
                            /*   set cursor at position it had before scroll */
                            v_cursor_get_caged_location(the_cursor, &inq_pos);
                            if (inq_pos.at_x > dx)
                                v_cursor_xy_set(the_cursor, &curs_loc);
                            }
                        break;

                    case tmlScrollRight:
                        /* right-box arrow key: */
                        /*   scroll right 1/4 viewport width */
                        /* get (x,y) position of cursor */
                        v_cursor_get_raw_location(the_cursor, &curs_loc);
                        /* get width of viewport -- it may have changed */
                        dx = view_port_get_width(display_viewport);
                        /* find character at pixel (0,0) */
                        view_port_xy_2_mpos(display_viewport,
                                            &zero_zero, &start_display);
                        if (start_display.in_node) {
                            /* get (x,y) of topmost character */
                            view_port_mpos_2_xy(display_viewport,
                                                &start_display, &inq_pos, 0);
                            /* decrement x coordinate by 1/4 viewport width */
                            inq_pos.at_x -= dx/4;
                            /* display at new coordinates */
                            view_port_draw_at(display_viewport,
                                              &start_display, &inq_pos);
                            /* if cursor character not on screen, */
                            /*   set cursor at position it had before scroll */
                            v_cursor_get_caged_location(the_cursor, &inq_pos);
                            if (inq_pos.at_x < 0)
                                v_cursor_xy_set(the_cursor, &curs_loc);
                            v_cursor_get_raw_location(the_cursor, &curs_loc);
                            v_cursor_get_model_pos(the_cursor, &curs_pos);
                            if (curs_loc.at_x < 0) {
                                /* reset cursor to right edge of character */
                                /*   if left edge is off screen */
                                on_phead = 
                                  format_mpos_2_phead(display_format, 
                                                      &curs_pos);
                                in_cline = 
                                  phead_mpos_2_cline(on_phead, &curs_pos);
                                in_char_set = 
                                  model_inq_state_at_node(
                                    node_get_model(curs_pos.in_node),
                                    curs_pos.in_node, tml_char_set_prop);
                                char_val = csm_c_char_set_as_ulong(
                                  in_char_set,
                                  ((unsigned char*) 
                                  node_get_contents(curs_pos.in_node))
                                  + curs_pos.char_index);
                                use_font = tline_get_font(in_cline);
                                curs_loc.at_x 
                                  += font_tbl_get_width(use_font, char_val)
                                  + font_tbl_get_icspace(use_font);
                                v_cursor_xy_set(the_cursor, &curs_loc);
                                }
                            }
                        break;

                    case tmlExit:
                        /* exit key: exit the program */
                        /* release dynamically allocated storage */
                        /* freeing viewport frees cursor */
                        view_port_free(display_viewport);
                        /* delete format */
                        format_free(display_format);
                        /* delete model */
                        model_free(display_model);
                        /* exit */
                        exit(0);

                    /*************************************************/
                    case tmlBackspace:
                        /* backspace key: delete character to left */
                        /*   of TMLib cursor */
                        /* try to move cursor left */
                        v_cursor_get_model_pos(the_cursor, &curs_pos);
                        v_cursor_move(the_cursor, vp_left, &curs_loc);
                        v_cursor_get_model_pos(the_cursor, &hold_pos);
                        /* quit on failure to move */
                        /*   (cursor at top of file) */
                        if (!model_pos_eq(&curs_pos, &hold_pos)) {
                            /* delete character */
                            view_port_delete_char(display_viewport,
                                 v_cursor_get_model_pos(the_cursor,
                                                        &curs_pos));
                            /* resetting cursor updates its model position */
                            v_cursor_xy_set(the_cursor, &curs_loc);
                            }
                        break;

                    case tmlReturn:
                        /* carriage return key: insert a return after cursor */
                        view_port_ins_after_cursor(display_viewport, "\n", 1);
                        break;

                    default:
                        /* any other key: insert character */
                        if (textlen > 0)
                            view_port_ins_after_cursor(display_viewport,
                                                       text, textlen);
                        break;
                    /*************************************************/
                }
                break;

            default:
               break;
            }
        }
}
