/*************************************************************************************
    csm_lock.hxx
    header file for the object definitions used for lock management in
        multiprocess applications - Dijkstra semaphore implementation

    (c) Copyright Apollo Computer Inc., 1989 (unpublished)
     --------------------------------------------------------------------
    |THE FOLLOWING PROGRAMS ARE THE SOLE PROPERTY OF APOLLO COMPUTER INC.|
    |      AND CONTAIN ITS PROPRIETARY AND CONFIDENTIAL INFORMATION.     |
     --------------------------------------------------------------------
**************************************************************************************/

#ifndef csm_lock_hxx
#define csm_lock_hxx

/**********************************************************************************
    classes defined in this file
**********************************************************************************/

class csm_proc_q;
class csm_lock;

/**********************************************************************************
    function to get pid of current process
**********************************************************************************/

extern void* csm_get_my_pid();

/**********************************************************************************
    process queue element representing one process waiting on a locked semaphore
**********************************************************************************/

class csm_proc_q {
    void*       pid;
    csm_proc_q* next;

    friend class csm_lock;
public:
    void*         get_pid(){return pid;}
    csm_proc_q*   get_next(){return next;}
    void          set_pid(void* new_pid){pid = new_pid;}
    void          set_next(csm_proc_q* new_next){next = new_next;}
    csm_proc_q(void* new_pid);
    ~csm_proc_q();
};

/**********************************************************************************
    semaphore for unique resources or vectors of resources
**********************************************************************************/

class csm_lock {
    void*       tml_sem;
    csm_proc_q* head;
    csm_proc_q* tail;
    void*       lockee;

public:
    void*       get_tml_sem(){return tml_sem;}
    csm_proc_q* get_head(){return head;}
    csm_proc_q* get_tail(){return tail;}
    void*       get_lockee(){return lockee;}
    void        set_tml_sem(void* new_tml_sem){tml_sem = new_tml_sem;}
    void        set_head(csm_proc_q* new_head){head = new_head;}
    void        set_tail(csm_proc_q* new_tail){tail = new_tail;}
    void        set_lockee(void* new_lockee){lockee = new_lockee;}
    csm_lock();                                     /* make lock for a unique resource */
    csm_lock(long pool_capacity);                   /* make lock for a resource pool */
    csm_lock(void* new_lockee);                     /* make lock for a specified resource pool */
    csm_lock(long pool_capacity, void* new_lockee); /* make lock for a specified resource pool */
    ~csm_lock();    /* destroy a lock, killing all processes DEAD that sleep on it */
    void attach(void* new_lockee){set_lockee(new_lockee);}
    void p_busy();                /* process busy waits on semaphore for resource */
    void p_suspend();             /* process queues self on semaphore if resource unavailable */
    long p_bounce();              /* process continues if resource unavailable */
    void v();                     /* process releases resource */
};

#endif
// csm_lock_hxx
