/**************************************************************************************
    font_traits.cxx
    basic traits for the font_tbl objects in the portable version
      of the text management library

    (c) Copyright Apollo Computer Inc., 1987 (unpublished)
     --------------------------------------------------------------------
    |THE FOLLOWING PROGRAMS ARE THE SOLE PROPERTY OF APOLLO COMPUTER INC.|
    |      AND CONTAIN ITS PROPRIETARY AND CONFIDENTIAL INFORMATION.     |
     --------------------------------------------------------------------
**************************************************************************************/

#include <tml/tml_objects.hxx>
#include <tml/tml_functions.hxx>
#include <tml/tml_dp_graph.hxx>
#include <tml/tml_os_depends.hxx>

/*constructor for width table for font*/
width_tbl::width_tbl(unsigned long default_char, unsigned long start, unsigned long max, short** width_tabs)
{
    start_index = start;
    max_index = max;
    dflt_char = default_char;
    if (width_tabs) {
        widths =    width_tabs[0];
        lbearings = width_tabs[1];
        rbearings = width_tabs[2];
        width_tabs[0] = 0;
        width_tabs[1] = 0;
        width_tabs[2] = 0;
        tml_os_free((char*) width_tabs);
        }
    else {
        widths = 0;
        lbearings = 0;
        rbearings = 0;
        }
}

width_tbl::~width_tbl()
{
    if (widths) {
        tml_os_free((char*) widths);
        widths = 0;
        }
    if (lbearings) {
        tml_os_free((char*) lbearings);
        lbearings = 0;
        }
    if (rbearings) {
        tml_os_free((char*) rbearings);
        rbearings = 0;
        }
}

/*constructor for a font table entry*/
font_tbl::font_tbl(vpftbl* func_tab)
{
    name = 0;
    name_count = 0;
    font_handle = 0;
    char_size = 1;     //default character size = 1 byte
    num_chars = 256;    //default number of characters = 256 (8 bit character set)
    first_char = 0;        //default first character index
    font_hi = 0;
    icspace = 0;
    baseline = 0;
    fixed = 0;
    is_loaded = 0;
    width = 0;
    ftbl = func_tab ? func_tab : tml_dp_get_dflt_vpftbl();
}

/*destructor for a font table entry*/
font_tbl::~font_tbl()
{
    /* free any width table associated with the font */
    delete width;
    /*if the font is loaded, unload it*/
    if (is_loaded)
        (*ftbl->ptr_2_tml_dp_funload)(font_handle);
    if (font_handle)
        (*ftbl->ptr_2_tml_dp_frelease)(font_handle);
    /*get rid of the name space allocated*/
    if (name)
        delete [name_count] name;
    ftbl = 0;
}
                                    
/*get all of the information about a font. return 0 if everything is ok, -1 otherwise (i.e., the font
    was not found)*/
long font_tbl::set_info(char *fname, long flen)
{
    font_handle = (*ftbl->ptr_2_tml_dp_get_fhandle)(fname, flen);
    if (!font_handle) return -1;

    name = new char[flen];
    tml_os_byte_copy(fname, name, flen);
    name_count = flen;

    char_size = (*ftbl->ptr_2_tml_dp_get_fcharsize)(font_handle);
    num_chars = (*ftbl->ptr_2_tml_dp_get_fcharnum)(font_handle);
    first_char = (*ftbl->ptr_2_tml_dp_get_fchar1)(font_handle);
    font_hi = (*ftbl->ptr_2_tml_dp_get_fhi)(font_handle);
    icspace = (*ftbl->ptr_2_tml_dp_get_ficspace)(font_handle);
    baseline = (*ftbl->ptr_2_tml_dp_get_fbline)(font_handle);
    fixed = (*ftbl->ptr_2_tml_dp_is_fixed)(font_handle);
    if (!fixed)
        width = new width_tbl((*ftbl->ptr_2_tml_dp_get_fspace)(font_handle),
                              (*ftbl->ptr_2_tml_dp_get_fchar1)(font_handle),
                              (*ftbl->ptr_2_tml_dp_get_fchar1)(font_handle)
                                + (*ftbl->ptr_2_tml_dp_get_fcharnum)(font_handle)
                                - 1,
                              (*ftbl->ptr_2_tml_dp_get_fwidths)(font_handle));
    return 0;
}

/*load a font -- return 0 if font loaded, -1 otherwise*/
long font_tbl::load_font()
{
    return (*ftbl->ptr_2_tml_dp_load_font)(font_handle);
}

/*return the current leading factor (in pixels)*/
long font_tbl::get_leading()
{
    return (font_hi - (*ftbl->ptr_2_tml_dp_get_fhi)(font_handle));
}

/*return the point size for a loaded font */
long font_tbl::get_point_size()
{
    return (*ftbl->ptr_2_tml_dp_get_fpoint )(font_handle);
}

/*return width table and set integer pointer argument's value */
/* to number of characters in table */
width_tbl* font_tbl::get_width_table(unsigned long* num_chars)
{
    *num_chars = (*ftbl->ptr_2_tml_dp_get_fcharnum)(font_handle);
    return width;
}

/*return vpftbl used by this font*/
vpftbl* font_tbl::get_ftbl()
{
    if (0 == ftbl)
        ftbl = tml_dp_get_dflt_vpftbl();
    return ftbl;
}

/*set the leading factor (given in pixels)*/
long font_tbl::set_leading(long leading_pixels)
{
    font_hi = (*ftbl->ptr_2_tml_dp_get_fhi)(font_handle) + leading_pixels;
    return font_hi;
}


