/**************************************************************************************
    format_traits.cxx
    member functions for the format class

    (c) Copyright Apollo Computer Inc., 1987 (unpublished)
     --------------------------------------------------------------------
    |THE FOLLOWING PROGRAMS ARE THE SOLE PROPERTY OF APOLLO COMPUTER INC.|
    |      AND CONTAIN ITS PROPRIETARY AND CONFIDENTIAL INFORMATION.     |
     --------------------------------------------------------------------
**************************************************************************************/

#include <tml/tml_objects.hxx>

/*constructor for a format*/
format::format()
{
    first_phead = new phead(this);
    first_phead->page_num = 1;
    first_phead->fline_no = 1;
    first_phead->line_cnt = 0;
    page_wide = page_hi = 0;
    max_line_width = max_used_ht = 0;
    last_phead = first_phead;
    this_model = 0;
    format_state = 0;
    ref_count = 1;
    retained_top = retained_bottom = 0;
    retain_pages = 5;
    top_free_rt_list = 0;
    attached_vps = 0;
}

/*destructor for a format*/
format::~format()
{
    register phead      *this_phead,
                        *next_phead;
    
    register rp_list   *cur_rp_ent,
                            *next_rp_ent;
             
    model_pos         zero_pos;
    /*if a model is attached, detach that model*/
    if (this_model)
        detach();

    /*detach this from any of the viewports with which it is currently associated*/
    /*note: view_port::detach_format() calls format::delete_viewport, which deletes the fv_list entry*/
    zero_pos.in_node = 0;
    zero_pos.char_index = -1;
    while (attached_vps){
        if (attached_vps->inq_vp()->get_cursor()){
            attached_vps->inq_vp()->get_cursor()->erase();
            attached_vps->inq_vp()->get_cursor()->set_model_pos(zero_pos);
            }
        attached_vps->inq_vp()->detach_format();
        }

    /*get rid of the property state*/
    if (format_state)
        delete format_state;
                             
    /*get rid of the retained pages*/
    for (cur_rp_ent = retained_top; cur_rp_ent; cur_rp_ent = next_rp_ent){
        next_rp_ent = cur_rp_ent->get_next();
        delete cur_rp_ent;
        }
    retained_top = retained_bottom = top_free_rt_list = 0;
    retain_pages = 0;

    /*finally, get rid of the page headers*/
    for(this_phead = first_phead; this_phead != 0; this_phead = next_phead){
        next_phead = this_phead->get_next();
        delete this_phead;
        }
}


/*attach a model to a format--note that this might
    get more complex when we introduce
    the notion of a model state*/
void format::attach(model *to_model)
{
    model_pos     new_first_pos;

    /*if a model is currently attached, detach it and wipe all but the first page head*/
    if (this_model)
        detach();

    this_model = to_model;
    new_first_pos.in_node = this_model->get_root();
    new_first_pos.char_index = 0;
    if (!new_first_pos.can_format())
        new_first_pos = new_first_pos.get_next();
    first_phead->set_start_pos(new_first_pos);
    new_first_pos.in_node = 0;
    new_first_pos.char_index = 0;
    first_phead->set_end_pos(new_first_pos);

    /*add the format to the mf_list of the model*/
    to_model->add_format(this);
}

/*detach a model from a format--note that
    this might get more complex when we introduce
    the notion of a model state*/
void format::detach()
{
    register phead    *this_phead,
                    *next_phead;
    model_pos null_pos;
    register rp_list   *current_rp,
                            *next_rp;
    register fv_list   *attach_vps;

    for (this_phead = first_phead->get_next(); this_phead; this_phead = next_phead){
        next_phead = this_phead->get_next();
        delete this_phead;
        }

    /*get rid of any traces of the old conceptual lines on any viewports*/
    null_pos.in_node = 0;
    null_pos.char_index = -1;
    for (attach_vps = attached_vps; attach_vps; attach_vps = attach_vps->inq_next()){
        if (attached_vps->inq_vp()->get_cursor()){
            attached_vps->inq_vp()->get_cursor()->erase();
            attached_vps->inq_vp()->get_cursor()->set_model_pos(null_pos);
            }
        attach_vps->inq_vp()->free_all_pos();
        }

    /*get rid of any retained pages*/
    for (current_rp = retained_top; current_rp; current_rp = next_rp){
        next_rp = current_rp->get_next();
        current_rp->set_page(0);
        current_rp->set_next(top_free_rt_list);
        top_free_rt_list = current_rp;
        }
    retained_top = retained_bottom = 0;

    first_phead->free_clines();
    first_phead->set_start_pos(null_pos);
    first_phead->set_end_pos(null_pos);
    first_phead->mark_reformat();
    last_phead = first_phead;
    
    /*take the mf_list element with this format off of the list for the model*/
    if (this_model)
        this_model->delete_format(this);
    this_model = 0;
}


/*set the size of the conceptual pages associated with the model*/
void format::set_page_size(short hi, short wide)
{
    page_hi = hi;
    page_wide = wide;
}

/*inquire on the property given in prop_ind.
 First check to see if it is set in the format;
 if not, return it's value in the model*/
void* format::inq_state(p_index prop_ind)
{
    void  *return_state;

    if (format_state && (return_state = format_state->inq_state(prop_ind)))
        return return_state;
    return ( this_model ? this_model->inq_state(prop_ind) : 0);
}

/*inquire on the property given in prop_ind and a node.
  First check to see if it is set in the format;
  if not, return it's value in the model at that node*/
void* format::inq_state(node *at_node, p_index prop_ind)
{
    void  *return_state;

    if (format_state && (return_state = format_state->inq_state(prop_ind)))
        return return_state;
    return ( this_model ? this_model->inq_state(at_node, prop_ind) : 0);
}
                                                   
/*add a viewport to the list of viewports to which this format is attached*/
void format::add_viewport(view_port *add_vp)
{
    register fv_list   *new_vp_list;

    new_vp_list = new fv_list;
    if (!new_vp_list) return;

    new_vp_list->set_next(attached_vps);
    new_vp_list->set_prior(0);
    if (attached_vps)
        attached_vps->set_prior(new_vp_list);
    new_vp_list->set_vp(add_vp);
    attached_vps = new_vp_list;
}

/*delete a viewport from the list of viewports to which this format is attached*/
void format::delete_viewport(view_port *del_vp)
{
    register fv_list   *hold_vp;
                  
    /*find the fv_list entry for the viewport being detached*/
    for (hold_vp = attached_vps; hold_vp; hold_vp = hold_vp->inq_next())
        if (hold_vp->inq_vp() == del_vp)
            break;
        ;
    /*if there is none, we're done*/
    if (!hold_vp) return;
                                               
    /*detach the found fv_list entry from the list of attached viewports*/
    if (hold_vp->inq_prior())
        hold_vp->inq_prior()->set_next(hold_vp->inq_next());
    else
        attached_vps = hold_vp->inq_next();

    if (hold_vp->inq_next())
        hold_vp->inq_next()->set_prior(hold_vp->inq_prior());

    /*delete the fv_list entry*/
    delete hold_vp;
}

/*set the format property state for the indicated property to the given value.
    Note that all format properties are default values*/
long format::set_state(p_index prop_ind, void* value)
{
    if (!format_state){
        format_state = new prop_state;
        if (!format_state)
            return -1;
        }
    return format_state->set_default(prop_ind, value);
}

/*unset the format property state for the indicated value. This is equivalent to setting
    the value for the property to 0*/
long format::unset_state(p_index prop_ind)
{
    if (!format_state)
        return 0;
    return format_state->set_default(prop_ind, 0);
}

/*make a retained page entry, and add it to the top of the list. Called by format_page*/
void format::make_rp_entry(phead *new_page)
{
    register rp_list   *new_rp_ent;
                  
    /*first, check to make sure that there isn't already an entry for the page--
        if there is, free the clines on the page, as it is about to be reformatted*/
    for (new_rp_ent = retained_top; new_rp_ent; new_rp_ent = new_rp_ent->get_next())
        if (new_rp_ent->get_page() == new_page){
            new_page->free_clines();
            return;
            }

    if (top_free_rt_list){
        new_rp_ent = top_free_rt_list;
        top_free_rt_list = new_rp_ent->get_next();
        }
    else
        new_rp_ent = new rp_list;

    new_rp_ent->set_prior(0);
    new_rp_ent->set_next(retained_top);
    if (retained_top)
        retained_top->set_prior(new_rp_ent);
    else
        retained_bottom = new_rp_ent;
    new_rp_ent->set_page(new_page);
    retained_top = new_rp_ent;
    purge_rp_list();
}

/*purge the retained page list--for any pages beyond the limit specified by the
    number of retained pages which are not visible, free the conceptual lines and
    delete from the retained page list*/
void format::purge_rp_list()
{
    register long i;
    register rp_list   *end_purge,
                        *purge_page;

    for (i = 0, end_purge = retained_top;
          end_purge && (i < retain_pages);
          end_purge = end_purge->get_next(), i++)
        ;

    if (!end_purge) return;

    for(purge_page = retained_bottom;
        purge_page != end_purge;
        purge_page = purge_page->get_prior()){
        if (!purge_page->get_page()->is_visible()){
            purge_page->get_page()->free_clines();
            purge_page->set_page(0);
            purge_page->get_prior()->set_next(purge_page->get_next());
            if (purge_page->get_next())
                purge_page->get_next()->set_prior(purge_page->get_prior());
            else
                retained_bottom = purge_page->get_prior();
            purge_page->set_next(top_free_rt_list);
            if (top_free_rt_list)
                top_free_rt_list->set_prior(purge_page);
            top_free_rt_list = purge_page;
            }
        }
}

          
void format::calc_max_line_width()
{
    register phead    *hold_page;
 
    max_line_width = 0;
    for (hold_page = first_phead; hold_page; hold_page = hold_page->get_next())
        if (max_line_width < hold_page->get_max_line_width())
            max_line_width = hold_page->get_max_line_width();
}


void format::calc_max_used_ht()
{
    register phead    *hold_page;
 
    max_used_ht = 0;
    for (hold_page = first_phead; hold_page; hold_page = hold_page->get_next())
            max_used_ht += hold_page->get_max_used_ht();
}

/*********************************************************************************
    navigation functions
*********************************************************************************/
/*return the page head of the page containing the supplied character*/
phead* format::mpos_2_phead(model_pos find_pos)
{
    register phead    *return_page,
                            *start_page,
                    *current_page;
    register rp_list   *retained_page;
    register long         i;
    model_pos     target,
                    holder;
                                 
    /*if find_pos isn't terminal, set the target to the next terminal*/
    if (!find_pos.in_node->can_format())
        target = find_pos.get_next();
    else
        target = find_pos;

    /*check the first page*/
    return_page = first_phead;
    if (!return_page->get_format_ok()){
        holder = return_page->get_start();
        if (holder.in_node == 0){
            holder.in_node = find_pos.in_node->get_model()->get_root();
            holder.char_index = 0;
            return_page->set_start_pos(holder);
            }
        return_page->format_page(return_page->get_start());
        }

    if (target <= return_page->get_end())
        return return_page;

    return_page = last_phead;
    /*easy (but ugly) case--character is on a page that hasn't been
        done yet*/
    if (return_page->get_end() < target){
        if (inq_state(target.in_node, tml_header_prop)){
            for (return_page = return_page->next_page(1);
                return_page;
                return_page = return_page->next_page(1))
                if (return_page->get_end() >= target)
                    break;
            }
        else
            return_page = last_phead->insert_page(target);
        return return_page;
        }
                                    
    /*check to see if it is the last page*/
    if (target >= return_page->get_start())
        return return_page;
                                         
    /*character is on a page which has been formatted--do a search to find it*/
    /*first, check to see if it is on one of the retained pages*/
    start_page = first_phead;
    for (retained_page = retained_top; retained_page; retained_page = retained_page->get_next()){
        current_page = retained_page->get_page();
        if (current_page->get_start() > target)
            continue;
        else if ((current_page->get_start() <= target) && (current_page->get_end() >= target))
            return current_page;
        else if (current_page->get_page_no() > start_page->get_page_no())
            start_page = current_page;
        }
    current_page = start_page;
    i = (current_page == first_phead) ? 1 : 0;
    if (inq_state(target.in_node, tml_header_prop)){
        for (return_page = first_phead->next_page(i);       //start with the first page, to insure any necessary reformatting
              return_page != last_phead;
              return_page = return_page->next_page(i)){
            if (return_page->get_end() >= target)
                return return_page;
            if (return_page == current_page) i = 1;
            }
        }
    else
        for (return_page = first_phead;
            return_page != last_phead;
            return_page = return_page->get_next()){
            if ((target >= return_page->get_start()) && (target <= return_page->get_end()))
                return return_page;
            else if ((target > return_page->get_end()) && (target < return_page->get_next()->get_start()))
                return (return_page->insert_page(target));
            }

    return 0;
}

/*return the page head of the indicated page if there is one, otherwise return 0*/
phead* format::get_page(long page_no)
{
    register phead *return_phead;

    if (page_no < 1) return first_phead;    //no page numbers less than 1; return first page

    if (page_no > last_phead->get_page_no()){
    /*the page hasn't been formatted yet--add pages until the page is found or we run out
        of text*/
        for (return_phead = last_phead;
            (return_phead);
            return_phead = return_phead->next_page(1))
            if (return_phead->get_page_no() >= page_no)
                break;
                ;
        return return_phead;
        }

    /*the page has been formatted--find it*/
    if (page_no > (last_phead->get_page_no() / 2))     //it is closer to the end than the beginning
        for (return_phead = last_phead;
             page_no < return_phead->get_page_no();
             return_phead = return_phead->prior_page(1))
             ;
    else
        for (return_phead = first_phead;
              page_no > return_phead->get_page_no();
              return_phead = return_phead->next_page(1))
              ;
    return return_phead;
}


