/**************************************************************************************
    model_traits.cxx
    functions for manipulating the model structure and the various heap lists
      associated with that structure for the portable version of the text
      management library

    (c) Copyright Apollo Computer Inc., 1987 (unpublished)
     --------------------------------------------------------------------
    |THE FOLLOWING PROGRAMS ARE THE SOLE PROPERTY OF APOLLO COMPUTER INC.|
    |      AND CONTAIN ITS PROPRIETARY AND CONFIDENTIAL INFORMATION.     |
     --------------------------------------------------------------------
**************************************************************************************/

#include <stdio.h>
#ifdef __cplusplus
#include <stdlib.h>
#endif
#include <tml_cxx2.0_linkage_decls.hxx>
#include <tml/tml_objects.hxx>
#include <tml/tml_functions.hxx>
#include <tml/tml_dp_graph.hxx>
#include <tml/tml_os_depends.hxx>
   
#define NULL 0

/*some default property values*/
const long    tml_default_wrap_val     = tml_false_prop;
const long    tml_default_margin[4]    = {0,0,0,0};
const long    tml_default_inv_val      = tml_false_prop;
const long    tml_default_psize_val[2] = {-1,-1};

/***************************************************************************

    Class functions for the model

**************************************************************************/
/*constructor for a model*/
model::model(long obj_heap_size , long goo_heap_size , long st_heap_size )
{
    extern long            tml_unix_search_exp(model_range, model_range&, void*, long);
    extern long            tml_unix_compile_exp(char*, long, void*, long, long);
    extern char*           tml_next_char(model_pos&);
    extern char*           tml_prior_char(model_pos&);
    extern char*           tml_this_char(model_pos&);

    char     *default_font;
    long      def_font_len;
    font_tbl *default_font_tbl;

    root_node = NULL;
    ref_count = 1;
    first_oheap = 0;
    first_gheap = 0;
    first_sheap = 0;
    oheap_size = obj_heap_size ? obj_heap_size : obj_heap_incr;
    current_oheap = new obj_heap(oheap_size);
    if (!current_oheap){
        /*this = 0;*/
        return;
        }
    gheap_size = goo_heap_size ? goo_heap_size : goo_heap_incr;
    current_gheap = new goo_heap(gheap_size);
    if (!current_gheap){
        delete current_oheap;
        /*this = 0;*/
        return;
        }
    sheap_size = st_heap_size ? st_heap_size : st_heap_incr;
    current_sheap = new st_heap(sheap_size);
    if (!current_sheap){
        delete current_gheap;
        delete current_oheap;
        /*this = 0;*/
        return;
        }
    model_state = new prop_state;
    if (!model_state){
        delete current_sheap;
        delete current_gheap;
        delete current_oheap;
        /*this = 0;*/
        return;
        }
    attach(current_oheap);
    attach(current_gheap);
    attach(current_sheap);
    attached_fonts = 0;
    past = new history;
    attached_formats = 0;
    attached_marks = 0;
    prop_pos = 0;
    char_sets = 0;
    model_lock = 0;
    heaps_lock = 0;

    /* set a default search environment */
    search_stack = new search_env(&tml_unix_search_exp,
                                  &tml_unix_compile_exp,
                                  &tml_next_char,
                                  &tml_prior_char,
                                  &tml_this_char);

    /* set the default union character set and the default character set */
    model_state->set_default(tml_union_char_set_prop, (void*) get_dflt_union_char_set());
    model_state->set_default(tml_char_set_prop, (void*) get_dflt_union_char_set()->get_sets());
    /*set the default font property*/
    default_font = (*tml_dp_get_dflt_vpftbl()->ptr_2_tml_dp_get_default_font)(def_font_len);
    if (!default_font){
        fprintf(stderr, "unable to find a default font--tmlib exiting\n");
        exit(1);
        }
    default_font_tbl = new font_tbl(tml_dp_get_dflt_vpftbl());
    if (default_font_tbl->set_info(default_font, def_font_len)){
        delete default_font_tbl;
        return;
        }
    attached_fonts = new ft_list;
    attached_fonts->set_next(0);
    attached_fonts->set_font(default_font_tbl);
    model_state->set_default(tml_font_prop, (void *)(default_font_tbl));
}
                                      
/*destructor for a model*/
model::~model()
{
    oheap_list *oh_current,  *oh_next;
    gheap_list *gh_current,  *gh_next;
    sheap_list *st_current,  *st_next;
    mf_list    *this_format, *next_format;
    ft_list    *this_font,   *next_font;
    mark_list  *this_mark,   *next_mark;
    search_env *current_se,  *next_se;

    if (--ref_count)
       return;

    /* go to sleep if model or its heaps are locked */
    if (model_lock) {
        lock_suspend();
        delete model_lock;
        // No need to v() the lock!
        // Those who sleep after me will sleep forever!
        }
    if (heaps_lock) {
        lock_heaps_suspend();
        delete heaps_lock;
        // No need to v() the lock!
        // Those who sleep after me will sleep forever!
        }

    /*update any format which contains this model*/
    for (this_format = attached_formats; this_format; this_format = next_format){
        next_format = this_format->inq_next();
        this_format->inq_format()->detach();
        }

    for(oh_current = first_oheap; oh_current != NULL; oh_current = oh_next){
    /*get rid of all of the node heaps associated with the model*/
        oh_next = oh_current->get_next();
        delete oh_current;
        }

    for(gh_current = first_gheap; gh_current != NULL; gh_current = gh_next){
    /*get rid of all of the goo heaps associated with the model*/
        gh_next = gh_current->get_next();
        delete gh_current;
        }
             
    /*get rid of all of the string heaps associated with the model*/
    for (st_current = first_sheap; st_current != NULL; st_current = st_next){
        st_next = st_current->get_next();
        delete st_current;
        }

    /*get rid of any font tables associated with the model*/
    for (this_font = attached_fonts; this_font; this_font = next_font){
        next_font = this_font->inq_next();
        delete this_font->inq_font();
        delete this_font;
        }
                                                    
    /*get rid of any marks associated with the model*/
    for (this_mark = attached_marks; this_mark; this_mark = next_mark){
        next_mark = this_mark->get_next();
        delete this_mark;
        }

    /*clear the search stack*/
    for (current_se = search_stack; current_se != NULL; current_se = next_se){
        next_se = current_se->get_next();
        delete current_se;
        }

    root_node = NULL;
    current_oheap = NULL;
    current_gheap = NULL;
    current_sheap = NULL;
    first_oheap = NULL;
    first_gheap = NULL;
    first_sheap = NULL;
    model_plid_table = NULL;
    delete model_state;
    model_state = NULL;
    prop_pos = NULL;
    delete past;
    past = NULL;
    attached_formats = NULL;
    attached_fonts = NULL;
    attached_marks = NULL;
    search_stack = NULL;
    delete char_sets;
    char_sets = NULL;
    model_lock = NULL;
    heaps_lock = NULL;
}
                                                                                                        
/**********************************************************************************

        Functions to directly manipulate the model data

**********************************************************************************/
/* this function added since node::set_node() is overloaded */
/*   because it is virtual and therefore a pointer to it */
/*   cannot be derived */
void fp_sn(node* subject, node* to_node, node_dir dir)
{
    subject->set_node(to_node, dir);
}

/*function to create a new root. If a root is already set, the old root becomes the child
    of the new root. Returns the new root*/
node* model::make_root()
{
    bnode    *new_root;

    history  *past;
    atom*     this_atom;
    atom_arg *arg1, *arg2, *arg3, *arg4;
                                 
    new_root = new bnode(this);
    if (!new_root) return 0;

    new_root->set_type(bobj);

    past = get_past();
    if (past->has_hist() && past->is_active()) {

        // if history is active, record this call

        arg1 = new atom_this_arg(past, new_root);
        arg2 = new atom_val_arg(past, new_root->get_node(n_down));
        arg3 = new atom_val_arg(past, root_node);
        arg4 = new atom_val_arg(past, (long)n_down);
        this_atom = new atom(past, (WHOLE_PTR) &fp_sn,
                                   (WHOLE_PTR) &fp_sn,
                                   arg1, arg2, arg4, NULL,
                                   arg1, arg3, arg4, NULL);
        past->add_atom(this_atom);
        if (root_node) {
            arg1 = new atom_this_arg(past, root_node);
            arg2 = new atom_val_arg(past, root_node->get_node(n_up));
            arg3 = new atom_val_arg(past, new_root);
            arg4 = new atom_val_arg(past,(long) n_up);
            this_atom = new atom(past, (WHOLE_PTR) &fp_sn,
                                       (WHOLE_PTR) &fp_sn,
                                       arg1, arg2, arg4, NULL,
                                       arg1, arg3, arg4, NULL);
            past->add_atom(this_atom);
            }

        }

    new_root->set_node(root_node, n_down);
    if (root_node)
        root_node->set_node(new_root, n_up);
    set_root(new_root);
    return new_root;
}

/*function to declare a node the root of the model. The node must already be
    associated with that model (i.e, be allocated in that model's object space).
    returns a pointer to the new root.*/
node* model::set_root(node* new_root)
{
    history  *past;
    atom     *this_atom;
    atom_arg *arg1, *arg2, *arg3;

    if (new_root) {
        if (new_root->get_model() != this)
            return 0;
        }

    past = get_past();
    if (past->has_hist() && past->is_active()) {

        // if history is active, record this call

        arg1 = new atom_this_arg(past, this);
        arg2 = new atom_val_arg(past, root_node);
        arg3 = new atom_val_arg(past, new_root);
        this_atom = new atom(past, (WHOLE_PTR) &model::set_root, (WHOLE_PTR) &model::set_root,
                             arg1, arg2, NULL, arg1, arg3, NULL);
        past->add_atom(this_atom);
        }
    root_node = new_root;
    return new_root;
}

/*attach an object heap to the list of object heaps associated with the model*/
void model::attach(obj_heap *new_heap)
{
    register oheap_list* new_list;

    new_list = new oheap_list;
    new_list->set_heap(new_heap);
    new_list->set_next(first_oheap);
    new_heap->attach(this);
    first_oheap = new_list;
}

/*attach a goo heap to the list of goo heaps associated with the model*/
void model::attach(goo_heap * new_heap)
{
    register gheap_list* new_list;

    new_list = new gheap_list;
    new_list->set_heap(new_heap);
    new_list->set_next(first_gheap);
    new_heap->attach(this);
    first_gheap = new_list;
}

/*attach a string heap to the list of string heaps associated with the model*/
void model::attach(st_heap * new_heap)
{
    register sheap_list* new_list;

    new_list = new sheap_list;
    new_list->set_heap(new_heap);
    new_list->set_next(first_sheap);
    new_heap->attach(this);
    first_sheap = new_list;
}

/*detach an object heap from the list of object heaps associated with a model*/
void model::detach(obj_heap *old_heap)
{
    register oheap_list* find_heap;

    find_heap = first_oheap;
    if ((find_heap->get_heap()) == old_heap){
        first_oheap = find_heap->get_next();
        if (old_heap == current_oheap)
            current_oheap = first_oheap?first_oheap->get_heap():0;
        old_heap->detach();
        return;
        }

    for(; find_heap->get_next() != NULL; find_heap = find_heap->get_next())
        if ((find_heap->get_next())->get_heap() == old_heap) break;
    if (find_heap->get_next() == NULL) return;

    find_heap->set_next((find_heap->get_next())->get_next());
    if (old_heap == current_oheap)
        current_oheap = first_oheap?first_oheap->get_heap():0;
    old_heap->detach();

}

/*detach a goo heap from the list of goo heaps associated with a model*/
void model::detach(goo_heap* old_heap)
{
    register gheap_list* find_heap;

    find_heap = first_gheap;
    if ((find_heap->get_heap()) == old_heap){
        first_gheap = find_heap->get_next();
        if (current_gheap == old_heap)
            current_gheap = first_gheap?first_gheap->get_heap():0;
        old_heap->detach();
        return;
        }

    for(; find_heap->get_next() != NULL; find_heap = find_heap->get_next())
        if ((find_heap->get_next())->get_heap() == old_heap) break;
    if (find_heap->get_next() == NULL) return;

    find_heap->set_next((find_heap->get_next())->get_next());
    if (current_gheap == old_heap)
        current_gheap = first_gheap?first_gheap->get_heap():0;
    old_heap->detach();

}

/*detach a string heap from the list of string heaps associated with a model*/
void model::detach(st_heap* old_heap)
{
    register sheap_list* find_heap;

    find_heap = first_sheap;
    if ((find_heap->get_heap()) == old_heap){
        first_sheap = find_heap->get_next();
        if (current_sheap == old_heap)
            current_sheap = first_sheap?first_sheap->get_heap():0;
        old_heap->detach();
        return;
        }
  
    for(; find_heap->get_next() != NULL; find_heap = find_heap->get_next())
        if ((find_heap->get_next())->get_heap() == old_heap) break;
    if (find_heap->get_next() == NULL) return;

    find_heap->set_next((find_heap->get_next())->get_next());
    if (current_sheap == old_heap)
        current_sheap = first_sheap?first_sheap->get_heap():0;
    old_heap->detach();

}
    
/*attach a union_char_set vector to the model's vector of union_char_set*/
void model::attach(union_char_set* new_union_set)
{
    register union_char_set* cur_union;
    register char_set*       cur_set;

    /* first union_char_set */
    if (NULL == char_sets) {
        char_sets = new_union_set;
        for (cur_union = char_sets; cur_union; cur_union = cur_union->get_next()) {
            cur_union->incr_ref_count();
            for (cur_set = cur_union->get_sets(); cur_set; cur_set = cur_set->get_next())
                cur_set->incr_ref_count();
            }
        return;
        }
    /* attach the new character set vector */
    char_sets->append(new_union_set);
}

/*detach single union_char_set element from the model's vector of union_char_sets */
void model::detach(union_char_set* old_union_set)
{
    register union_char_set* cur_union_set;

    if ((NULL == char_sets) || (NULL == old_union_set))
        return;
    /* find old_union_set in the model's vector */
    for (cur_union_set = char_sets; cur_union_set->get_next();
      cur_union_set = cur_union_set->get_next())
        if (cur_union_set->get_next() == old_union_set)
            break;
    /* not found */
    if (NULL == cur_union_set->get_next())
        return;
    /* found */
    else
        cur_union_set->detach(old_union_set);
}

/*add an mf_list element for a format--called by format::attach(model)*/
void model::add_format(format *new_format)
{
    register mf_list   *new_ele;

    new_ele = new mf_list;
    if (!new_ele) return;

    new_ele->set_next(attached_formats);
    if (attached_formats)
        attached_formats->set_prior(new_ele);
    new_ele->set_format(new_format);
    new_ele->set_prior(0);
    attached_formats = new_ele;
}

/*detach an mf_list element for a format--called by format::detach(model*)*/
void model::delete_format(format *del_format)
{
    register mf_list   *target_mf;

    for (target_mf = attached_formats;
          target_mf;
          target_mf = target_mf->inq_next())
          if (target_mf->inq_format() == del_format)
            break;
        ;

    if (target_mf->inq_prior())
        target_mf->inq_prior()->set_next(target_mf->inq_next());
    else
        attached_formats = target_mf->inq_next();
    if (target_mf->inq_next())
        target_mf->inq_next()->set_prior(target_mf->inq_prior());
    delete target_mf;
}

/**********************************************************************************

        Functions manipulating the property state of a model

**********************************************************************************/

/*inquire on the property value of the given property at the given node*/
void* model::inq_state(node *at_node, p_index prop_ind)
{
    register void  *return_val;

/*    model_state->set_lock();*/
    if (at_node == prop_pos)
        return_val = model_state->inq_state(prop_ind);
    else {
        prop_pos = move_prop(at_node);
        if (prop_pos != at_node)
            return_val = 0;
        else
            return_val = model_state->inq_state(prop_ind);
        }
/*    model_state->release_lock();*/
    return return_val;
}

/*move the current property position to the specified node*/
node* model::move_prop(node *to_node)
{
    register long      current_level,
                       to_level;
    register node     *hold_node;
                                         
    /*if to_node is the current property position, return it*/
    if (prop_pos == to_node) return prop_pos;
                                                                                  
    /*if to_node is nil, pop the props to above the root*/
    if (!to_node){

        for (hold_node = prop_pos; hold_node; hold_node = hold_node->get_node(n_up))
            hold_node->exit_props();
        prop_pos = 0;
        return 0;
        }

    /*if to_node is in a different model, return 0*/
    if (this != to_node->get_model())
        return 0;

    /*if there is no current property position, just set the properties to all those active at to_node*/
    if (!prop_pos){
        prop_pos = move_prop(to_node->get_node(n_up));
        to_node->enter_props();
        prop_pos = to_node;
        return prop_pos;
        }

    /*if to_node is higher in the tree then the current property position,
      unset properties until prop_pos is at the same level as to_node*/
    to_level = to_node->get_level();
    for(current_level = prop_pos->get_level();
         current_level > to_level;
         prop_pos = prop_pos->get_node(n_up), current_level--)
         prop_pos->exit_props();
                                                                                                                                          
    /*if to_node is lower in the tree than the current property position, set the properties to its parent,
        then set the properties of to_node*/
    if(prop_pos->get_level() < to_level){
        prop_pos = move_prop(to_node->get_node(n_up));
        to_node->enter_props();
        prop_pos = to_node;
        return prop_pos;
        }
      
    /*prop_pos and to_node are of the same level--unset the current property position's properties,
        then set them to the parent of to_node, then set to_node's properties (recurse like mad)*/

    prop_pos->exit_props();
    prop_pos = prop_pos->get_node(n_up);
    prop_pos = move_prop(to_node->get_node(n_up));
    to_node->enter_props();
    prop_pos = to_node;
    return prop_pos;
}

/**********************************************************************************

        Functions which allocate space in the model's heaps

**********************************************************************************/


/*create a property vector within a model*/
prop_vec* model::make_pvec()
{
    register prop_vec   *new_pvec;

    new_pvec = new prop_vec(this);
    return new_pvec;
}

/*create a property within a model*/
property* model::make_prop(long property_size)
{
    register property    *new_prop;

    new_prop = new property(this, property_size);
    return new_prop;
}

/*create an argument slot with an integer value*/
argument* model::make_arg(long new_value)
{
    register argument    *new_arg;

    new_arg = new int_arg(this);
    if (!new_arg) return 0;
                                
    ((int_arg*)new_arg)->set_val(new_value);
    return new_arg;
}

/*create an argument slot with a float value*/
argument* model::make_arg(float new_value)
{
    register argument    *new_arg;

    new_arg = new float_arg(this);
    if (!new_arg) return 0;
                              
    ((float_arg*)new_arg)->set_val(new_value);
    return new_arg;
}

/*create an argument slot with a string value*/
argument* model::make_arg(char *new_value, long new_count)
{
    register argument    *new_arg;

    new_arg = new string_arg(this);
    if (!new_arg) return 0;
                                
    ((string_arg*)new_arg)->set_val(new_value, new_count);
    return new_arg;
}

/*create an argument slot for a subtree*/
argument* model::make_arg(node *subtree)
{
    register argument    *new_arg;

    new_arg = new subtree_arg(this);
    if (!new_arg) return 0;
                                
    ((subtree_arg*)new_arg)->set_val(subtree);
    return new_arg;
}
                                                                          
/*create an argument slot for a char_set*/
argument* model::make_arg(char_set* new_char_set)
{
    register argument    *new_arg;

    new_arg = new char_set_arg(this);
    if (!new_arg) return 0;
                                
    ((char_set_arg*)new_arg)->set_val(new_char_set);
    return new_arg;
}
                                                                          
/*create an argument slot for a union_char_set*/
argument* model::make_arg(union_char_set *new_union_char_set)
{
    register argument    *new_arg;

    new_arg = new union_char_set_arg(this);
    if (!new_arg) return 0;
                                
    ((union_char_set_arg*)new_arg)->set_val(new_union_char_set);
    return new_arg;
}
                                                                          
/*create a node of the indicated size*/
node* model::make_node(long node_size)
{
    register node    *new_node;

    new_node = new node(node_size, this);
    return new_node;
}

/*allocate a chunk of the goo heap of the indicated size*/
char* model::make_goo(long count, char *contents)
{
    register char  *new_goo;

    new_goo = current_gheap->string_alloc(count, contents);
    return new_goo;
}
                                                                          
/*allocate an (unknown) object in the object heap*/
char* model::make_obj(long obj_size)
{
    register char* new_obj;

    new_obj = current_oheap->alloc(obj_size);
    return new_obj;
}

/***************************************************************************
              cut and paste functions
***************************************************************************/

/*cut a node and it's subtree out of the model tree, and update any displays*/
/*returns the node which was cut out*/
node* model::cut(node *cut_node)
{                
    model_range      cut_range;
    model_pos         prior_char,
                        next_char,
                        start_redraw;
    register format           *attached_format;
    register view_port      *attached_viewport;
    register mf_list   *in_format;
    register fv_list   *in_viewport;

    /*first, do a garbage check*/
    if (!this || !cut_node)
        return 0;

    /*get the endpoints of the cut, and find the next and prior characters*/
    cut_range = cut_node->get_span();
    prior_char = cut_range.get_start().get_prior();
    next_char = cut_range.get_end().get_next();
    if (prior_char.in_node)
        start_redraw = prior_char;
    else if (next_char.in_node)
        start_redraw = next_char;
    else {
        start_redraw.in_node = new gnode(this);
        if (!start_redraw.in_node)
            return 0;
        start_redraw.in_node->attach_data("\n", 1);
        start_redraw.char_index = 0;
        cut_node->connect_node(start_redraw.in_node, n_right);
        }
                
    /*now cut the node*/
    cut_node->detach();

    /*now update any displays which contained the node*/
    for (in_format = attached_formats; in_format; in_format = in_format->inq_next()){
        attached_format = in_format->inq_format();
        for (in_viewport = attached_format->get_fv_list(); in_viewport; in_viewport = in_viewport->inq_next()){
            attached_viewport = in_viewport->inq_vp();
            /*if we have cut beginning of the first line, do an incremental redraw beginning with start_redraw*/
            if ((attached_viewport->get_top_line()) && 
                ((cut_range.contains(attached_viewport->get_top_line()->get_start())) ||
                (attached_viewport->get_top_line()->get_start().in_node->get_max_index() < 
                    attached_viewport->get_top_line()->get_start().char_index)))
                attached_viewport->redraw(start_redraw, start_redraw, attached_viewport->get_top_line());
            else
                attached_viewport->redraw();
            }
        }

    return cut_node;
}

/*paste a node in the position indicated, and update any displays.*/
/*returns the node which has been pasted in*/
node* model::paste(node* paste_node, node *paste_to, node_dir paste_dir)
{
    register node  *return_node;

    /*first, check for garbage*/
    if (!this || !paste_node || !paste_to || (paste_to->get_model() != this))
        return 0;

    return_node = paste_to->connect_node(paste_node, paste_dir);
    update_display();
    return return_node;
}
 
/***************************************************************************
    functions which make marks, delete marks, and display marks

**************************************************************************/
                                                              
/*make a mark out of the supplied range. Note that the model_range
    supplied is copied into storage which is used by the mark*/
mark_list* model::make_mark(model_range init_range)
{
    mark_list     *new_mark;
    model_range   *new_mark_range;

    /*first, check for incoming garbage*/
    if (!this || !init_range.get_start().in_node || !init_range.get_end().in_node)
        return 0;

    /*make sure that the mark is part of this model*/
    if ((init_range.get_start().in_node->get_model() != this) ||
        (init_range.get_end().in_node->get_model() != this))
        return 0;
                                                  
    /*now make a new mark_list entry, and fill it in with the new mark range*/
    new_mark = new mark_list;
    new_mark_range = new_mark->set_range(init_range);
    if (!new_mark_range){
        delete new_mark;
        return 0;
        }
    
    /*now add the mark_list item to the list kept for the model*/
    new_mark->set_next(attached_marks);
    if (attached_marks)
        attached_marks->set_prior(new_mark);
    attached_marks = new_mark;
    return new_mark;
}

                     
/*delete a mark from the list of marks associated with the model. Make sure that it is
    erased from any viewport in which it is currently visible*/
void model::delete_mark(mark_list *to_delete)
{
    register mark_list    *find_mark;

    /*first, check for garbage*/
    if (!this || !to_delete)
        return;

    /*find the mark in the list of marks for the model*/
    for (find_mark = attached_marks; find_mark; find_mark = find_mark->get_next())
        if (find_mark == to_delete)
            break;

    if (!find_mark)        //it wasn't on the list for the model
        return;

    /*detach the item from the model's list of marks*/
    if (find_mark->get_prior())
        find_mark->get_prior()->set_next(find_mark->get_next());
    else
        attached_marks = find_mark->get_next();

    if (find_mark->get_next())
        find_mark->get_next()->set_prior(find_mark->get_prior());

    /*erase the mark in any viewport in which it is visible*/
    find_mark->erase();

    /*range in mark is deleted by destructor*/
    delete find_mark;
}

/*draw the mark in all viewports*/
long model::mark_draw(mark_list *to_draw)
{
    /*first, check for garbage*/
    if (!this || !to_draw)
        return 0;

    /*draw it*/
    return (to_draw->draw());
}

/*erase the mark in all viewports*/
long model::mark_erase(mark_list *to_erase)
{
    /*first, check for garbage*/
    if (!this || !to_erase)
        return 0;

    /*erase it*/
    return (to_erase->erase());
}

/*update the display of the model, reflecting any changes in the model*/
/*NOTE--this implementation is crude--it redraws any part of the model,
    even if that part has not been changed*/
void model::update_display()
{
    register mf_list   *in_format;
    register fv_list   *in_viewport;

    for (in_format = attached_formats; in_format; in_format = in_format->inq_next()){
        for (in_viewport = in_format->inq_format()->get_fv_list(); in_viewport; in_viewport = in_viewport->inq_next()){
            in_viewport->inq_vp()->redraw();
            }
        }
}
     

/*search functions*/
/*search a range for a pre-compiled pattern*/
model_range model::search(char *search_exp, long search_exp_len,
                          model_range search_range, long forward, long case_sens)
{
    model_range    return_range;
    const long     max_exp_len  = 4096;
    register char *pattern_ptr;

    /*check for trash*/
    if (!search_range.get_start().in_node ||
        !search_range.get_end().in_node ||
        !search_exp)
        return return_range;

    if (!search_stack)
        return return_range;

    if (!search_stack->get_search() || !search_stack->get_compile())
        return return_range;

    pattern_ptr = tml_os_alloc(max_exp_len, 1);
    (*(search_stack->get_compile()))(search_exp, search_exp_len, (void*)pattern_ptr, max_exp_len, case_sens);
    (*(search_stack->get_search()))(search_range, return_range, pattern_ptr, forward);
    tml_os_free(pattern_ptr);

    return return_range;
}

/*compile a search expression*/
long model::compile_exp(char *search_exp, long search_exp_len,
                                void *compiled_exp, long max_comp_len, long case_sens)
{
    /*check for trash*/
    if (!search_stack || !search_stack->get_compile())
        return 0;
    if (!search_exp || !compiled_exp)
        return 0;

    return ((*(search_stack->get_compile()))(search_exp, search_exp_len, compiled_exp, max_comp_len, case_sens));
}
                    
/*search the indicated range, after compiling the regular expression handed in*/
model_range model::search_exp(void *pattern, model_range search_range,  long forward )
{
    model_range     return_range;

    if (!search_range.get_start().in_node ||
        !search_range.get_end().in_node ||
        !pattern)
        return return_range;

    if (!search_stack)
        return return_range;

    if (!search_stack->get_search())
        return return_range;
                                                  
    (*(search_stack->get_search()))(search_range, return_range, pattern, forward);
    return return_range;
}

/***************************************************************************
    Functions to implement mutual exclusion
**************************************************************************/

long model::lock_bounce()
{
    if (model_lock)
        return model_lock->p_bounce();
    return 0;
}

long model::lock_heaps_bounce()
{
    if (heaps_lock)
        return heaps_lock->p_bounce();
    return 0;
}

/***************************************************************************
    Class functions for heap_list (base class for lists of heaps associated
    with a particular model) -- should never be called directly

**************************************************************************/

/*constructor for the heap list element*/
heap_list::heap_list()
{
    next = NULL;
    this_heap = NULL;
}

/*destructor for the heap list element*/
heap_list::~heap_list()
{
    if (this_heap != NULL) delete this_heap;
    next = NULL;
    this_heap = NULL;
}

/***************************************************************************
    Destructors for heap_list derived classes

***************************************************************************/

/*destructor for oheap_list*/
oheap_list::~oheap_list()
{
    delete ((obj_heap*) this_heap);
    this_heap = NULL;
}

/*destructor for gheap_list*/
gheap_list::~gheap_list()
{
    delete ((goo_heap*) this_heap);
    this_heap = NULL;
}

/*destructor for sheap_list*/
sheap_list::~sheap_list()
{
    delete ((st_heap*) this_heap);
    this_heap = NULL;
}

/*destructor for sqheap_list*/
sqheap_list::~sqheap_list()
{
    delete ((squeeze_heap*) this_heap);
    this_heap = NULL;
}

