/**************************************************************************************
    prop_ent_traits.cxx
    member functions for the class prop_func_ent and the location of all
      of the entry and exit functions for TML

    (c) Copyright Apollo Computer Inc., 1987 (unpublished)
     --------------------------------------------------------------------
    |THE FOLLOWING PROGRAMS ARE THE SOLE PROPERTY OF APOLLO COMPUTER INC.|
    |      AND CONTAIN ITS PROPRIETARY AND CONFIDENTIAL INFORMATION.     |
     --------------------------------------------------------------------
**************************************************************************************/

#include <tml/tml_objects.hxx>
#include <tml/tml_functions.hxx>
#include <tml/tml_dp_graph.hxx>

/**************************************************************************************
    Global tml_prop_func_tbl
**************************************************************************************/

// To make this global per-process, must be declared with attribute specifier
//   for Apollo builds to allow use in global and inlib'd libraries.
#ifdef apollo
prop_func_ent *tml_prop_func_tbl[prop_tbl_size] __attribute((section(tml_prop_func_tbl))) = {0};
#else
prop_func_ent *tml_prop_func_tbl[prop_tbl_size] = {0};
#endif

/**************************************************************************************
    Service functions for tml_prop_func_tbl
**************************************************************************************/

prop_func_ent *get_prop_func_tbl_entry(long tbl_index)
{
    return(tml_prop_func_tbl[tbl_index]);
}


void set_prop_func_tbl_entry(long tbl_index, prop_func_ent *entry)
{
    
    tml_prop_func_tbl[tbl_index] = entry;
}


/**************************************************************************************
    Property entry and exit functions
**************************************************************************************/

/*activation function for word wrap property*/
void* wrap_entry(argument *wrap_on)
{
    long *i;

    i = new long;
    *i = *((long*)wrap_on->get_val());
    return ((void*) i);
}
 
void* wrap_exit(argument *wrap_on)
{
    long *i;
     
    if (!wrap_on)
        return 0;
    i = (long *)wrap_on->get_model()->get_state()->inq_state(tml_wrap_prop);
    if (i) delete i;
    return 0;
}

/*activation function for margin property*/
void* margin_entry(argument *margin_set)
{
    register argument *margin_arg;
    register long i;
    long *margin_array;

    margin_array = new long[4];
    
    for (i=0, margin_arg = margin_set;
          margin_arg && i < 4;
          i++, margin_arg = margin_arg->get_next())
        *(margin_array + i) = *((long*)margin_arg->get_val());

    for ( ; i < 4; i++)
        *(margin_array + i) = 0;

    return (void *)(margin_array);
}

/*deactivation function for margin property*/
void* margin_exit(argument *margin_unset)
{
    long *margin_array;
    
    if (!margin_unset)
        return 0;

    margin_array = (long *)(margin_unset->get_model()->get_state()->inq_state(tml_margin_prop));
    if (margin_array)
        delete[4] margin_array;
    return 0;
}

/*activation function for inverse property*/
void* inv_entry(argument *inverse_on)
{
    return (*((long*)inverse_on->get_val())) ? (void *)(&tml_true_prop)
                                          : (void *)(&tml_false_prop);
}

/*activation function for the font property*/
void *font_entry(argument *font_arg)
{
              char *font_name,
                   *old_font_name;
    register  long i;
    long      f_len;
    font_tbl *new_font;
    ft_list  *model_fonts;
    vpftbl   *ftbl;

    font_name = (char*)font_arg->get_val();
    f_len = ((string_arg*)font_arg)->get_scount();
    if (font_arg->get_next()
      && (0 == (ftbl = (vpftbl*) font_arg->get_next()->get_val())))
            ftbl = tml_dp_get_dflt_vpftbl();
    else
        ftbl = tml_dp_get_dflt_vpftbl();

    for (model_fonts = font_arg->get_model()->get_fonts();
            model_fonts;
            model_fonts = model_fonts->inq_next()){
        new_font = model_fonts->inq_font();
        if (new_font->get_name_count() != f_len)
            continue;
        old_font_name = new_font->get_name();
        for (i = 0; (i < f_len) && (*(font_name + i) == *(old_font_name + i)); i++)
            ;
        if (f_len <= i)
            break;
        }
    if (model_fonts)
        return (void *)(new_font);

    new_font = new font_tbl(ftbl);
    if (new_font->set_info(font_name, f_len)){
        delete new_font;
        return 0;
        }
    model_fonts = new ft_list;
    model_fonts->set_next(font_arg->get_model()->get_fonts());
    model_fonts->set_font(new_font);
    font_arg->get_model()->set_font_tbl(model_fonts);
    return (void*)(new_font);
}
                                                      
/*activation function for page size property*/
void *psize_entry(argument *psizes)
{
    long *size_array;
    register long i;
    register argument *size_args;

    size_array = new long[2];

    for (i = 0, size_args = psizes;
            size_args && i < 2;
            i++, size_args = size_args->get_next())
        *(size_array + i) = *((long*)size_args->get_val());
                        
    for ( ; i < 2; i++)
        *(size_array + i) = -1;

    return (void *)size_array;
}

/*deactivation function for page size property*/
void* psize_exit(argument *psizes)
{
    long *size_array;

    if (!psizes) return 0;

    size_array = (long *)(psizes->get_model()->get_state()->inq_state(tml_pgsize_prop));
    if (size_array)
        delete[2] size_array;
    return 0;
}

/*activation function for header property*/
void* header_entry(argument *header_set)
{
    register argument *header_arg;
    register long i;
    node **header_array;
           
    header_array = (node**)new node*[3];
                 
    // Left,Center, Right header

    for (i=0, header_arg = header_set;
          header_arg && i < 3;
          i++, header_arg = header_arg->get_next())
          header_array[i] = (node*)header_arg->get_val();

    return (void *)(header_array);
}

/*deactivation function for header property*/
void* header_exit(argument *header_unset)
{
    node **hdr_ptr;
    long  i;
    
    if (!header_unset)
        return 0;

    hdr_ptr = (node**)(header_unset->get_model()->get_state()->inq_state(tml_header_prop));

    if (hdr_ptr)
    {
       for (i=0 ; i < 3; i++)
           delete *(hdr_ptr + i);

       delete hdr_ptr;
    }

    return 0;
}

/*activation function for footer property*/
void* footer_entry(argument *footer_set)
{
    register argument *footer_arg;
    register long i;
    node **footer_array;
           
    footer_array = (node**)new node*[3];
                            
    // Left,Center, Right footer

    for (i=0, footer_arg = footer_set;
          footer_arg && i < 3;
          i++, footer_arg = footer_arg->get_next())
          footer_array[i] = (node*)footer_arg->get_val();

    return (void *)(footer_array);
}

/*deactivation function for footer property*/
void* footer_exit(argument *footer_unset)
{
    node **ftr_ptr;
    long  i;
    
    if (!footer_unset)
        return 0;

    ftr_ptr = (node**)(footer_unset->get_model()->get_state()->inq_state(tml_footer_prop));

    if (ftr_ptr)
    {
       for (i=0 ; i < 3; i++)
           delete *(ftr_ptr + i);

       delete ftr_ptr;
    }

    return 0;
}


/*activation function for form feed property*/
void* form_feed_entry(argument *form_feed_on)
{
    long   *i = new long;

    *i = *((long*)form_feed_on->get_val());
    return ((void*) i);
}
 
void* form_feed_exit(argument *form_feed_on)
{
    long   *i;
     
    if (!form_feed_on)
        return 0;
    i = (long *)form_feed_on->get_model()->get_state()->inq_state(tml_form_feed_prop);
    if (i) delete i;
    return 0;
}

/*activation function for character set property*/
void* char_set_entry(argument *cur_char_set)
{
    return(cur_char_set->get_val());
}

/*activation function for union character set property*/
void* union_char_set_entry(argument *cur_union)
{
    return(cur_union->get_val());
}

/**************************************************************************************
    Member functions for the class prop_func_ent
**************************************************************************************/

/*constructor for a property function table entry*/
prop_func_ent::prop_func_ent()
{
    next = 0;
    prop_index = undefined_prop;
    entry = exit = 0;
    loaded = 0;
}

/*destructor for a property function table entry*/
prop_func_ent::~prop_func_ent()
{
    register prop_func_ent *hold_ent,
                           *next_ent,
                           *last_ent;
    register long            tbl_index;

    /*first, find if it is still in the table, and if so remove it*/
    tbl_index = ((((unsigned long) prop_index) >> PROP_HASH_SHIFT) & PROP_HASH_MASK);
    last_ent = tml_prop_func_tbl[tbl_index];
    while (last_ent == this){
        tml_prop_func_tbl[tbl_index] = last_ent->next;
        last_ent = last_ent->next;
        }
    for (last_ent = tml_prop_func_tbl[tbl_index]; last_ent; last_ent = next_ent){
        hold_ent = last_ent->next;
        if (hold_ent != this)
            next_ent = hold_ent;
        else{
            last_ent->next = next;
            next_ent = last_ent;
            }
        }
    entry = exit = 0;
    loaded = 0;
}


/*activation function for tab property*/
void* tab_entry(argument *tab_set)
{
    return ((void*) tab_set);
}

/*load the function associated with a property into the property function table*/
void prop_func_ent::load()
{
    /*note--this will get added to as the properties get defined*/
    if (prop_index == tml_wrap_prop) {
        entry = &wrap_entry;
        exit = &wrap_exit;
        }
    else if (prop_index == tml_margin_prop) {
        entry = &margin_entry;
        exit = &margin_exit;
        }
    else if (prop_index == tml_font_prop)
        entry = &font_entry;
    else if (prop_index == tml_inv_prop)
        entry = &inv_entry;
    else if (prop_index == tml_pgsize_prop) {
        entry = &psize_entry;
        exit = &psize_exit;
        }
    else if (prop_index == tml_header_prop) {
        entry = &header_entry;
        exit = &header_exit;
        }
    else if (prop_index == tml_footer_prop) {
        entry = &footer_entry;
        exit = &footer_exit;
        }
    else if (prop_index == tml_form_feed_prop) {
        entry = &form_feed_entry;
        exit = &form_feed_exit;
        }
    else if (prop_index == tml_char_set_prop)
        entry = &char_set_entry;
    else if (prop_index == tml_union_char_set_prop)
        entry = &union_char_set_entry;
    else if (prop_index == tml_tab_prop) {
        entry = &tab_entry;
        }
   loaded = 1;
}


