/**************************************************************************************
     range_traits.cxx
     functions for the model range class in the text management library
       modelling component

    (c) Copyright Apollo Computer Inc., 1987 (unpublished)
     --------------------------------------------------------------------
    |THE FOLLOWING PROGRAMS ARE THE SOLE PROPERTY OF APOLLO COMPUTER INC.|
    |      AND CONTAIN ITS PROPRIETARY AND CONFIDENTIAL INFORMATION.     |
     --------------------------------------------------------------------
**************************************************************************************/

#include <stdio.h>
#include <ctype.h>
#include <tml/tml_objects.hxx>
#include <tml/tml_functions.hxx>
#include <tml/tml_os_depends.hxx>

/*constructor for the class model_range*/
model_range::model_range()
{
    start.in_node = end.in_node = 0;
    start.char_index = end.char_index = -1;
}


/*isolate a model range so that it is dominated by a single node,
    return that node*/
node* model_range::isolate()
{
    model_pos     range_start,
                        range_end;
    register node  *hold_node;

    /*check to make sure that we haven't been handed trash*/
    if ((start.in_node == 0) || (end.in_node == 0))
        return 0;
         
    /*make sure that range_start is before range_end*/
    if (start > end){
        range_start = end;
        range_end = start;
        }
    else {
        range_start = start;
        range_end = end;
        }
        
    /*if we need to split the range_start node, do so and update the range_start*/
    if (range_start.char_index > 0){
        hold_node = range_start.in_node->split(range_start.char_index);
        if (range_start.in_node == range_end.in_node){
            range_end.in_node = hold_node;
            range_end.char_index -= range_start.char_index;
            }
        range_start.in_node = hold_node;
        range_start.char_index = 0;
        }

    /*if we need to split the end node, do so and update the end*/
    if (range_end.char_index < range_end.in_node->get_max_index()){
        range_end.in_node->split(range_end.char_index + 1);
        range_end.char_index = range_end.in_node->get_max_index();
        }

    /*now, isolate the subtree*/
    return (make_subtree(range_start.in_node, range_end.in_node));
}

                                                                           
/*see if the indicated target position is within the bounds of the model_range.
    return 1 if it is, 0 otherwise*/
long model_range::contains(model_pos target_pos)
{
    model_range target_range;
    model_pos     range_start,
                        range_end;

    /*first, check to see if we have been handed trash*/
    if ((start.in_node == 0) || (end.in_node == 0) || (target_pos.in_node == 0))
        return 0;

    /*make sure that range_start is before range_end*/
    if (start > end){
        range_start = end;
        range_end = start;
        }
    else {
        range_start = start;
        range_end = end;
        }

    if (target_pos.in_node->get_last() == target_pos.in_node)
        return((range_start <= target_pos) && (range_end >= target_pos));

    target_range = target_pos.in_node->get_span();
    return ((range_start <= target_range.start) && (range_end >= target_range.end));
}
