/**************************************************************************************
    tml_c_internals.cxx
    definitions for functions declared in tml_c_types.h for the C language
      interface of the portable version of the text management library
      used in other parts of tml.

    (c) Copyright Apollo Computer Inc., 1987 (unpublished)
     --------------------------------------------------------------------
    |THE FOLLOWING PROGRAMS ARE THE SOLE PROPERTY OF APOLLO COMPUTER INC.|
    |      AND CONTAIN ITS PROPRIETARY AND CONFIDENTIAL INFORMATION.     |
     --------------------------------------------------------------------
**************************************************************************************/

#include <stdio.h>
#include <tml/tml_cxx2.0_linkage_decls.hxx>
#include <tml/tml_objects.hxx>
#include <tml/tml_functions.hxx>
#include <tml/tml_os_depends.hxx>

#ifdef __cplusplus
extern "C" {
#endif

// perform initial draw as indicated - all graphics systems
void tml_dp_init_draw(void* viewport_handle)
{
    view_port   *vp;
    phead       *first_page;

    vp = (view_port*) viewport_handle;
    if (vp->get_format()) {
        if (first_page = vp->get_format()->get_first_page()) {
            if (vp->get_format()->get_model() && vp->get_format()->get_model()->get_root()) {
                first_page->format_page(vp->get_format()->get_model()->get_root()->get_span().get_start());
                vp->draw(first_page->get_tline(), zero_zero);
                }
            }
        }
    return;
}

// XWindows only
// use viewport offset to enlarge rectangle described by four long arguments
// then repaint any conceptual line that impinges on the rectangle
void tml_c_repaint_screen (view_port *vp, long ulx, long uly, long lrx, long lry)
{
    register cline       *next_line,
                         *last_line;
    register long         delta_x, delta_y;
    register long         crit_y;
    register screen_xy    start_xy;
             draw_mark   *cur_mark;
             model_range *cur_range;
             draw_m_draw  cur_draw_func;
             core_m_draw  cur_core_func;
             model_pos    start_mpos, end_mpos, correct_start, ending_mpos;
             view_pos    *start_vpos, *end_vpos;
             v_cursor    *cursor;

    // perform initial draw if required

    // view_port::xy_2_cline() returns zero if nothing on screen
    next_line = vp->xy_2_cline(zero_zero);
    if (!next_line) {
        tml_dp_init_draw(vp);
        return;
        }

    // nominally clears entire viewport, but clip is on

    (*(vp->get_ftbl()->ptr_2_tml_dp_clear))(vp->get_handle(),
                                            0, 0,
                                            vp->get_width(), vp->get_hi());

    // get viewport shift (from scroll) if any

    delta_x = start_xy.at_x = vp->get_top_line()->get_x();
    delta_y = start_xy.at_y = vp->get_top_line()->get_y();
     
    // enlarge rectangle to account for any scrolling
    //   but keep it within bounds of viewport

    if (delta_x < 0) {
        ulx += delta_x;
        if (ulx < 0)
            ulx = 0;
        }
    else if (delta_x > 0) {
        lrx += delta_x;
        if (lrx > vp->get_width())
            lrx = vp->get_width();
        }
    if (delta_y < 0) {
        uly += delta_y;
        if (uly < 0)
            uly = 0;
        }
    else if (delta_y > 0) {
        lry += delta_y;
        if (lry > vp->get_hi())
            lry = vp->get_hi();
        }

    // test for lines impinging on box and draw those that do

    for (; next_line; next_line = next_line->get_down()) {

        // skip line that cannot be drawn
        if ( ! next_line->can_draw())
            continue;

        // get upper left screen coordinates of start character on line
        start_xy = vp->mpos_2_xy(next_line->get_start());

        // line not on screen
        if ((start_xy.at_x == 0) && (start_xy.at_y == -1))
            break;

        // adjust y coordinate to upper left corner of line's bounding box
        if (cl_text == next_line->get_type())
            start_xy.at_y -= ((tline*)next_line)->get_baseline();

        // get crit_y, value that tells whether line may impinge
        //   on vertical bounds of box (allows handling of case
        //   where top of line is higher than box but bottom is not)

        crit_y = start_xy.at_y;
        if (crit_y < uly)
            crit_y += next_line->get_hi();
            
        // line's top or top-plus-line-height falls below top of box

        if (crit_y >= uly) {

            // at least part of line falls above bottom of box

            if (start_xy.at_y <= lry) {

                // test if line impinges on horizontal bounds of box 

                // line starts left of box

                if (start_xy.at_x < ulx) {
                    if (ulx <= (start_xy.at_x + next_line->get_contents_width()))
                        next_line->draw_line(vp, start_xy);
                    }

                // line starts on left edge of box

                else if (start_xy.at_x == ulx)
                    next_line->draw_line(vp, start_xy);

                // line starts before or on right edge of box

                else if (ulx < start_xy.at_x) {
                    if (start_xy.at_x <= lrx)
                        next_line->draw_line(vp, start_xy);
                    }
                }

            // this and all following lines are below box
            else
                break;
            }
        }

    // test for marked lines impinging on the box and remark those that do
    // logic is same as for loop above except that start and end of marks
    //   may not be on tline boundaries
    for (cur_mark = vp->get_marks(); cur_mark; cur_mark = cur_mark->get_next()) {

        // get range covered by current mark
        cur_range = cur_mark->get_model_mark()->get_range();

        // check for correct order of start and end
        start_mpos = cur_range->get_start();
        end_mpos = cur_range->get_end();
        if (end_mpos < start_mpos) {
            correct_start = end_mpos;
            end_mpos = start_mpos;
            start_mpos = correct_start;
            }

        // get model position start character for repaint
        // use top-of-viewport for start if mark starts above screen
        // skip to next mark if mark starts below screen
        start_vpos = vp->mpos_2_vpos(start_mpos);
        if (!start_vpos) {
            start_vpos = vp->get_top_line();
            if (start_mpos <= start_vpos->get_cline()->get_start())
                start_mpos = start_vpos->get_cline()->get_start();
            else
                continue;
            }

        // get model position of end character for repaint
        // use bottom-of-viewport for end if mark ends below screen
        // skip to next mark if mark ends above screen
        end_vpos = vp->mpos_2_vpos(end_mpos);
        if (!end_vpos) {
            end_vpos = vp->get_bottom_line();
            if (end_mpos >= end_vpos->get_cline()->get_end())
                end_mpos = end_vpos->get_cline()->get_end();
            else
                continue;
            }

        // get cline associated with start character's screen position
        next_line = vp->xy_2_cline(vp->mpos_2_xy(start_mpos));

        // get cline associated with end character's screen position
        last_line = vp->xy_2_cline(vp->mpos_2_xy(end_mpos));

        // re-mark lines that impinge on box
        // only those portions that fall in clipped regions
        //   will be affected (i.e. inverted, if mark has 
        //   default draw routine)
        cur_draw_func = cur_mark->get_draw_func();
        cur_core_func = cur_mark->get_core_func();

        for (; next_line && (next_line != last_line->get_next());
          next_line = next_line->get_next(),
          start_mpos = next_line->get_start()) {
            ending_mpos = (next_line == last_line) ? end_mpos : next_line->get_end();
            start_xy = vp->mpos_2_xy(start_mpos);
            if ((start_xy.at_x == 0) && (start_xy.at_y == -1))
                continue;
            if (cl_text == next_line->get_type())
                start_xy.at_y -= ((tline*) next_line)->get_baseline();
            crit_y = start_xy.at_y;
            if (crit_y < uly)
                crit_y += next_line->get_hi();
            if (crit_y >= uly) {
                if (start_xy.at_y <= lry) {
                    if (start_xy.at_x < ulx) {
                        if (ulx <= (vp->mpos_2_xy(next_line->get_start()).at_x + next_line->get_wide()))
                            (*cur_draw_func)(vp, cur_core_func, start_mpos, ending_mpos);
                        }
                    else if (start_xy.at_x == ulx)
                        (*cur_draw_func)(vp, cur_core_func, start_mpos, ending_mpos);
                    else if (ulx < start_xy.at_x) {
                        if (start_xy.at_x <= lrx)
                            (*cur_draw_func)(vp, cur_core_func, start_mpos, ending_mpos);
                        }
                    }
                else
                    break;
                }
            }
        }

    if (cursor = vp->get_cursor()) {
        // Cursor exists. Sets visibility of cursor to 
        //   zero to make draw effective.
        // Draw affects only obscured portions of cursor
        //   because clip is set.
        cursor->set_visible(0);
        cursor->draw();
        }
}

char* tml_c_new_wrapper(unsigned long num_bytes)
{
    return new char[num_bytes];
}

void tml_c_delete_wrapper(char* ptr)
{
    delete ptr;
}

long v_cursor_get_visible(void *v_cursor_p)
{
    return(((v_cursor *) v_cursor_p)->get_visible());
}

long v_cursor_draw(void *v_cursor_p)
{
    return (((v_cursor *) v_cursor_p)->draw());
}

long v_cursor_erase(void *v_cursor_p)
{
    return (((v_cursor *) v_cursor_p)->erase());
}

void *view_port_inq_state(void *view_port_p, p_index prop_ind)
{
    return((void *) ((view_port*) view_port_p)->inq_state(prop_ind));
}

void *view_port_get_cursor(void *view_port_p)
{
    return((void *) ((view_port*) view_port_p)->get_cursor());
}

void view_port_redraw(void *view_port_p)
{
    ((view_port *)view_port_p)->redraw();
}

void *view_port_xy_2_cline(void *view_port_p, C_SCREEN_XY *where)
{
    return((void *) ((view_port*) view_port_p)->xy_2_cline(*((screen_xy *) where)));
}

char *font_tbl_get_name(void *font_tbl_p)
{
    return(((font_tbl *) font_tbl_p)->get_name());
}

long font_tbl_get_name_count(void *font_tbl_p)
{
    return(((font_tbl *) font_tbl_p)->get_name_count());
}

C_SCREEN_XY *c_screen_xy_alloc()
{
    return((C_SCREEN_XY *) tml_os_alloc(1, (long) sizeof(C_SCREEN_XY)));
}

void c_screen_xy_free(C_SCREEN_XY *this_c_screen_xy)
{
    tml_os_free((char*) this_c_screen_xy);
}

void *model_inq_state_at_node(void *model_p, void *node_p, void* i)
{
    return((void *) ((model *) model_p)->inq_state((node *) node_p, (p_index) i));
}

C_MODEL_POS *c_model_pos_alloc()
{
    return((C_MODEL_POS *) tml_os_alloc(1, (long) sizeof(C_MODEL_POS)));
}

void c_model_pos_free(C_MODEL_POS *this_c_model_pos)
{
    tml_os_free((char*) this_c_model_pos);
}

void *node_get_contents(void *node_p)
{
    return((void *) ((node *) node_p)->get_contents());
}

void *node_get_model(void *node_p)
{
    return((void *) ((node *) node_p)->get_model());
}

C_MODEL_POS* view_port_xy_2_mpos(void *view_port_p, C_SCREEN_XY *where, C_MODEL_POS *model_pos_p)
{
    model_pos pos;
    char *new_model_pos;

    if ((long) sizeof(model_pos) != (long) sizeof(C_MODEL_POS)) {
        printf("view_port_xy_2_mpos: sizes of class model_pos and type C_MODEL_POS are incompatible\n");
        return(NULL);
        }
    else {
        pos = ((view_port *) view_port_p)->xy_2_mpos(*((screen_xy *) where));
        new_model_pos = tml_os_byte_copy((char *) &pos, (char *) model_pos_p, (long) sizeof(model_pos));
        return((C_MODEL_POS *) new_model_pos);
        }
}

void *view_port_get_format(void *view_port_p)
{
    return((void *) ((view_port*) view_port_p)->get_format());
}

void *format_get_first_page(void *format_p)
{
    return((void *) ((format *) format_p)->get_first_page());
}

C_MODEL_POS *cline_get_start(void *cline_p, C_MODEL_POS *model_pos_p)
{
    char *new_model_pos;
    model_pos pos;

    if ((long) sizeof(model_pos) != (long) sizeof(C_MODEL_POS)) {
        printf("cline_get_start: sizes of class model_pos and type C_MODEL_POS are incompatible\n");
        return(NULL);
        }
    else {
        pos = ((cline *) cline_p)->get_start();
        new_model_pos = tml_os_byte_copy((char *) &pos, (char *) model_pos_p, (long) sizeof(model_pos));
        return((C_MODEL_POS *) new_model_pos);
        }
}

C_MODEL_POS *phead_get_start(void *phead_p, C_MODEL_POS *pos_p)
{
    char *new_mod_pos;
    model_pos pos;

    if ((long) sizeof(model_pos) != (long) sizeof(C_MODEL_POS)) {
        printf("phead_get_start: sizes of class model_pos and type C_MODEL_POS are incompatible\n");
        return(NULL);
        }
    else {
        pos = ((phead *) phead_p)->get_start();
        new_mod_pos = tml_os_byte_copy((char *) &pos, (char *) pos_p, (long) sizeof(model_pos));
        return((C_MODEL_POS *) new_mod_pos);
        }
}

void *view_port_get_top_line(void *view_port_p)
{
    return((void *) ((view_port*) view_port_p)->get_top_line());
}

void *view_pos_get_cline(void *view_pos_p)
{
    return((void *) ((view_pos *) view_pos_p)->get_cline());
}

#ifdef __cplusplus
}
#endif

