/**************************************************************************************
    tml_d_ptr.hxx
    header file for objects used to convert pointers to and from
        disk format

    (c) Copyright Apollo Computer Inc., 1989 (unpublished)
     --------------------------------------------------------------------
    |THE FOLLOWING PROGRAMS ARE THE SOLE PROPERTY OF APOLLO COMPUTER INC.|
    |      AND CONTAIN ITS PROPRIETARY AND CONFIDENTIAL INFORMATION.     |
     --------------------------------------------------------------------
**************************************************************************************/

#ifndef tml_d_ptr_hxx
#define tml_d_ptr_hxx

#ifndef arch_neutral_hxx
#include <tml/arch_neutral.hxx>
#endif

#ifndef endian_hxx
#include <tml/endian.hxx>
#endif

/*heap/offset incarnation of a pointer, used to store address-independent pointers on disk*/
class d_ptr {
    HALF_PTR heap;                 // an index into an array of heaps in class d_model_parts (disk save)
    HALF_PTR offset;               // an offset into the heap indexed by the previous member

public:
    HALF_PTR  get_heap(){return heap;}
    HALF_PTR  get_offset(){return offset;}
    long      is_null(){return ((((HALF_PTR)-1)==heap) && (((HALF_PTR)-1)==offset));}
    void      get_d_ptr(char *ptr, CHARP *table);  // obtain the disk-format pointer associated with a memory pointer
    void*     get_m_ptr(CHARP *table);            // obtain the memory pointer associated with a d_ptr
    WHOLE_PTR as_whole_ptr();                 // obtain and architecture-neutral version of a d_ptr
    void      operator=(void *source){heap = ((HALF_PTR)(((WHOLE_PTR) source) >> BITS_IN_HALF_PTR));offset = ((HALF_PTR)(HALF_PTR_MASK & ((WHOLE_PTR) source)));}
    void      unscramble(endian* their_endian){heap = their_endian->my_half_ptr(heap); offset = their_endian->my_half_ptr(offset);}
    };

#endif 
// tml_d_ptr_hxx

