/**************************************************************************************
    File tml_inlib_entry.cxx
    Initializer entry point for inlib libraries

    (c) Copyright Apollo Computer Inc., 1987 (unpublished)
     --------------------------------------------------------------------
    |THE FOLLOWING PROGRAMS ARE THE SOLE PROPERTY OF APOLLO COMPUTER INC.|
    |      AND CONTAIN ITS PROPRIETARY AND CONFIDENTIAL INFORMATION.     |
     --------------------------------------------------------------------
**************************************************************************************/

#include <stdio.h>
#include <tml/tml_cxx2.0_linkage_decls.hxx>
#include <tml/tml_objects.hxx>
#include <tml/tml_os_depends.hxx>
#include <tml/dflt_view_funcs.hxx>

/* Initialize global (non-per-process) objects for inlib'd libraries */
/*   (Apollo only). NOTE: The division of tml_ext_type initialization into */
/*   init_tml_ext_types_part_1() and init_tml_ext_types_part_2() is purely */
/*   arbitrary. It is a workaround for a cfxx input buffer size limitation */
/*   that causes an internal error during compilation if the code in the two */
/*   functions is included in a single function. */
/* To cause tml_inlib_entry() to be called when the library is loaded by inlib */
/*   build the library with the "-entry tml_inlib_entry" option to /com/bind. */
/*   (The -entry option must appear AFTER the object module that contains the */
/*   routine it designates as the alternate entry point.) */
/* This routine is not called when a library built in that manner is bound */
/*    into a program. */
/* The work of this module is parallel to the work of extend.cxx in bound-in */
/*   libraries; variables declared extern here are defined there. */

/* Declare pointers to preexisting global objects */
extern vpftbl       dflt_view_port_function_table;
#ifdef apollo
extern GEN_FUNCP    gpr_dflt_view_port_function_table[];
#endif

/* Declare pointers to new global objects */

vpftbl              *tml_sample_xwindows_grafsys_p;
#ifdef apollo
vpftbl              *tml_sample_gpr_grafsys_p;
#endif

/* objects used by CSM initialized by call to csm_inlib_entry() */

union_char_set_arg  *tml_sample_union_char_set_arg_p;
char_set_arg        *tml_sample_char_set_arg_p;
subtree_arg         *tml_sample_subtree_arg_p;
string_arg          *tml_sample_string_arg_p;
float_arg           *tml_sample_float_arg_p;
int_arg             *tml_sample_int_arg_p;
argument            *tml_sample_argument_p;
font_property       *tml_sample_font_prop_p;
property            *tml_sample_prop_p;
pgnode              *tml_sample_pgnode_p;
gnode               *tml_sample_gnode_p;
bnode               *tml_sample_bnode_p;

extern tml_ext_type tml_xwindows_grafsys_ext_t;
#ifdef apollo
extern tml_ext_type tml_gpr_grafsys_ext_t;
#endif

/* objects used by CSM initialized by call to csm_inlib_entry() */

extern tml_ext_type tml_union_char_set_val_ext_t;
extern tml_ext_type tml_char_set_val_ext_t;
extern tml_ext_type tml_subtree_val_ext_t;
extern tml_ext_type tml_string_val_ext_t;
extern tml_ext_type tml_float_val_ext_t;
extern tml_ext_type tml_int_val_ext_t;
extern tml_ext_type tml_no_val_ext_t;
extern tml_ext_type tml_tab_prop_ext_t;
extern tml_ext_type tml_union_char_set_prop_ext_t;
extern tml_ext_type tml_char_set_prop_ext_t;
extern tml_ext_type tml_form_feed_prop_ext_t;
extern tml_ext_type tml_footer_prop_ext_t;
extern tml_ext_type tml_header_prop_ext_t;
extern tml_ext_type tml_pgsize_prop_ext_t;
extern tml_ext_type tml_inv_prop_ext_t;
extern tml_ext_type tml_font_prop_ext_t;
extern tml_ext_type tml_margin_prop_ext_t;
extern tml_ext_type tml_wrap_prop_ext_t;
extern tml_ext_type tml_undefined_prop_ext_t;
extern tml_ext_type tml_free_ext_t;
extern tml_ext_type tml_leaf_ext_t;
extern tml_ext_type tml_pgobj_ext_t;
extern tml_ext_type tml_gobj_ext_t;
extern tml_ext_type tml_branch_ext_t;
extern tml_ext_type tml_bobj_ext_t;
extern tml_ext_type tml_undefined_node_ext_t;

extern tml_type tml_xwindows_grafsys_t;
#ifdef apollo
extern tml_type tml_gpr_grafsys_t;
#endif
extern tml_type tml_union_char_set_val_t;
extern tml_type tml_char_set_val_t;
extern tml_type tml_subtree_val_t;
extern tml_type tml_string_val_t;
extern tml_type tml_float_val_t;
extern tml_type tml_int_val_t;
extern tml_type tml_no_val_t;
extern tml_type tml_tab_prop_t;
extern tml_type tml_union_char_set_prop_t;
extern tml_type tml_char_set_prop_t;
extern tml_type tml_form_feed_prop_t;
extern tml_type tml_footer_prop_t;
extern tml_type tml_header_prop_t;
extern tml_type tml_pgsize_prop_t;
extern tml_type tml_inv_prop_t;
extern tml_type tml_font_prop_t;
extern tml_type tml_margin_prop_t;
extern tml_type tml_wrap_prop_t;
extern tml_type tml_undefined_prop_t;
extern tml_type tml_free_t;
extern tml_type tml_leaf_t;
extern tml_type tml_pgobj_t;
extern tml_type tml_gobj_t;
extern tml_type tml_branch_t;
extern tml_type tml_bobj_t;
extern tml_type tml_undefined_node_t;

extern tml_type* const tml_xwindows_grafsys;
#ifdef apollo
extern tml_type* const tml_gpr_grafsys;
#endif
extern tml_type* const tml_syll_frag_char_set;
extern tml_type* const tml_alpha_char_set;
extern tml_type* const tml_char_set;
extern tml_type* const union_char_set_val_arg;
extern tml_type* const char_set_val_arg;
extern tml_type* const subtree_val_arg;
extern tml_type* const string_val_arg;
extern tml_type* const float_val_arg;
extern tml_type* const int_val_arg;
extern tml_type* const no_val_arg;
extern tml_type* const tml_tab_prop;
extern tml_type* const tml_union_char_set_prop;
extern tml_type* const tml_char_set_prop;
extern tml_type* const tml_form_feed_prop;
extern tml_type* const tml_footer_prop;
extern tml_type* const tml_header_prop;
extern tml_type* const tml_pgsize_prop;
extern tml_type* const tml_inv_prop;
extern tml_type* const tml_font_prop;
extern tml_type* const tml_margin_prop;
extern tml_type* const tml_wrap_prop;
extern tml_type* const undefined_prop;
extern tml_type* const is_free;
extern tml_type* const leaf;
extern tml_type* const pgobj;
extern tml_type* const gobj;
extern tml_type* const branch;
extern tml_type* const bobj;
extern tml_type* const undefined_node;

extern type_queue type_q;

extern tml_ext_type_v1 tml_pgsize_prop_ext_t_v1;
extern tml_ext_type_v1 tml_inv_prop_ext_t_v1;
extern tml_ext_type_v1 tml_font_prop_ext_t_v1;
extern tml_ext_type_v1 tml_margin_prop_ext_t_v1;
extern tml_ext_type_v1 tml_wrap_prop_ext_t_v1;
extern tml_ext_type_v1 gobj_ext_t_v1;
extern tml_ext_type_v1 bobj_ext_t_v1;

extern type_trans_v1 tml_pgsize_trans;
extern type_trans_v1 tml_inv_trans;
extern type_trans_v1 tml_font_trans;
extern type_trans_v1 tml_margin_trans;
extern type_trans_v1 tml_wrap_trans;
extern type_trans_v1 tml_gobj_trans;
extern type_trans_v1 tml_bobj_trans;

extern void csm_inlib_entry();

extern type_trans_queue_v1 type_trans_q_v1;

void tml_inlib_entry()
{
    void init_tml_ext_types_part_1();
    void init_tml_ext_types_part_2();
    void init_tml_types();
    void init_tml_type_pointers();
    void init_v1_data();

    csm_inlib_entry();

    dflt_view_port_function_table.ptr_2_tml_dp_get_default_font      = &tml_dp_x_get_default_font;
    dflt_view_port_function_table.ptr_2_tml_dp_get_fhandle           = &tml_dp_x_get_fhandle;
    dflt_view_port_function_table.ptr_2_tml_dp_load_font             = &tml_dp_x_load_font;
    dflt_view_port_function_table.ptr_2_tml_dp_get_fcharsize         = &tml_dp_x_get_fcharsize;
    dflt_view_port_function_table.ptr_2_tml_dp_get_fcharnum          = &tml_dp_x_get_fcharnum;
    dflt_view_port_function_table.ptr_2_tml_dp_get_fchar1            = &tml_dp_x_get_fchar1;
    dflt_view_port_function_table.ptr_2_tml_dp_get_fhi               = &tml_dp_x_get_fhi;
    dflt_view_port_function_table.ptr_2_tml_dp_get_fpoint            = &tml_dp_x_get_fpoint;
    dflt_view_port_function_table.ptr_2_tml_dp_get_fbline            = &tml_dp_x_get_fbline;
    dflt_view_port_function_table.ptr_2_tml_dp_is_fixed              = &tml_dp_x_is_fixed;
    dflt_view_port_function_table.ptr_2_tml_dp_get_fwidths           = &tml_dp_x_get_fwidths;
    dflt_view_port_function_table.ptr_2_tml_dp_get_ficspace          = &tml_dp_x_get_ficspace;
    dflt_view_port_function_table.ptr_2_tml_dp_set_ficspace          = &tml_dp_x_set_ficspace;
    dflt_view_port_function_table.ptr_2_tml_dp_get_fspace            = &tml_dp_x_get_fspace;
    dflt_view_port_function_table.ptr_2_tml_dp_set_fspace            = &tml_dp_x_set_fspace;
    dflt_view_port_function_table.ptr_2_tml_dp_funload               = &tml_dp_x_funload;
    dflt_view_port_function_table.ptr_2_tml_dp_frelease              = &tml_dp_x_frelease;
    dflt_view_port_function_table.ptr_2_tml_dp_make_graphic_handle   = &tml_dp_x_make_graphic_handle;
    dflt_view_port_function_table.ptr_2_tml_dp_delete_graphic_handle = &tml_dp_x_delete_graphic_handle;
    dflt_view_port_function_table.ptr_2_tml_dp_get_hi                = &tml_dp_x_get_hi;
    dflt_view_port_function_table.ptr_2_tml_dp_get_wide              = &tml_dp_x_get_wide;
    dflt_view_port_function_table.ptr_2_tml_dp_get_foreg_id          = &tml_dp_x_get_foreg_id;
    dflt_view_port_function_table.ptr_2_tml_dp_get_backg_id          = &tml_dp_x_get_backg_id;
    dflt_view_port_function_table.ptr_2_tml_dp_get_fill_backg_id     = &tml_dp_x_get_fill_backg_id;
    dflt_view_port_function_table.ptr_2_tml_dp_get_fill_obj_id       = &tml_dp_x_get_fill_obj_id;
    dflt_view_port_function_table.ptr_2_tml_dp_set_foreg_id          = &tml_dp_x_set_foreg_id;
    dflt_view_port_function_table.ptr_2_tml_dp_set_backg_id          = &tml_dp_x_set_backg_id;
    dflt_view_port_function_table.ptr_2_tml_dp_set_fill_backg_id     = &tml_dp_x_set_fill_backg_id;
    dflt_view_port_function_table.ptr_2_tml_dp_set_fill_obj_id       = &tml_dp_x_set_fill_obj_id;
    dflt_view_port_function_table.ptr_2_tml_dp_set_clip              = &tml_dp_x_set_clip;
    dflt_view_port_function_table.ptr_2_tml_dp_clear_clip            = &tml_dp_x_clear_clip;
    dflt_view_port_function_table.ptr_2_tml_dp_clear                 = &tml_dp_x_clear;
    dflt_view_port_function_table.ptr_2_tml_dp_copy                  = &tml_dp_x_copy;
    dflt_view_port_function_table.ptr_2_tml_dp_blt                   = &tml_dp_x_blt;
    dflt_view_port_function_table.ptr_2_tml_dp_string_draw           = &tml_dp_x_string_draw;
    dflt_view_port_function_table.ptr_2_tml_dp_string_size           = &tml_dp_x_string_size;
    dflt_view_port_function_table.ptr_2_tml_dp_damage_repair         = &tml_dp_x_damage_repair;
    dflt_view_port_function_table.ptr_2_tml_dp_dflt_cursor_draw      = &tml_dp_x_dflt_cursor_draw;
    dflt_view_port_function_table.ptr_2_tml_dp_get_grafsys_obj       = &tml_dp_x_get_grafsys_obj;
    dflt_view_port_function_table.ptr_2_get_default_font_name        = (char* (*)(...)) &get_x_default_font_name;
    dflt_view_port_function_table.ptr_2_set_default_font_name        = (char* (*)(...)) &set_x_default_font_name;
    dflt_view_port_function_table.grafsys_type                       = tml_xwindows_grafsys;

#ifdef apollo
    gpr_dflt_view_port_function_table[0]  = (GEN_FUNCP) &tml_dp_gpr_get_default_font;
    gpr_dflt_view_port_function_table[1]  = (GEN_FUNCP) &tml_dp_gpr_get_fhandle;
    gpr_dflt_view_port_function_table[2]  = (GEN_FUNCP) &tml_dp_gpr_load_font;
    gpr_dflt_view_port_function_table[3]  = (GEN_FUNCP) &tml_dp_gpr_get_fcharsize;
    gpr_dflt_view_port_function_table[4]  = (GEN_FUNCP) &tml_dp_gpr_get_fcharnum;
    gpr_dflt_view_port_function_table[5]  = (GEN_FUNCP) &tml_dp_gpr_get_fchar1;
    gpr_dflt_view_port_function_table[6]  = (GEN_FUNCP) &tml_dp_gpr_get_fhi;
    gpr_dflt_view_port_function_table[7]  = (GEN_FUNCP) &tml_dp_gpr_get_fpoint;
    gpr_dflt_view_port_function_table[8]  = (GEN_FUNCP) &tml_dp_gpr_get_fbline;
    gpr_dflt_view_port_function_table[9]  = (GEN_FUNCP) &tml_dp_gpr_is_fixed;
    gpr_dflt_view_port_function_table[10] = (GEN_FUNCP) &tml_dp_gpr_get_fwidths;
    gpr_dflt_view_port_function_table[11] = (GEN_FUNCP) &tml_dp_gpr_get_ficspace;
    gpr_dflt_view_port_function_table[12] = (GEN_FUNCP) &tml_dp_gpr_set_ficspace;
    gpr_dflt_view_port_function_table[13] = (GEN_FUNCP) &tml_dp_gpr_get_fspace;
    gpr_dflt_view_port_function_table[14] = (GEN_FUNCP) &tml_dp_gpr_set_fspace;
    gpr_dflt_view_port_function_table[15] = (GEN_FUNCP) &tml_dp_gpr_funload;
    gpr_dflt_view_port_function_table[16] = (GEN_FUNCP) &tml_dp_gpr_frelease;
    gpr_dflt_view_port_function_table[17] = (GEN_FUNCP) &tml_dp_gpr_make_graphic_handle;
    gpr_dflt_view_port_function_table[18] = (GEN_FUNCP) &tml_dp_gpr_delete_graphic_handle;
    gpr_dflt_view_port_function_table[19] = (GEN_FUNCP) &tml_dp_gpr_get_hi;
    gpr_dflt_view_port_function_table[20] = (GEN_FUNCP) &tml_dp_gpr_get_wide;
    gpr_dflt_view_port_function_table[21] = (GEN_FUNCP) &tml_dp_gpr_get_foreg_id;
    gpr_dflt_view_port_function_table[22] = (GEN_FUNCP) &tml_dp_gpr_get_backg_id;
    gpr_dflt_view_port_function_table[23] = (GEN_FUNCP) &tml_dp_gpr_get_fill_backg_id;
    gpr_dflt_view_port_function_table[24] = (GEN_FUNCP) &tml_dp_gpr_get_fill_obj_id;
    gpr_dflt_view_port_function_table[25] = (GEN_FUNCP) &tml_dp_gpr_set_foreg_id;
    gpr_dflt_view_port_function_table[26] = (GEN_FUNCP) &tml_dp_gpr_set_backg_id;
    gpr_dflt_view_port_function_table[27] = (GEN_FUNCP) &tml_dp_gpr_set_fill_backg_id;
    gpr_dflt_view_port_function_table[28] = (GEN_FUNCP) &tml_dp_gpr_set_fill_obj_id;
    gpr_dflt_view_port_function_table[29] = (GEN_FUNCP) &tml_dp_gpr_set_clip;
    gpr_dflt_view_port_function_table[30] = (GEN_FUNCP) &tml_dp_gpr_clear_clip;
    gpr_dflt_view_port_function_table[31] = (GEN_FUNCP) &tml_dp_gpr_clear;
    gpr_dflt_view_port_function_table[32] = (GEN_FUNCP) &tml_dp_gpr_copy;
    gpr_dflt_view_port_function_table[33] = (GEN_FUNCP) &tml_dp_gpr_blt;
    gpr_dflt_view_port_function_table[34] = (GEN_FUNCP) &tml_dp_gpr_string_draw;
    gpr_dflt_view_port_function_table[35] = (GEN_FUNCP) &tml_dp_gpr_string_size;
    gpr_dflt_view_port_function_table[36] = (GEN_FUNCP) &tml_dp_gpr_damage_repair;
    gpr_dflt_view_port_function_table[37] = (GEN_FUNCP) &tml_dp_gpr_dflt_cursor_draw;
    gpr_dflt_view_port_function_table[38] = (GEN_FUNCP) &tml_dp_gpr_get_grafsys_obj;
    gpr_dflt_view_port_function_table[39] = (GEN_FUNCP) &get_gpr_default_font_name;
    gpr_dflt_view_port_function_table[40] = (GEN_FUNCP) &set_gpr_default_font_name;

#endif

    tml_sample_xwindows_grafsys_p   = &dflt_view_port_function_table;
#ifdef apollo
    tml_sample_gpr_grafsys_p        = (vpftbl*) gpr_dflt_view_port_function_table;
#endif

    /* objects used by CSM initialized by call to csm_inlib_entry(); */

    tml_sample_subtree_arg_p        = new subtree_arg ;
    tml_sample_string_arg_p         = new string_arg;
    tml_sample_float_arg_p          = new float_arg;
    tml_sample_int_arg_p            = new int_arg;
    tml_sample_argument_p           = new argument;
    tml_sample_font_prop_p          = new font_property;
    tml_sample_prop_p               = new property;
    tml_sample_pgnode_p             = new pgnode;
    tml_sample_gnode_p              = new gnode;
    tml_sample_bnode_p              = new bnode;

    init_tml_ext_types_part_1();
    init_tml_ext_types_part_2();
    init_tml_types();
    init_tml_type_pointers();

    type_q.head = (tml_type*) &tml_bobj_t;

    init_v1_data();
}

void init_tml_ext_types_part_1()
{
    tml_xwindows_grafsys_ext_t.time_high    = 0x48fd7f15;
    tml_xwindows_grafsys_ext_t.time_low     = 0x5000;
    tml_xwindows_grafsys_ext_t.reserved     = 0x0000;
    tml_xwindows_grafsys_ext_t.family       = 0x0d;
    tml_xwindows_grafsys_ext_t.host.host[0] = 0x00; 
    tml_xwindows_grafsys_ext_t.host.host[1] = 0x02; 
    tml_xwindows_grafsys_ext_t.host.host[2] = 0x2e; 
    tml_xwindows_grafsys_ext_t.host.host[3] = 0xcb; 
    tml_xwindows_grafsys_ext_t.host.host[4] = 0x00; 
    tml_xwindows_grafsys_ext_t.host.host[5] = 0x00; 
    tml_xwindows_grafsys_ext_t.host.host[6] = 0x00;

#ifdef apollo
    tml_gpr_grafsys_ext_t.time_high    = 0x48fd7f20;
    tml_gpr_grafsys_ext_t.time_low     = 0x7000;
    tml_gpr_grafsys_ext_t.reserved     = 0x0000;
    tml_gpr_grafsys_ext_t.family       = 0x0d;
    tml_gpr_grafsys_ext_t.host.host[0] = 0x00; 
    tml_gpr_grafsys_ext_t.host.host[1] = 0x02; 
    tml_gpr_grafsys_ext_t.host.host[2] = 0x2e; 
    tml_gpr_grafsys_ext_t.host.host[3] = 0xcb; 
    tml_gpr_grafsys_ext_t.host.host[4] = 0x00; 
    tml_gpr_grafsys_ext_t.host.host[5] = 0x00; 
    tml_gpr_grafsys_ext_t.host.host[6] = 0x00;
#endif

    tml_union_char_set_val_ext_t.time_high    = 0x41442d1c;
    tml_union_char_set_val_ext_t.time_low     = 0xa000;
    tml_union_char_set_val_ext_t.reserved     = 0x0000;
    tml_union_char_set_val_ext_t.family       = 0x0d;
    tml_union_char_set_val_ext_t.host.host[0] = 0x00;
    tml_union_char_set_val_ext_t.host.host[1] = 0x01;
    tml_union_char_set_val_ext_t.host.host[2] = 0x9f;
    tml_union_char_set_val_ext_t.host.host[3] = 0xab;
    tml_union_char_set_val_ext_t.host.host[4] = 0x00;
    tml_union_char_set_val_ext_t.host.host[5] = 0x00;
    tml_union_char_set_val_ext_t.host.host[6] = 0x00;
    
    tml_char_set_val_ext_t.time_high    = 0x41442c21;
    tml_char_set_val_ext_t.time_low     = 0x5000;
    tml_char_set_val_ext_t.reserved     = 0x0000;
    tml_char_set_val_ext_t.family       = 0x0d;
    tml_char_set_val_ext_t.host.host[0] = 0x00;
    tml_char_set_val_ext_t.host.host[1] = 0x01;
    tml_char_set_val_ext_t.host.host[2] = 0x9f;
    tml_char_set_val_ext_t.host.host[3] = 0xab;
    tml_char_set_val_ext_t.host.host[4] = 0x00;
    tml_char_set_val_ext_t.host.host[5] = 0x00;
    tml_char_set_val_ext_t.host.host[6] = 0x00;
    
    tml_subtree_val_ext_t.time_high    = 0x405b4f33;
    tml_subtree_val_ext_t.time_low     = 0x4000;
    tml_subtree_val_ext_t.reserved     = 0x0000;
    tml_subtree_val_ext_t.family       = 0x0d;
    tml_subtree_val_ext_t.host.host[0] = 0x00;
    tml_subtree_val_ext_t.host.host[1] = 0x01;
    tml_subtree_val_ext_t.host.host[2] = 0x9f;
    tml_subtree_val_ext_t.host.host[3] = 0xab;
    tml_subtree_val_ext_t.host.host[4] = 0x00;
    tml_subtree_val_ext_t.host.host[5] = 0x00;
    tml_subtree_val_ext_t.host.host[6] = 0x00;
    
    tml_string_val_ext_t.time_high    = 0x405b4f0a;
    tml_string_val_ext_t.time_low     = 0xf000;
    tml_string_val_ext_t.reserved     = 0x0000;
    tml_string_val_ext_t.family       = 0x0d;
    tml_string_val_ext_t.host.host[0] = 0x00;
    tml_string_val_ext_t.host.host[1] = 0x01;
    tml_string_val_ext_t.host.host[2] = 0x9f;
    tml_string_val_ext_t.host.host[3] = 0xab;
    tml_string_val_ext_t.host.host[4] = 0x00;
    tml_string_val_ext_t.host.host[5] = 0x00;
    tml_string_val_ext_t.host.host[6] = 0x00;
    
    tml_float_val_ext_t.time_high    = 0x405b4ec2;
    tml_float_val_ext_t.time_low     = 0x9000;
    tml_float_val_ext_t.reserved     = 0x0000;
    tml_float_val_ext_t.family       = 0x0d;
    tml_float_val_ext_t.host.host[0] = 0x00;
    tml_float_val_ext_t.host.host[1] = 0x01;
    tml_float_val_ext_t.host.host[2] = 0x9f;
    tml_float_val_ext_t.host.host[3] = 0xab;
    tml_float_val_ext_t.host.host[4] = 0x00;
    tml_float_val_ext_t.host.host[5] = 0x00;
    tml_float_val_ext_t.host.host[6] = 0x00;
    
    tml_int_val_ext_t.time_high    = 0x405b4e8f;
    tml_int_val_ext_t.time_low     = 0x5000;
    tml_int_val_ext_t.reserved     = 0x0000;
    tml_int_val_ext_t.family       = 0x0d;
    tml_int_val_ext_t.host.host[0] = 0x00;
    tml_int_val_ext_t.host.host[1] = 0x01;
    tml_int_val_ext_t.host.host[2] = 0x9f;
    tml_int_val_ext_t.host.host[3] = 0xab;
    tml_int_val_ext_t.host.host[4] = 0x00;
    tml_int_val_ext_t.host.host[5] = 0x00;
    tml_int_val_ext_t.host.host[6] = 0x00;
    
    tml_no_val_ext_t.time_high    = 0x405b4e60;
    tml_no_val_ext_t.time_low     = 0x2000;
    tml_no_val_ext_t.reserved     = 0x0000;
    tml_no_val_ext_t.family       = 0x0d;
    tml_no_val_ext_t.host.host[0] = 0x00;
    tml_no_val_ext_t.host.host[1] = 0x01;
    tml_no_val_ext_t.host.host[2] = 0x9f;
    tml_no_val_ext_t.host.host[3] = 0xab;
    tml_no_val_ext_t.host.host[4] = 0x00;
    tml_no_val_ext_t.host.host[5] = 0x00;
    tml_no_val_ext_t.host.host[6] = 0x00;
    
    tml_tab_prop_ext_t.time_high    = 0x42fbf155;
    tml_tab_prop_ext_t.time_low     = 0x5000;
    tml_tab_prop_ext_t.reserved     = 0x0000;
    tml_tab_prop_ext_t.family       = 0x0d;
    tml_tab_prop_ext_t.host.host[0] = 0x00;
    tml_tab_prop_ext_t.host.host[1] = 0x01;
    tml_tab_prop_ext_t.host.host[2] = 0x19;
    tml_tab_prop_ext_t.host.host[3] = 0x9f;
    tml_tab_prop_ext_t.host.host[4] = 0x00;
    tml_tab_prop_ext_t.host.host[5] = 0x00;
    tml_tab_prop_ext_t.host.host[6] = 0x00;
    
    tml_union_char_set_prop_ext_t.time_high    = 0x4115f074;
    tml_union_char_set_prop_ext_t.time_low     = 0x1000;
    tml_union_char_set_prop_ext_t.reserved     = 0x0000;
    tml_union_char_set_prop_ext_t.family       = 0x0d;
    tml_union_char_set_prop_ext_t.host.host[0] = 0x00;
    tml_union_char_set_prop_ext_t.host.host[1] = 0x01;
    tml_union_char_set_prop_ext_t.host.host[2] = 0x9f;
    tml_union_char_set_prop_ext_t.host.host[3] = 0xab;
    tml_union_char_set_prop_ext_t.host.host[4] = 0x00;
    tml_union_char_set_prop_ext_t.host.host[5] = 0x00;
    tml_union_char_set_prop_ext_t.host.host[6] = 0x00;
    
    tml_char_set_prop_ext_t.time_high    = 0x40f60484;
    tml_char_set_prop_ext_t.time_low     = 0xa000;
    tml_char_set_prop_ext_t.reserved     = 0x0000;
    tml_char_set_prop_ext_t.family       = 0x0d;
    tml_char_set_prop_ext_t.host.host[0] = 0x00;
    tml_char_set_prop_ext_t.host.host[1] = 0x01;
    tml_char_set_prop_ext_t.host.host[2] = 0x9f;
    tml_char_set_prop_ext_t.host.host[3] = 0xab;
    tml_char_set_prop_ext_t.host.host[4] = 0x00;
    tml_char_set_prop_ext_t.host.host[5] = 0x00;
    tml_char_set_prop_ext_t.host.host[6] = 0x00;
    
    tml_form_feed_prop_ext_t.time_high    = 0x405b4416;
    tml_form_feed_prop_ext_t.time_low     = 0xe000;
    tml_form_feed_prop_ext_t.reserved     = 0x0000;
    tml_form_feed_prop_ext_t.family       = 0x0d;
    tml_form_feed_prop_ext_t.host.host[0] = 0x00;
    tml_form_feed_prop_ext_t.host.host[1] = 0x01;
    tml_form_feed_prop_ext_t.host.host[2] = 0x9f;
    tml_form_feed_prop_ext_t.host.host[3] = 0xab;
    tml_form_feed_prop_ext_t.host.host[4] = 0x00;
    tml_form_feed_prop_ext_t.host.host[5] = 0x00;
    tml_form_feed_prop_ext_t.host.host[6] = 0x00;
}
    
void init_tml_ext_types_part_2()
{
    tml_footer_prop_ext_t.time_high    = 0x405b43d9;
    tml_footer_prop_ext_t.time_low     = 0xa000;
    tml_footer_prop_ext_t.reserved     = 0x0000;
    tml_footer_prop_ext_t.family       = 0x0d;
    tml_footer_prop_ext_t.host.host[0] = 0x00;
    tml_footer_prop_ext_t.host.host[1] = 0x01;
    tml_footer_prop_ext_t.host.host[2] = 0x9f;
    tml_footer_prop_ext_t.host.host[3] = 0xab;
    tml_footer_prop_ext_t.host.host[4] = 0x00;
    tml_footer_prop_ext_t.host.host[5] = 0x00;
    tml_footer_prop_ext_t.host.host[6] = 0x00;
    
    tml_header_prop_ext_t.time_high    = 0x405b439d;
    tml_header_prop_ext_t.time_low     = 0x7000;
    tml_header_prop_ext_t.reserved     = 0x0000;
    tml_header_prop_ext_t.family       = 0x0d;
    tml_header_prop_ext_t.host.host[0] = 0x00;
    tml_header_prop_ext_t.host.host[1] = 0x01;
    tml_header_prop_ext_t.host.host[2] = 0x9f;
    tml_header_prop_ext_t.host.host[3] = 0xab;
    tml_header_prop_ext_t.host.host[4] = 0x00;
    tml_header_prop_ext_t.host.host[5] = 0x00;
    tml_header_prop_ext_t.host.host[6] = 0x00;
    
    tml_pgsize_prop_ext_t.time_high    = 0x405b437b;
    tml_pgsize_prop_ext_t.time_low     = 0x4000;
    tml_pgsize_prop_ext_t.reserved     = 0x0000;
    tml_pgsize_prop_ext_t.family       = 0x0d;
    tml_pgsize_prop_ext_t.host.host[0] = 0x00;
    tml_pgsize_prop_ext_t.host.host[1] = 0x01;
    tml_pgsize_prop_ext_t.host.host[2] = 0x9f;
    tml_pgsize_prop_ext_t.host.host[3] = 0xab;
    tml_pgsize_prop_ext_t.host.host[4] = 0x00;
    tml_pgsize_prop_ext_t.host.host[5] = 0x00;
    tml_pgsize_prop_ext_t.host.host[6] = 0x00;
    
    tml_inv_prop_ext_t.time_high    = 0x405b4345;
    tml_inv_prop_ext_t.time_low     = 0x1000;
    tml_inv_prop_ext_t.reserved     = 0x0000;
    tml_inv_prop_ext_t.family       = 0x0d;
    tml_inv_prop_ext_t.host.host[0] = 0x00;
    tml_inv_prop_ext_t.host.host[1] = 0x01;
    tml_inv_prop_ext_t.host.host[2] = 0x9f;
    tml_inv_prop_ext_t.host.host[3] = 0xab;
    tml_inv_prop_ext_t.host.host[4] = 0x00;
    tml_inv_prop_ext_t.host.host[5] = 0x00;
    tml_inv_prop_ext_t.host.host[6] = 0x00;
    
    tml_font_prop_ext_t.time_high    = 0x405b430e;
    tml_font_prop_ext_t.time_low     = 0xd000;
    tml_font_prop_ext_t.reserved     = 0x0000;
    tml_font_prop_ext_t.family       = 0x0d;
    tml_font_prop_ext_t.host.host[0] = 0x00;
    tml_font_prop_ext_t.host.host[1] = 0x01;
    tml_font_prop_ext_t.host.host[2] = 0x9f;
    tml_font_prop_ext_t.host.host[3] = 0xab;
    tml_font_prop_ext_t.host.host[4] = 0x00;
    tml_font_prop_ext_t.host.host[5] = 0x00;
    tml_font_prop_ext_t.host.host[6] = 0x00;
    
    tml_margin_prop_ext_t.time_high    = 0x405b42b2;
    tml_margin_prop_ext_t.time_low     = 0xa000;
    tml_margin_prop_ext_t.reserved     = 0x0000;
    tml_margin_prop_ext_t.family       = 0x0d;
    tml_margin_prop_ext_t.host.host[0] = 0x00;
    tml_margin_prop_ext_t.host.host[1] = 0x01;
    tml_margin_prop_ext_t.host.host[2] = 0x9f;
    tml_margin_prop_ext_t.host.host[3] = 0xab;
    tml_margin_prop_ext_t.host.host[4] = 0x00;
    tml_margin_prop_ext_t.host.host[5] = 0x00;
    tml_margin_prop_ext_t.host.host[6] = 0x00;
    
    tml_wrap_prop_ext_t.time_high    = 0x405b4287;
    tml_wrap_prop_ext_t.time_low     = 0x7000;
    tml_wrap_prop_ext_t.reserved     = 0x0000;
    tml_wrap_prop_ext_t.family       = 0x0d;
    tml_wrap_prop_ext_t.host.host[0] = 0x00;
    tml_wrap_prop_ext_t.host.host[1] = 0x01;
    tml_wrap_prop_ext_t.host.host[2] = 0x9f;
    tml_wrap_prop_ext_t.host.host[3] = 0xab;
    tml_wrap_prop_ext_t.host.host[4] = 0x00;
    tml_wrap_prop_ext_t.host.host[5] = 0x00;
    tml_wrap_prop_ext_t.host.host[6] = 0x00;
    
    tml_undefined_prop_ext_t.time_high    = 0x405b41fe;
    tml_undefined_prop_ext_t.time_low     = 0x3000;
    tml_undefined_prop_ext_t.reserved     = 0x0000;
    tml_undefined_prop_ext_t.family       = 0x0d;
    tml_undefined_prop_ext_t.host.host[0] = 0x00;
    tml_undefined_prop_ext_t.host.host[1] = 0x01;
    tml_undefined_prop_ext_t.host.host[2] = 0x9f;
    tml_undefined_prop_ext_t.host.host[3] = 0xab;
    tml_undefined_prop_ext_t.host.host[4] = 0x00;
    tml_undefined_prop_ext_t.host.host[5] = 0x00;
    tml_undefined_prop_ext_t.host.host[6] = 0x00;
    
    tml_free_ext_t.time_high    = 0x405b3cda;
    tml_free_ext_t.time_low     = 0x8000;
    tml_free_ext_t.reserved     = 0x0000;
    tml_free_ext_t.family       = 0x0d;
    tml_free_ext_t.host.host[0] = 0x00;
    tml_free_ext_t.host.host[1] = 0x01;
    tml_free_ext_t.host.host[2] = 0x9f;
    tml_free_ext_t.host.host[3] = 0xab;
    tml_free_ext_t.host.host[4] = 0x00;
    tml_free_ext_t.host.host[5] = 0x00;
    tml_free_ext_t.host.host[6] = 0x00;
    
    tml_leaf_ext_t.time_high    = 0x405b3c4c;
    tml_leaf_ext_t.time_low     = 0x2000;
    tml_leaf_ext_t.reserved     = 0x0000;
    tml_leaf_ext_t.family       = 0x0d;
    tml_leaf_ext_t.host.host[0] = 0x00;
    tml_leaf_ext_t.host.host[1] = 0x01;
    tml_leaf_ext_t.host.host[2] = 0x9f;
    tml_leaf_ext_t.host.host[3] = 0xab;
    tml_leaf_ext_t.host.host[4] = 0x00;
    tml_leaf_ext_t.host.host[5] = 0x00;
    tml_leaf_ext_t.host.host[6] = 0x00;
    
    tml_pgobj_ext_t.time_high    = 0x405b3c99;
    tml_pgobj_ext_t.time_low     = 0x5000;
    tml_pgobj_ext_t.reserved     = 0x0000;
    tml_pgobj_ext_t.family       = 0x0d;
    tml_pgobj_ext_t.host.host[0] = 0x00;
    tml_pgobj_ext_t.host.host[1] = 0x01;
    tml_pgobj_ext_t.host.host[2] = 0x9f;
    tml_pgobj_ext_t.host.host[3] = 0xab;
    tml_pgobj_ext_t.host.host[4] = 0x00;
    tml_pgobj_ext_t.host.host[5] = 0x00;
    tml_pgobj_ext_t.host.host[6] = 0x00;
    
    tml_gobj_ext_t.time_high    = 0x405b3c0b;
    tml_gobj_ext_t.time_low     = 0xf000;
    tml_gobj_ext_t.reserved     = 0x0000;
    tml_gobj_ext_t.family       = 0x0d;
    tml_gobj_ext_t.host.host[0] = 0x00;
    tml_gobj_ext_t.host.host[1] = 0x01;
    tml_gobj_ext_t.host.host[2] = 0x9f;
    tml_gobj_ext_t.host.host[3] = 0xab;
    tml_gobj_ext_t.host.host[4] = 0x00;
    tml_gobj_ext_t.host.host[5] = 0x00;
    tml_gobj_ext_t.host.host[6] = 0x00;
    
    tml_branch_ext_t.time_high    = 0x405b3b66;
    tml_branch_ext_t.time_low     = 0xb000;
    tml_branch_ext_t.reserved     = 0x0000;
    tml_branch_ext_t.family       = 0x0d;
    tml_branch_ext_t.host.host[0] = 0x00;
    tml_branch_ext_t.host.host[1] = 0x01;
    tml_branch_ext_t.host.host[2] = 0x9f;
    tml_branch_ext_t.host.host[3] = 0xab;
    tml_branch_ext_t.host.host[4] = 0x00;
    tml_branch_ext_t.host.host[5] = 0x00;
    tml_branch_ext_t.host.host[6] = 0x00;
    
    tml_bobj_ext_t.time_high    = 0x405b3b06;
    tml_bobj_ext_t.time_low     = 0x7000;
    tml_bobj_ext_t.reserved     = 0x0000;
    tml_bobj_ext_t.family       = 0x0d;
    tml_bobj_ext_t.host.host[0] = 0x00;
    tml_bobj_ext_t.host.host[1] = 0x01;
    tml_bobj_ext_t.host.host[2] = 0x9f;
    tml_bobj_ext_t.host.host[3] = 0xab;
    tml_bobj_ext_t.host.host[4] = 0x00;
    tml_bobj_ext_t.host.host[5] = 0x00;
    tml_bobj_ext_t.host.host[6] = 0x00;
    
    tml_undefined_node_ext_t.time_high    = 0x405b39e7;
    tml_undefined_node_ext_t.time_low     = 0x3000;
    tml_undefined_node_ext_t.reserved     = 0x0000;
    tml_undefined_node_ext_t.family       = 0x0d;
    tml_undefined_node_ext_t.host.host[0] = 0x00;
    tml_undefined_node_ext_t.host.host[1] = 0x01;
    tml_undefined_node_ext_t.host.host[2] = 0x9f;
    tml_undefined_node_ext_t.host.host[3] = 0xab;
    tml_undefined_node_ext_t.host.host[4] = 0x00;
    tml_undefined_node_ext_t.host.host[5] = 0x00;
    tml_undefined_node_ext_t.host.host[6] = 0x00;
}

void init_tml_types()
{
   /* Shadow pointers are necessary due to changes in declarations of the 
      following constants from const tml_type* to tml_type* const.
       Changes resulted from cxx2.0 port */

    tml_xwindows_grafsys_t.type        = &tml_xwindows_grafsys_ext_t;
    tml_xwindows_grafsys_t.instance    = tml_sample_xwindows_grafsys_p;
    tml_xwindows_grafsys_t.next        = 0;

#ifdef apollo
    tml_gpr_grafsys_t.type             = &tml_gpr_grafsys_ext_t;
    tml_gpr_grafsys_t.instance         = tml_sample_gpr_grafsys_p;
    tml_gpr_grafsys_t.next             = &tml_xwindows_grafsys_t;
#endif

    tml_subtree_val_t.type             = &tml_subtree_val_ext_t;
    tml_subtree_val_t.instance         = tml_sample_subtree_arg_p;
    tml_subtree_val_t.next             = &tml_char_set_val_t;
    
    tml_string_val_t.type              = &tml_string_val_ext_t;
    tml_string_val_t.instance          = tml_sample_string_arg_p;
    tml_string_val_t.next              = &tml_subtree_val_t;
    
    tml_float_val_t.type               = &tml_float_val_ext_t;
    tml_float_val_t.instance           = tml_sample_float_arg_p;
    tml_float_val_t.next               = &tml_string_val_t;
    
    tml_int_val_t.type                 = &tml_int_val_ext_t;
    tml_int_val_t.instance             = tml_sample_int_arg_p;
    tml_int_val_t.next                 = &tml_float_val_t;
    
    tml_no_val_t.type                  = &tml_no_val_ext_t;
    tml_no_val_t.instance              = tml_sample_argument_p;
    tml_no_val_t.next                  = &tml_int_val_t;
    
    tml_tab_prop_t.type                = &tml_tab_prop_ext_t;
    tml_tab_prop_t.instance            = tml_sample_prop_p;
    tml_tab_prop_t.next                = &tml_no_val_t;
    
    tml_union_char_set_prop_t.type     = &tml_union_char_set_prop_ext_t;
    tml_union_char_set_prop_t.instance = tml_sample_prop_p;
    tml_union_char_set_prop_t.next     = &tml_tab_prop_t;
    
    tml_char_set_prop_t.type           = &tml_char_set_prop_ext_t;
    tml_char_set_prop_t.instance       = tml_sample_prop_p;
    tml_char_set_prop_t.next           = &tml_union_char_set_prop_t;
    
    tml_form_feed_prop_t.type          = &tml_form_feed_prop_ext_t;
    tml_form_feed_prop_t.instance      = tml_sample_prop_p;
    tml_form_feed_prop_t.next          = &tml_char_set_prop_t;
    
    tml_footer_prop_t.type             = &tml_footer_prop_ext_t;
    tml_footer_prop_t.instance         = tml_sample_prop_p;
    tml_footer_prop_t.next             = &tml_form_feed_prop_t;
    
    tml_header_prop_t.type             = &tml_header_prop_ext_t;
    tml_header_prop_t.instance         = tml_sample_prop_p;
    tml_header_prop_t.next             = &tml_footer_prop_t;
    
    tml_pgsize_prop_t.type             = &tml_pgsize_prop_ext_t;
    tml_pgsize_prop_t.instance         = tml_sample_prop_p;
    tml_pgsize_prop_t.next             = &tml_header_prop_t;
    
    tml_inv_prop_t.type                = &tml_inv_prop_ext_t;
    tml_inv_prop_t.instance            = tml_sample_prop_p;
    tml_inv_prop_t.next                = &tml_pgsize_prop_t;
    
    tml_font_prop_t.type               = &tml_font_prop_ext_t;
    tml_font_prop_t.instance           = tml_sample_font_prop_p;
    tml_font_prop_t.next               = &tml_inv_prop_t;
    
    tml_margin_prop_t.type             = &tml_margin_prop_ext_t;
    tml_margin_prop_t.instance         = tml_sample_prop_p;
    tml_margin_prop_t.next             = &tml_font_prop_t;
    
    tml_wrap_prop_t.type               = &tml_wrap_prop_ext_t;
    tml_wrap_prop_t.instance           = tml_sample_prop_p;
    tml_wrap_prop_t.next               = &tml_margin_prop_t;
    
    tml_undefined_prop_t.type          = &tml_undefined_prop_ext_t;
    tml_undefined_prop_t.instance      = 0;
    tml_undefined_prop_t.next          = 0;
    
    tml_free_t.type                    = &tml_free_ext_t;
    tml_free_t.instance                = 0;
    tml_free_t.next                    = 0;
    
    tml_leaf_t.type                    = &tml_leaf_ext_t;
    tml_leaf_t.instance                = 0;
    tml_leaf_t.next                    = 0;
    
    tml_pgobj_t.type                   = &tml_pgobj_ext_t;
    tml_pgobj_t.instance               = tml_sample_pgnode_p;
    tml_pgobj_t.next                   = &tml_wrap_prop_t;
    
    tml_gobj_t.type                    = &tml_gobj_ext_t;
    tml_gobj_t.instance                = tml_sample_gnode_p;
    tml_gobj_t.next                    = &tml_pgobj_t;
    
    tml_branch_t.type                  = &tml_branch_ext_t;
    tml_branch_t.instance              = 0;
    tml_branch_t.next                  = 0;
    
    tml_bobj_t.type                    = &tml_bobj_ext_t;
    tml_bobj_t.instance                = tml_sample_bnode_p;
    tml_bobj_t.next                    = &tml_gobj_t;
    
    tml_undefined_node_t.type          = &tml_undefined_node_ext_t;
    tml_undefined_node_t.instance      = 0;
    tml_undefined_node_t.next          = 0;
}

void init_tml_type_pointers()
{
    tml_type** ptr_2_tml_xwindows_grafsys;
#ifdef apollo
    tml_type** ptr_2_tml_gpr_grafsys;
#endif
    tml_type** ptr_2_union_char_set_val_arg;
    tml_type** ptr_2_char_set_val_arg;
    tml_type** ptr_2_subtree_val_arg;
    tml_type** ptr_2_string_val_arg;
    tml_type** ptr_2_float_val_arg;
    tml_type** ptr_2_int_val_arg;
    tml_type** ptr_2_no_val_arg;
    tml_type** ptr_2_tml_tab_prop;
    tml_type** ptr_2_tml_union_char_set_prop;
    tml_type** ptr_2_tml_char_set_prop;
    tml_type** ptr_2_tml_form_feed_prop;
    tml_type** ptr_2_tml_footer_prop;
    tml_type** ptr_2_tml_header_prop;
    tml_type** ptr_2_tml_pgsize_prop;
    tml_type** ptr_2_tml_inv_prop;
    tml_type** ptr_2_tml_font_prop;
    tml_type** ptr_2_tml_margin_prop;
    tml_type** ptr_2_tml_wrap_prop;
    tml_type** ptr_2_undefined_prop;
    tml_type** ptr_2_is_free;
    tml_type** ptr_2_leaf;
    tml_type** ptr_2_pgobj;
    tml_type** ptr_2_gobj;
    tml_type** ptr_2_branch;
    tml_type** ptr_2_bobj;
    tml_type** ptr_2_undefined_node;

    ptr_2_tml_xwindows_grafsys     = (tml_type**)&tml_xwindows_grafsys;
    *ptr_2_tml_xwindows_grafsys    = &tml_xwindows_grafsys_t;
#ifdef apollo
    ptr_2_tml_gpr_grafsys          = (tml_type**)&tml_gpr_grafsys;
    *ptr_2_tml_gpr_grafsys         = &tml_gpr_grafsys_t;
#endif
    ptr_2_union_char_set_val_arg   = (tml_type**)&union_char_set_val_arg;
    *ptr_2_union_char_set_val_arg  = &tml_union_char_set_val_t;
    ptr_2_char_set_val_arg         = (tml_type**)&char_set_val_arg;
    *ptr_2_char_set_val_arg        = &tml_char_set_val_t;
    ptr_2_subtree_val_arg          = (tml_type**)&subtree_val_arg;
    *ptr_2_subtree_val_arg         = &tml_subtree_val_t;
    ptr_2_string_val_arg           = (tml_type**)&string_val_arg;
    *ptr_2_string_val_arg          = &tml_string_val_t;
    ptr_2_float_val_arg            = (tml_type**)&float_val_arg;
    *ptr_2_float_val_arg           = &tml_float_val_t;
    ptr_2_int_val_arg              = (tml_type**)&int_val_arg;
    *ptr_2_int_val_arg             = &tml_int_val_t;
    ptr_2_no_val_arg               = (tml_type**)&no_val_arg;
    *ptr_2_no_val_arg              = &tml_no_val_t;
    ptr_2_tml_tab_prop             = (tml_type**)&tml_tab_prop;
    *ptr_2_tml_tab_prop            = &tml_tab_prop_t;
    ptr_2_tml_union_char_set_prop  = (tml_type**)&tml_union_char_set_prop;
    *ptr_2_tml_union_char_set_prop = &tml_union_char_set_prop_t;
    ptr_2_tml_char_set_prop        = (tml_type**)&tml_char_set_prop;
    *ptr_2_tml_char_set_prop       = &tml_char_set_prop_t;
    ptr_2_tml_form_feed_prop       = (tml_type**)&tml_form_feed_prop;
    *ptr_2_tml_form_feed_prop      = &tml_form_feed_prop_t;
    ptr_2_tml_footer_prop          = (tml_type**)&tml_footer_prop;
    *ptr_2_tml_footer_prop         = &tml_footer_prop_t;
    ptr_2_tml_header_prop          = (tml_type**)&tml_header_prop;
    *ptr_2_tml_header_prop        = &tml_header_prop_t;
    ptr_2_tml_pgsize_prop          = (tml_type**)&tml_pgsize_prop;
    *ptr_2_tml_pgsize_prop         = &tml_pgsize_prop_t;
    ptr_2_tml_inv_prop             = (tml_type**)&tml_inv_prop;
    *ptr_2_tml_inv_prop            = &tml_inv_prop_t;
    ptr_2_tml_font_prop            = (tml_type**)&tml_font_prop;
    *ptr_2_tml_font_prop           = &tml_font_prop_t;
    ptr_2_tml_margin_prop          = (tml_type**)&tml_margin_prop;
    *ptr_2_tml_margin_prop         = &tml_margin_prop_t;
    ptr_2_tml_wrap_prop            = (tml_type**)&tml_wrap_prop;
    *ptr_2_tml_wrap_prop           = &tml_wrap_prop_t;
    ptr_2_undefined_prop           = (tml_type**)&undefined_prop;
    *ptr_2_undefined_prop          = &tml_undefined_prop_t;
    ptr_2_is_free                  = (tml_type**)&is_free;
    *ptr_2_is_free                 = &tml_free_t;
    ptr_2_leaf                     = (tml_type**)&leaf;
    *ptr_2_leaf                    = &tml_leaf_t;
    ptr_2_pgobj                    = (tml_type**)&pgobj;
    *ptr_2_pgobj                   = &tml_pgobj_t;
    ptr_2_gobj                     = (tml_type**)&gobj;
    *ptr_2_gobj                    = &tml_gobj_t;
    ptr_2_branch                   = (tml_type**)&branch;
    *ptr_2_branch                  = &tml_branch_t;
    ptr_2_bobj                     = (tml_type**)&bobj;
    *ptr_2_bobj                    = &tml_bobj_t;
    ptr_2_undefined_node           = (tml_type**)&undefined_node;
    *ptr_2_undefined_node          = &tml_undefined_node_t;
}

void init_v1_data()
{
    extern bnode* extend_cxx_make_bnode(model* in_model);
    extern gnode* extend_cxx_make_gnode(model* in_model);
    extern property* extend_cxx_make_property(model* in_model);
    extern font_property* extend_cxx_make_font_property(model* in_model);

    tml_pgsize_prop_ext_t_v1.high      = 0;
    tml_pgsize_prop_ext_t_v1.low       = 11;
    
    tml_inv_prop_ext_t_v1.high         = 0;
    tml_inv_prop_ext_t_v1.low          = 10;
    
    tml_font_prop_ext_t_v1.high        = 0;
    tml_font_prop_ext_t_v1.low         = 9;
    
    tml_margin_prop_ext_t_v1.high      = 0;
    tml_margin_prop_ext_t_v1.low       = 8;
    
    tml_wrap_prop_ext_t_v1.high        = 0;
    tml_wrap_prop_ext_t_v1.low         = 7;
    
    gobj_ext_t_v1.high                 = 0;
    gobj_ext_t_v1.low                  = 3;
    
    bobj_ext_t_v1.high                 = 0;
    bobj_ext_t_v1.low                  = 1;

    tml_pgsize_trans.type_v1  = &tml_pgsize_prop_ext_t_v1;
    tml_pgsize_trans.type     = &tml_pgsize_prop_t;
    tml_pgsize_trans.make_func = (GEN_FUNCP) &extend_cxx_make_property;
    tml_pgsize_trans.next     = 0;
    
    tml_inv_trans.type_v1     = &tml_inv_prop_ext_t_v1;
    tml_inv_trans.type        = &tml_inv_prop_t;
    tml_inv_trans.make_func   = (GEN_FUNCP) &extend_cxx_make_property;
    tml_inv_trans.next        = &tml_pgsize_trans;
    
    tml_font_trans.type_v1    = &tml_font_prop_ext_t_v1;
    tml_font_trans.type       = &tml_font_prop_t;
    tml_font_trans.make_func  = (GEN_FUNCP) &extend_cxx_make_font_property;
    tml_font_trans.next       = &tml_inv_trans;
    
    tml_margin_trans.type_v1  = &tml_margin_prop_ext_t_v1;
    tml_margin_trans.type     = &tml_margin_prop_t;
    tml_margin_trans.make_func = (GEN_FUNCP) &extend_cxx_make_property;
    tml_margin_trans.next     = &tml_font_trans;
    
    tml_wrap_trans.type_v1    = &tml_wrap_prop_ext_t_v1;
    tml_wrap_trans.type       = &tml_wrap_prop_t;
    tml_wrap_trans.make_func  = (GEN_FUNCP) &extend_cxx_make_property;
    tml_wrap_trans.next       = &tml_margin_trans;
    
    tml_gobj_trans.type_v1    = &gobj_ext_t_v1;
    tml_gobj_trans.type       = &tml_gobj_t;
    tml_gobj_trans.make_func  = (GEN_FUNCP) &extend_cxx_make_gnode;
    tml_gobj_trans.next       = &tml_wrap_trans;
    
    tml_bobj_trans.type_v1    = &bobj_ext_t_v1;
    tml_bobj_trans.type       = &tml_bobj_t;
    tml_bobj_trans.make_func  = (GEN_FUNCP) &extend_cxx_make_bnode;
    tml_bobj_trans.next       = &tml_gobj_trans;

    type_trans_q_v1.head = &tml_bobj_trans;
    type_trans_q_v1.tail = &tml_pgsize_trans;
}

