/**************************************************************************************
    tml_os_depends.h
    definitions of system- and machine-dependent macros for tml++
    declarations of the operating system dependent routines for tml++

    (c) Copyright Apollo Computer Inc., 1987 (unpublished)
     --------------------------------------------------------------------
    |THE FOLLOWING PROGRAMS ARE THE SOLE PROPERTY OF APOLLO COMPUTER INC.|
    |      AND CONTAIN ITS PROPRIETARY AND CONFIDENTIAL INFORMATION.     |
     --------------------------------------------------------------------
**************************************************************************************/

#ifndef tml_os_depends_h
#define tml_os_depends_h

#ifdef apollo
#define ALIGN_BLOCK 4   /* modulus in bytes of universal alignment boundary */
#endif

#ifdef bsd
#include <fcntl.h>
#endif

#ifdef sys5
#include <fcntl.h>
#endif

#ifdef aegis
#include <fcntl.h>
#endif

#ifdef __STDC__
/* From here to "#else", declarations for ANSI C programmers

    /* copies count bytes from array from to array to--does */
    /*  not check sizes of either. Returns "to" if all goes well */

extern char* tml_c_tml_os_byte_copy(char* from, char* to, long count);

    /* typedef for comparison function argument to tml_c_tml_os_qsort() */

    typedef long *(CMP)(void*, void*);

    /* sorts array base of nel elements each of size size using comparison */
    /*   function pointed to by cmp */

extern void tml_c_tml_os_qsort(char *base, unsigned long nel, unsigned long size, CMP cmp);

    /* open file named by (possibly non-null-terminated) pathname of length */
    /*   name_len for purposes and under conditions specified by oflag with */
    /*   permissions mode. Returns file handle suitable for host system, or */
    /*   NULL on failure. */

extern void *tml_c_tml_os_open(char *pathname, long name_len, long oflag, long mode);

    /* read nbyte characters into buf from file associated with file_handle */
    /*   obtained by a call to tml_os_open(); return bytes read, 0 on EOF,  */
    /*   -1 on failure */

extern long tml_c_tml_os_read(void *file_handle, char *buf, unsigned long nbyte);

    /* write nbyte characters from buf into file associated with file_handle */
    /*   obtained by a call to tml_os_open(); return bytes written, 0 on EOF, */
    /*   -1 on failure */

extern long tml_c_tml_os_write(void *file_handle, char *buf, unsigned long nbyte);

    /* close file associated with file_handle (obtained from tml_os_open); */
    /*   return -1 on failure, zero otherwise. */

extern long tml_c_tml_os_close(void *file_handle);

    /* rewind file associated with file_handle (obtained from tml_os_open); */
    /*   return -1 on failure, zero otherwise. */

extern long tml_c_tml_os_rewind(void *file_handle);

    /* return NULL on failure or pointer to nelem * elsize bytes of free space. */
    /*   Each byte is initialized to zero. */

extern char* tml_c_tml_os_alloc(unsigned long nelem, unsigned long elsize);

    /* free the space at ptr, which must have been obtained from a call to */
    /*   tml_os_alloc(), or the result is unpredictible. */

extern void tml_c_tml_os_free(char* ptr);

    /* return -1 if address at ptr1 precedes address at ptr2, 0 if they */
    /*   are equal, 1 if address at ptr1 follows address at ptr2. The */
    /*   comparison algorithm and definitions of precedes and follows are */
    /*   system dependent */

extern long tml_c_tml_os_ptr_cmp(char* *ptr1, char* *ptr2);

    /* initialize a semaphore structure with a given capacity */

extern void* tml_os_init_sem(long pool_capacity);

    /* free a semaphore structure */

extern void tml_os_free_sem(void* old_sem);

    /* busy wait on a semaphore */

extern long tml_os_busy_wait(void* on_sem);

    /* lock a semaphore */

extern void tml_os_p_sem(void* on_sem);

    /* continue process waiting for semaphore */

extern void* tml_c_tml_os_v_sem(void* proc_id, void* on_sem);

    /* check availability of a resource via its semaphore */

extern long tml_os_resource_available(void* on_sem);

    /* get process identifier for the current process -- returns a pointer */
    /*   to the appropriate object cast as a void* */

extern void* tml_c_tml_os_getpid(void);

    /* reduce vulnerability of the current process to interrupts during critical */
    /*   code sections -- returns status of interrupts before the call */

extern void* tml_c_tml_os_ignore_interrupts(void);

    /* restore vulnerability of the current process to interrupts after completing */
    /*   critical code sections -- returns status of interrupts after the call */

extern void* tml_c_tml_os_restore_interrupts(void* interrupts);

    /* kill a process dead -- returns a pointer to the appropriate object */
    /*   cast as a void* */

extern void* tml_c_tml_os_kill_proc(void* proc_id);

#else

/* From here to "#endif", declarations for programmers using vanilla C */

    /* copies count bytes from array from to array to--does */
    /*  not check sizes of either. Returns "to" if all goes well */
    /* ARGS: char* from, char* to, long count */

extern char* tml_c_tml_os_byte_copy();

    /* typedef for comparison function argument to tml_c_tml_os_qsort() */

        /* ARGS: char *, char * */

    typedef long *(CMP)();

    /* sorts array base of nel elements each of size size using comparison */
    /*   function pointed to by cmp */
    /* ARGS: char *base, unsigned long nel, unsigned long size, CMP cmp */

extern  tml_c_tml_os_qsort();

    /* open file named by (possibly non-null-terminated) pathname of length */
    /*   name_len for purposes and under conditions specified by oflag with */
    /*   permissions mode. Returns file handle suitable for host system, or */
    /*   NULL on failure. */
    /* ARGS: char *pathname, long name_len, long oflag, long mode */

extern char *tml_c_tml_os_open();

    /* read nbyte characters into buf from file associated with file_handle */
    /*   obtained by a call to tml_os_open(); return bytes read, 0 on EOF,  */
    /*   -1 on failure */
    /* ARGS: char *file_handle, char *buf, unsigned long nbyte */

extern long tml_c_tml_os_read();

    /* write nbyte characters from buf into file associated with file_handle */
    /*   obtained by a call to tml_os_open(); return bytes written, 0 on EOF, */
    /*   -1 on failure */
    /* ARGS: char *file_handle, char *buf, unsigned long nbyte */

extern long tml_c_tml_os_write();

    /* close file associated with file_handle (obtained from tml_os_open); */
    /*   return -1 on failure, zero otherwise. */
    /* ARGS: char *file_handle */

extern long tml_c_tml_os_close();

    /* rewind file associated with file_handle (obtained from tml_os_open); */
    /*   return -1 on failure, zero otherwise. */
    /* ARGS: char *file_handle */

extern long tml_c_tml_os_rewind();

    /* return NULL on failure or pointer to nelem * elsize bytes of free space. */
    /*   Each byte is initialized to zero. */
    /* ARGS: unsigned long nelem, unsigned long elsize */

extern char* tml_c_tml_os_alloc();

    /* free the space at ptr, which must have been obtained from a call to */
    /*   tml_os_alloc(), or the result is unpredictible. */
    /* ARGS: char* ptr */

extern  tml_c_tml_os_free();

    /* return -1 if address at ptr1 precedes address at ptr2, 0 if they */
    /*   are equal, 1 if address at ptr1 follows address at ptr2. The */
    /*   comparison algorithm and definitions of precedes and follows are */
    /*   system dependent */
    /* ARGS: char* *ptr1, char* *ptr2 */

extern long tml_c_tml_os_ptr_cmp();

    /* initialize a semaphore structure with a given capacity */
    /* ARGS: long pool_capacity */

extern char *tml_os_init_sem();

    /* free a semaphore structure */
    /* ARGS: char *old_sem */

extern  tml_os_free_sem();

    /* busy wait on a semaphore */
    /* ARGS: char *on_sem */

extern long tml_os_busy_wait();

    /* lock a semaphore */
    /* ARGS: char *on_sem */

extern  tml_os_p_sem();

    /* continue process waiting for semaphore */
    /* ARGS: char *proc_id, char *on_sem */

extern char *tml_c_tml_os_v_sem();

    /* check availability of a resource via its semaphore */
    /* ARGS: char *on_sem */

extern long tml_os_resource_available();

    /* get process identifier for the current process -- returns a pointer */
    /*   to the appropriate object cast as a char **/
    /* ARGS:  */

extern char *tml_c_tml_os_getpid();

    /* reduce vulnerability of the current process to interrupts during critical */
    /*   code sections -- returns status of interrupts before the call */
    /* ARGS:  */

extern char *tml_c_tml_os_ignore_interrupts();

    /* restore vulnerability of the current process to interrupts after completing */
    /*   critical code sections -- returns status of interrupts after the call */
    /* ARGS: char *interrupts */

extern char *tml_c_tml_os_restore_interrupts();

    /* kill a process dead -- returns a pointer to the appropriate object */
    /*   cast as a char **/
    /* ARGS: char *proc_id */

extern char *tml_c_tml_os_kill_proc();

#endif

#endif 
/* tml_os_depends_h */
