/**************************************************************************************
    tml_os_depends.hxx
    definitions of system- and machine-dependent macros for tml++
    declarations of the operating system dependent routines for tml++

    (c) Copyright Apollo Computer Inc., 1987 (unpublished)
     --------------------------------------------------------------------
    |THE FOLLOWING PROGRAMS ARE THE SOLE PROPERTY OF APOLLO COMPUTER INC.|
    |      AND CONTAIN ITS PROPRIETARY AND CONFIDENTIAL INFORMATION.     |
     --------------------------------------------------------------------
**************************************************************************************/

#ifndef tml_os_depends_hxx
#define tml_os_depends_hxx

#define TML_ALIGN_BLOCK 4   /* modulus in bytes of universal alignment boundary */

#ifndef O_RDONLY        /* true when fcntl.h has not been included */

#ifdef bsd
#include <fcntl.h>
#include <sys/file.h>
#include <signal.h>
#endif

#ifdef sys5
#include <fcntl.h>
#include <signal.h>
#endif

#ifdef aegis
#ifndef sys5
#include <fcntl.h>
#include <signal.h>
#endif
#endif

#endif

#ifndef csm_lock_hxx
#include <tml/csm_lock.hxx>
#endif

#include <tml/csm_cxx2.0_linkage_decls.hxx>

    /* copies count bytes from array from to array to--does */
    /*  not check sizes of either. Returns "to" if all goes well */

extern char* tml_os_byte_copy(char* from, char* to, long count);

    /* sorts array base of nel elements each of size size using comparison */
    /*   function pointed to by cmp */

extern void tml_os_qsort(char *base, unsigned long nel, unsigned long size, long (*cmp)(void*, void*));

    /* open file named by (possibly non-null-terminated) pathname of length */
    /*   name_len for purposes and under conditions specified by oflag with */
    /*   permissions mode. Returns file handle suitable for host system, or */
    /*   NULL on failure. */

extern void *tml_os_open(char *pathname, long name_len, long oflag, long mode);

    /* read nbyte characters into buf from file associated with file_handle */
    /*   obtained by a call to tml_os_open(); return bytes read, 0 on EOF,  */
    /*   -1 on failure */

extern long tml_os_read(void *file_handle, char *buf, unsigned long nbyte);

    /* write nbyte characters from buf into file associated with file_handle */
    /*   obtained by a call to tml_os_open(); return bytes written, 0 on EOF, */
    /*   -1 on failure */

extern long tml_os_write(void *file_handle, char *buf, unsigned long nbyte);

    /* close file associated with file_handle (obtained from tml_os_open); */
    /*   return -1 on failure, zero otherwise. */

extern long tml_os_close(void *file_handle);

    /* rewind file associated with file_handle (obtained from tml_os_open); */
    /*   return -1 on failure, zero otherwise. */

extern long tml_os_rewind(void *file_handle);

    /* return NULL on failure or pointer to nelem * elsize bytes of free space. */
    /*   Each byte is initialized to zero. */

extern char* tml_os_alloc(unsigned long nelem, unsigned long elsize);

    /* free the space at ptr, which must have been obtained from a call to */
    /*   tml_os_alloc(), or the result is unpredictible. */

extern void tml_os_free(char* ptr);

    /* return -1 if address at ptr1 precedes address at ptr2, 0 if they are */
    /* equal, 1 if address at ptr1 follows address at ptr2. The comparison */
    /* algorithm and definitions of precedes and follows are system dependent */

extern long tml_os_ptr_cmp(char* *ptr1, char* *ptr2);

    /* initialize a semaphore structure with a given capacity */

extern void* tml_os_init_sem(long pool_capacity);

    /* free a semaphore structure */

extern void tml_os_free_sem(void* old_sem);

    /* busy wait on a semaphore */

extern long tml_os_busy_wait(void* on_sem);

    /* lock a semaphore */

extern void tml_os_p_sem(void* on_sem);

    /* continue process waiting on semaphore */

extern void* tml_os_v_sem(void* proc_id, void* on_sem);

    /* check availability of a resource via its semaphore */

extern long tml_os_resource_available(void* on_sem);

    /* get process identifier for the current process -- returns a pointer */
    /*   to the appropriate object cast as a void* */

extern void* tml_os_getpid();

    /* reduce vulnerability of the current process to interrupts during critical */
    /*   code sections -- returns status of interrupts before the call */

extern void* tml_os_ignore_interrupts();

    /* restore vulnerability of the current process to interrupts after completing */
    /*   critical code sections -- returns status of interrupts after the call */

extern void* tml_os_restore_interrupts(void* interrupts);

    /* kill a process dead -- returns a pointer to the appropriate object */
    /*   cast as a void* */

extern void* tml_os_kill_proc(void* proc_id);

    /* issue command to write the byte arrays, of sizes given, */
    /*   and associated with given locks, asynchronously, */
    /*   i.e. without waiting for write call to return */
    /* the write operations are performed in sequence given */
    /* the destination file is identified by file_handle, */
    /*   which must have been returned from tml_os_open() */

extern long tml_os_asynch_write(void* file_handle, long num_bufs, char** bufs, long* nbyte_arr, csm_lock** locks);

#endif
// tml_os_depends_hxx

