/**************************************************************************************
    tml_os_io_bsd.cxx
    Operating system dependent input/output routines for UNIX BSD4.2

    (c) Copyright Apollo Computer Inc., 1987 (unpublished)
     --------------------------------------------------------------------
    |THE FOLLOWING PROGRAMS ARE THE SOLE PROPERTY OF APOLLO COMPUTER INC.|
    |      AND CONTAIN ITS PROPRIETARY AND CONFIDENTIAL INFORMATION.     |
     --------------------------------------------------------------------
**************************************************************************************/

#include <stdio.h>
#ifdef __cplusplus
#include <stdlib.h>
#include <osfcn.h>
#endif
#include <string.h>
#ifndef apollo
#include <strings.h>
#endif
#include <tml/endian.hxx>
#include <tml/tml_os_depends.hxx>

#define FRIENDLY_MODE   0775

/* return indicates file type */
/*  -1: call failed */
/*   1: tml object */
/*   3: unstruct object */
long tml_file_type(char* filename, short name_len )
{
    FILE* fid;
    long  tml_magic;
    long  type = -1;
    char* name_ptr = NULL;
               
    if ('\0' == *(filename + name_len - 1))
        name_ptr = filename;
    else {
        name_ptr = new char[name_len + 1];
        strncpy(name_ptr, filename, (int) name_len);
        *(name_ptr + name_len) = '\0';
        }

    if (fid = fopen(name_ptr, "r")) {
        if (fgets((char*) &tml_magic, sizeof (long), fid)) {
            switch (tml_magic) {
                case TML_MAGIC_0123:
                case TML_MAGIC_1032:
                case TML_MAGIC_2301:
                case TML_MAGIC_3210:
                    type = 1;
                    break;
                default:
                    type = 2;
                    break;
                }
            }
        fclose(fid);
        }

    if (name_ptr != filename)
        delete name_ptr;

    return type;
}

/* dummy routines */
char *tml_make_fhandle(char *file_ident, long file_type)
{
    return file_ident;
}
void tml_free_fhandle(char *fhandle)
{
}

/* returns pointer to UNIX file descriptor cast as a void * */
/* returns NULL on failure of open() */
/* return value is void* for operating system independence */
/* "0" mode value invokes "FRIENDLY_MODE" */
void *tml_os_open(char *filename, long name_len, long oflag, long mode)
{
    long *fdp;
    char *pathname;

    fdp = (long *) tml_os_alloc(1, (long) sizeof(long));
    pathname = (char *) tml_os_alloc(name_len + 1, (long) sizeof(char));
    pathname = tml_os_byte_copy(filename, pathname, name_len);

    if (0 == mode)
        mode = FRIENDLY_MODE;

    *fdp = open(pathname, (int) oflag, (int) mode);

    tml_os_free(pathname);

    if (-1 == *fdp) {
        tml_os_free((char*) fdp);
        return((void *) 0);
        }
    else
        return((void *) fdp);
}

/* return: */
/*    number of bytes read if successful */
/*    0 on eof */
/*    -1 on failure of read() */
long tml_os_read(void *file_handle, char *buf, unsigned long nbyte)
{
    long fildes;
    long bytes_read = 0;
    long total_bytes_read = 0;

    fildes = *((long *) file_handle);
    while (nbyte) {
        bytes_read = read((int) fildes, buf, (int) nbyte);
        if (0 == bytes_read)                   // EOF
            break;
        else if (-1 == bytes_read) {           // IO ERROR
            total_bytes_read = -1;
            break;
            }
        buf += bytes_read;
        nbyte -= bytes_read;
        total_bytes_read += bytes_read;
        }
    return(total_bytes_read);
}

/* return: */
/*   number of bytes written if successful */
/*   -1 on failure of write() */
long tml_os_write(void *file_handle, char *buf, unsigned long nbyte)
{
    long fildes;
    long bytes_written = 0;
    long total_bytes_written = 0;

    fildes = *((long *) file_handle);
    while (nbyte) {
        bytes_written = write((int) fildes, buf, (int) nbyte);

        // note no special case for return of 0; infinite retry

        if (-1 == bytes_written) {           // IO ERROR
            total_bytes_written = -1;
            break;
            }
        buf += bytes_written;
        nbyte -= bytes_written;
        total_bytes_written += bytes_written;
        }
    return(total_bytes_written);
}

/* rewind current file position to BOF */
/* return: */
/*    current position if successful */
/*    -1 on failure of lseek */
long tml_os_rewind(void *file_handle)
{
    long fildes;

    fildes = *((long *) file_handle);
    return lseek((int) fildes, 0L, 0);
}

long tml_os_close(void *file_handle)
{
    long fildes;
    long retval;

    fildes = *((long *) file_handle);
    retval = close((int) fildes);
    tml_os_free((char*) file_handle);
    return(retval);
}

