/**************************************************************************************
    unix_re.cxx
    Unix regular expression search function components

    (c) Copyright Apollo Computer Inc., 1987 (unpublished)
     --------------------------------------------------------------------
    |THE FOLLOWING PROGRAMS ARE THE SOLE PROPERTY OF APOLLO COMPUTER INC.|
    |      AND CONTAIN ITS PROPRIETARY AND CONFIDENTIAL INFORMATION.     |
     --------------------------------------------------------------------
**************************************************************************************/
#include <stdio.h>
#include <string.h>

// @(#)regexp.h 1.2
// 3.0 SID #    1.3

// modified 17-Mar-88 LCS for TML

#define	CBRA	2
#define	CCHR	4
#define	CDOT	8
#define	CCL	12
#define	CDOL	20
#define	CCEOF	22
#define	CKET	24
#define	CBACK	36

#define	STAR	01
#define RNGE	03

#define	NBRA	9

// Following macros are used by compile for "flexible" access to
// the characters in the original regular expression string to be
// compiled - why such flexibility was deemed desirable, I don't
// know, but TML has no use for it, so I define the macros here
// so no client program need worry about the sucker.
#define INIT         register char *sp = instring;
#define GETC()       ((sp == max_index) ? (char)0 : *sp++)
#define PEEKC()      (*sp)
#define UNGETC(c)    (--sp)
#define RETURN(c)    {return c;}
#define ERROR(c)     {printf("bad return from compile_exp\n"); return 0;}
#define FAILURE      return(1)
#define SUCCESS      return(0)
#define OKAY(x)      ((x) == 0)

// The following macros are used to update bits in a vector
// bvp ::= BitVectorPointer -- a pointer to array of char
// bn  ::= BitNumber        -- number of bit to set
#define SETBIT(bn) (ep[bn >> 3] |= ((bn & 0x7) << 1))
#define CHKBIT(bn) (ep[bn >> 3] &  ((bn & 0x7) << 1))
#define ecmp(s1, s2, n)	(!strncmp(s1, s2, n))

char *compile(char *instring, char *ep, char *endbuf, long seof, long instring_len)
{
	INIT	/* Dependent declarations and initializations */
	register c;
	register eof = seof;
    char *max_index = instring+instring_len;
	char *lastep = instring;
	long cclcnt;
	char bracket[NBRA], *bracketp;
	long closed;
	long neg;
	long lc;
	long i, cflg;
	int	nbra;
	int	ebra;

	lastep = 0;
	if((c = GETC()) == eof || c == '\n') {
		if(c == '\n') {
			UNGETC(c);
		}
		if(ep == 0)
			ERROR(41);
		RETURN(ep);
	}
	bracketp = bracket;
	closed = nbra = ebra = 0;
	if(c == '^')
		*ep = 1;
	else {
		UNGETC(c);
        *ep = 0;
    }
    ep++;
	while(1) {
		if(ep >= endbuf)
			ERROR(50);
		c = GETC();
		if(c != '*' && ((c != '\\') || (PEEKC() != '{')))
			lastep = ep;
		if(c == eof) {
			*ep++ = CCEOF;
			RETURN(ep);
		}
		switch(c) {

		case '.':
			*ep++ = CDOT;
			continue;

		case '\n':
			UNGETC(c);
			*ep++ = CCEOF;
			RETURN(ep);

		case '*':
			if(lastep == 0 || *lastep == CBRA || *lastep == CKET)
				goto defchar;
			*lastep |= STAR;
			continue;

		case '$':
			if(PEEKC() != eof && PEEKC() != '\n')
				goto defchar;
			*ep++ = CDOL;
			continue;

		case '[':
			if(&ep[17] >= endbuf)
				ERROR(50);

			*ep++ = CCL;
			lc = 0;
			for(i = 0; i < 16; i++)
				ep[i] = 0;

			neg = 0;
			if((c = GETC()) == '^') {
				neg = 1;
				c = GETC();
			}

			do {
				if(c == '\0' || c == '\n')
					ERROR(49);
				if(c == '-' && lc != 0) {
					if((c = GETC()) == ']') {
						SETBIT('-');
						break;
					}
					while(lc < c) {
						SETBIT(lc);
						lc++;
					}
				}
				lc = c;
				SETBIT(c);
			} while((c = GETC()) != ']');
			if(neg) {
				for(cclcnt = 0; cclcnt < 16; cclcnt++)
					ep[cclcnt] ^= -1;
				ep[0] &= 0376;
			}

			ep += 16;

			continue;

		case '\\':
			switch(c = GETC()) {

			case '(':
				if(nbra >= NBRA)
					ERROR(43);
				*bracketp++ = nbra;
				*ep++ = CBRA;
				*ep++ = nbra++;
				continue;

			case ')':
				if(bracketp <= bracket || ++ebra != nbra)
					ERROR(42);
				*ep++ = CKET;
				*ep++ = *--bracketp;
				closed++;
				continue;

			case '{':
				if(lastep == (char *) 0)
					goto defchar;
				*lastep |= RNGE;
				cflg = 0;
			nlim:
				c = GETC();
				i = 0;
				do {
					if('0' <= c && c <= '9')
						i = 10 * i + c - '0';
					else
						ERROR(16);
				} while(((c = GETC()) != '\\') && (c != ','));
				if(i > 255)
					ERROR(11);
				*ep++ = i;
				if(c == ',') {
					if(cflg++)
						ERROR(44);
					if((c = GETC()) == '\\')
						*ep++ = 255;
					else {
						UNGETC(c);
						goto nlim;
						/* get 2'nd number */
					}
				}
				if(GETC() != '}')
					ERROR(45);
				if(!cflg)	/* one number */
					*ep++ = i;
				else if((ep[-1] & 0377) < (ep[-2] & 0377))
					ERROR(46);
				continue;

			case '\n':
				ERROR(36);

			case 'n':
				c = '\n';
				goto defchar;

			default:
				if(c >= '1' && c <= '9') {
					if((c -= '1') >= closed)
						ERROR(25);
					*ep++ = CBACK;
					*ep++ = c;
					continue;
				}
			}
	/* Drop through to default to use \ to turn off special chars */

		defchar:
		default:
			lastep = ep;
			*ep++ = CCHR;
			*ep++ = c;
		}
	}
}

long step(char *p1, char *p2, long p1_len)
{
	register c;
	char *braslist[NBRA];
	char *braelist[NBRA];
	char *loc1, *loc2;
    char *p1_start = p1;
    char *p1_max = p1+p1_len;
    long advance(
      char  *lp,
      char  *ep,
      char **braslist,
      char **braelist,
      char **loc2);

	if(*p2 == 1 /* find a circf? */) {
		loc1 = p1;
		if(OKAY(advance(p1, p2++, braslist, braelist, &loc2))) return (long)(loc1-p1_start);
        else return (-1);
	}
    p2++;
	/* fast check for first character */
	if(*p2 == CCHR) {
		c = p2[1];
		do {
			if(*p1 != c)
				continue;
			if(OKAY(advance(p1, p2, braslist, braelist, &loc2))) {
				loc1 = p1;
                return (long)(loc1-p1_start);
			}
		} while((p1<p1_max) && (*p1++));
        return -1;
	}
		/* regular algorithm */
	do {
		if(OKAY(advance(p1, p2, braslist, braelist, &loc2))) {
			loc1 = p1;
            return (long)(loc1-p1_start);
		}
	} while((p1 < p1_max) && (*p1++));
    return -1;
}

long advance(
  char  *lp,
  char  *ep,
  char **braslist,
  char **braelist,
  char **loc2)
{
	register char *curlp;
	long c;
	char *bbeg;
	long ct;
    long low;
    long size;
    void getrnge(char *str, long& low, long& size);

	while(1) {
		switch(*ep++) {

		case CCHR:
			if(*ep++ == *lp++)
				continue;
			FAILURE;
	
		case CDOT:
			if(*lp++)
				continue;
			FAILURE;
	
		case CDOL:
			if(*lp == 0)
				continue;
			FAILURE;
	
		case CCEOF:
			*loc2 = lp;
			SUCCESS;
	
		case CCL:
			c = *lp++ & 0177;
			if(CHKBIT(c)) {
				ep += 16;
				continue;
			}
			FAILURE;

		case CBRA:
			braslist[*ep++] = lp;
			continue;
	
		case CKET:
			braelist[*ep++] = lp;
			continue;
	
		case CCHR | RNGE:
			c = *ep++;
			getrnge(ep,low,size);
			while(low--)
				if(*lp++ != c)
					FAILURE;
			curlp = lp;
			while(size--)
				if(*lp++ != c)
					break;
			if(size < 0)
				lp++;
			ep += 2;
			goto star;
	
		case CDOT | RNGE:
			getrnge(ep,low,size);
			while(low--)
				if(*lp++ == '\0')
					FAILURE;
			curlp = lp;
			while(size--)
				if(*lp++ == '\0')
					break;
			if(size < 0)
				lp++;
			ep += 2;
			goto star;
	
		case CCL | RNGE:
			getrnge(ep + 16,low,size);
			while(low--) {
				c = *lp++ & 0177;
				if(!CHKBIT(c))
					FAILURE;
			}
			curlp = lp;
			while(size--) {
				c = *lp++ & 0177;
				if(!CHKBIT(c))
					break;
			}
			if(size < 0)
				lp++;
			ep += 18;		/* 16 + 2 */
			goto star;
	
		case CBACK:
			bbeg = braslist[*ep];
			ct = braelist[*ep++] - bbeg;
	
			if(ecmp(bbeg, lp, ct)) {
				lp += ct;
				continue;
			}
			FAILURE;
	
		case CBACK | STAR:
			bbeg = braslist[*ep];
			ct = braelist[*ep++] - bbeg;
			curlp = lp;
			while(ecmp(bbeg, lp, ct))
				lp += ct;
	
			while(lp >= curlp) {
				if(OKAY(advance(lp, ep,braslist,braelist,loc2)))
                	SUCCESS;
				lp -= ct;
			}
			FAILURE;
	
		case CDOT | STAR:
			curlp = lp;
			while(*lp++);
			goto star;
	
		case CCHR | STAR:
			curlp = lp;
			while(*lp++ == *ep);
			ep++;
			goto star;
	
		case CCL | STAR:
			curlp = lp;
			do {
				c = *lp++ & 0177;
			} while(CHKBIT(c));
			ep += 16;
			goto star;
	
		star:
			do {
				--lp;
				if(OKAY(advance(lp, ep,braslist,braelist,loc2)))
					SUCCESS;
			} while(lp > curlp);
			FAILURE;

		}
	}
}

static void getrnge(char *str, long& low, long& size)
{
	low = *str++ & 0377;
	size = (*str == 255)? 20000: (*str &0377) - low;
}

// glue routines to fit into nifty search envelope done by JW
char *tml_unix_compile_re(char *compile_st,long compile_len,void *out_pat,long out_pat_max,long case_sens=1)
{
  return (compile(compile_st, (char *)out_pat, (char *)out_pat+out_pat_max, 0, compile_len));
}

long tml_unix_match_re(char *search_st, long search_len, void *use_pat)
{
  return (step((char *)search_st, (char *)use_pat, search_len));
}

long tml_unix_min_match(void *pattern)
{
  return 10;
}




