/**************************************************************************************
    viewport_traits.cxx
    Basic traits for the viewport object in tml++

    (c) Copyright Apollo Computer Inc., 1987 (unpublished)
     --------------------------------------------------------------------
    |THE FOLLOWING PROGRAMS ARE THE SOLE PROPERTY OF APOLLO COMPUTER INC.|
    |      AND CONTAIN ITS PROPRIETARY AND CONFIDENTIAL INFORMATION.     |
     --------------------------------------------------------------------
**************************************************************************************/

#include <tml/tml_cxx2.0_linkage_decls.hxx>
#include <tml/tml_objects.hxx>
#include <tml/tml_functions.hxx>
#include <tml/tml_dp_graph.hxx>
#include <tml/dflt_view_funcs.hxx>
    
#define NULL 0

/*useful constants for these traits*/
const    long       tml_view_pos_alloc_size    = 10;
const    screen_xy zero_zero = { 0, 0 };

/*default table of rendering functions*/
#ifdef apollo
vpftbl* dflt_vpftbl __attribute((section(dflt_vpftbl))) = NULL;
#else
vpftbl* dflt_vpftbl = NULL;
#endif

/*initializer for default contents of default table*/
#ifdef apollo_global
vpftbl dflt_view_port_function_table;
#else
vpftbl dflt_view_port_function_table = {
    &tml_dp_x_get_default_font,
    &tml_dp_x_get_fhandle,
    &tml_dp_x_load_font,
    &tml_dp_x_get_fcharsize,
    &tml_dp_x_get_fcharnum,
    &tml_dp_x_get_fchar1,
    &tml_dp_x_get_fhi,
    &tml_dp_x_get_fpoint,
    &tml_dp_x_get_fbline,
    &tml_dp_x_is_fixed,
    &tml_dp_x_get_fwidths,
    &tml_dp_x_get_ficspace,
    &tml_dp_x_set_ficspace,
    &tml_dp_x_get_fspace,
    &tml_dp_x_set_fspace,
    &tml_dp_x_funload,
    &tml_dp_x_frelease,
    &tml_dp_x_make_graphic_handle,
    &tml_dp_x_delete_graphic_handle,
    &tml_dp_x_get_hi,
    &tml_dp_x_get_wide,
    &tml_dp_x_get_foreg_id,
    &tml_dp_x_get_backg_id,
    &tml_dp_x_get_fill_backg_id,
    &tml_dp_x_get_fill_obj_id,
    &tml_dp_x_set_foreg_id,
    &tml_dp_x_set_backg_id,
    &tml_dp_x_set_fill_backg_id,
    &tml_dp_x_set_fill_obj_id,
    &tml_dp_x_set_clip,
    &tml_dp_x_clear_clip,
    &tml_dp_x_clear,
    &tml_dp_x_copy,
    &tml_dp_x_blt,
    &tml_dp_x_string_draw,
    &tml_dp_x_string_size,
    &tml_dp_x_damage_repair,
    &tml_dp_x_dflt_cursor_draw,
    &tml_dp_x_get_grafsys_obj,
    (char* (*)(...)) &get_x_default_font_name,
    (char* (*)(...)) &set_x_default_font_name,
    /* ftbl field is filled by initializations in tml_extend.cxx */
    /*   for bound libraries. For Apollo inlib'd libraries, the  */
    /*   alternate entry point tml_inlib_entry() initializes     */
    /*   this field automatically. For Apollo global librarires  */
    /*   the user initializes TMLib via an explicit call to that */
    /*   function.                                               */
};
#endif

/* function returning default vpftbl for GPR */
extern void* tml_dp_get_gpr_genfuncp_tbl();

/* function to return dflt_vpftbl as a void* */
vpftbl* tml_dp_get_dflt_vpftbl()
{
    if (NULL == dflt_vpftbl) {
        dflt_view_port_function_table.grafsys_type = tml_xwindows_grafsys;
        tml_dp_set_dflt_vpftbl(&dflt_view_port_function_table);
        }
    return dflt_vpftbl;
}

/*function to set dflt_vpftbl from a void* */
void tml_dp_set_dflt_vpftbl(vpftbl* new_table)
{
    dflt_vpftbl = new_table;
}

/*function to get default vpftbl for XWindows */
vpftbl* tml_dp_get_x_vpftbl()
{
    if (0 == dflt_view_port_function_table.grafsys_type)
        dflt_view_port_function_table.grafsys_type = tml_xwindows_grafsys;
    return &dflt_view_port_function_table;
}

#ifdef apollo

/*gpr table of rendering functions*/
vpftbl* gpr_dflt_vpftbl __attribute((section(gpr_dflt_vpftbl))) = NULL;

/*function to get default vpftbl for GPR */
vpftbl* tml_dp_get_gpr_vpftbl()
{
    if (NULL == gpr_dflt_vpftbl) {
        gpr_dflt_vpftbl = (vpftbl*) tml_dp_get_gpr_genfuncp_tbl(); 
        /* This initializes a NULL member of the gpr GEN_FUNCP array */
        /*   defined in tml_dp_gpr_graph.c. Assumes that a pointer */
        /*   to function (GEN_FUNCP) and a pointer to object */
        /*   (tml_gpr_grafsys) are the same size. */
        gpr_dflt_vpftbl->grafsys_type = tml_gpr_grafsys;
        }
    return gpr_dflt_vpftbl;
}
#endif

/* tml_dp_get_dflt_grafsys_type() is in file tml_extend.cxx */

/**************************************************************************************
        constructor and destructor
**************************************************************************************/

/*basic constructor for a view_port object*/
view_port::view_port()
{
    register long i;
    register view_pos *hold_pos;

    bitmap_handle = 0;
    dx = dy = -1;
    h_offset = 0;
    top_line = bottom_line = 0;
    this_format = 0;
    cursor = 0;
    foreg_id = backg_id = -1;
    first_mark = 0;
    tf_vp = new view_pos;
    tf_vp->in_viewport = this;
    for (i = 0; i < tml_view_pos_alloc_size; i++){
        hold_pos = new view_pos;
        hold_pos->in_viewport = this;
        hold_pos->next = tf_vp;
        tf_vp = hold_pos;
        }
    view_state = new prop_state;
    ftbl = dflt_vpftbl;
}

/*basic destructor for a view_port object*/
view_port::~view_port()
{
    register view_pos   *current_top,
                        *next_top;
    register draw_mark *this_mark,
                                *next_mark;

    if (this_format)
        this_format->delete_viewport(this);

    if (cursor)
        delete cursor;

    free_all_pos();
    for (current_top = tf_vp; current_top; current_top = next_top){
        next_top = current_top->get_next();
        delete current_top;
        }

    for (this_mark = first_mark; this_mark; this_mark = next_mark){
        next_mark = this_mark->get_next();
        delete this_mark;
        }

    if (bitmap_handle) (*(ftbl->ptr_2_tml_dp_delete_graphic_handle))(bitmap_handle);

    ftbl = NULL;

    delete view_state;
}

/**************************************************************************************
        constructors for preloaded derived classes of class view_port
**************************************************************************************/

xw_view_port::xw_view_port()
{
    set_ftbl(tml_dp_get_x_vpftbl());
}

#ifdef apollo
gpr_view_port::gpr_view_port()
{
    set_ftbl(tml_dp_get_gpr_vpftbl());
}
#endif

/**************************************************************************************
        data access traits
        NOTE--most of the data access  traits are inline, defined in tml_objects.hxx
**************************************************************************************/

/*inquire on the property given a prop_ind. First check to see if it is set in the viewport;
 if not, return it's value in the format; if unset in the format the check on the format will
 return its value in the model*/
void* view_port::inq_state(p_index prop_ind)
{
    void  *return_state;

    if (return_state = view_state->inq_state(prop_ind))
        return return_state;
    return (this_format ? this_format->inq_state(prop_ind) : 0);
}

/*inquire on the property settings at a given node given a prop_ind. First check to see
if it is set in the viewport; if not, return it's value in the format; if unset in the format
the check on the format will return its value in the model*/
void* view_port::inq_state(node *at_node, p_index prop_ind)
{
    void  *return_state;

    if (return_state = view_state->inq_state(prop_ind))
        return return_state;
    return (this_format ? this_format->inq_state(at_node, prop_ind) : 0);
}

/**************************************************************************************
        data setting traits
        NOTE--some of the data setting  traits are inline, defined in tml_objects.hxx
**************************************************************************************/
                                                      
/*set_handle takes a pre-allocated bitmap, assigns it to the viewport, and
    fills in the width and height. It assumes (currently) that the viewport fills
    the bitmap*/
void view_port::set_handle(void* bitmap_id)
{
    bitmap_handle = (*(ftbl->ptr_2_tml_dp_make_graphic_handle))(bitmap_id, this);
    dx = (*(ftbl->ptr_2_tml_dp_get_wide))(bitmap_handle);
    dy = (*(ftbl->ptr_2_tml_dp_get_hi))(bitmap_handle);
    foreg_id = (*(ftbl->ptr_2_tml_dp_get_foreg_id))(bitmap_handle);
    backg_id = (*(ftbl->ptr_2_tml_dp_get_backg_id))(bitmap_handle);
}


/*attach format associates a format with a viewport. If the format supplied is 0, a new
    format is allocated. If the width or depth is set as 0, that dimension is set to the
    size of the viewport.*/
long view_port::attach(format* to_format, long fx_size , long fy_size)
{
    long format_x,
        format_y;
                                              
    if (this_format) detach_format();

    if (!to_format) {
        this_format = new format;
        if (!this_format) return -1;
        if (fx_size || fy_size){
            format_x = fx_size ? fx_size : -1;
            format_y = fy_size ? fy_size : -1;
            this_format->set_page_size((short) format_y, (short) format_x);
            }
        }
    else {
        this_format = to_format;
        if (this_format->get_ref_count() < 2 ) {
            if (fx_size || fy_size){
                format_x = fx_size ? fx_size : -1;
                format_y = fy_size ? fy_size : -1;
                this_format->set_page_size((short) format_y, (short) format_x);
                }
            }
        }
    this_format->inc_ref_count();
    this_format->add_viewport(this);
    return 0;
}
                                                   
/*detach format disassociates a format from a viewport. If the ref count on the format
    goes to 0, the format is deleted*/
void view_port::detach_format()
{
    register format   *old_format;

    old_format = this_format;
    this_format = 0;
    old_format->dec_ref_count();
    old_format->delete_viewport(this);
    free_all_pos();
}

/*dummy used since default arguments cannot be assigned aggregate values*/
long view_port::attach(v_cursor *to_cursor)
{
    return attach(to_cursor, zero_zero);
}

/*attach a cursor to a viewport. If the cursor is attached to another viewport, an
    error is returned. The cursor is set to a given screen_xy; if none is
    specified the default setting is to 0,0 (upper left hand corner of the screen)*/
long view_port::attach(v_cursor *to_cursor, screen_xy set_to)
{
    if (!to_cursor) return -1;
    if (to_cursor->get_viewport()) return -1;
    cursor = to_cursor;
    to_cursor->attach(this);
    to_cursor->set(set_to);
    if (0 == to_cursor->get_draw_func())
        to_cursor->set_draw_func(ftbl->ptr_2_tml_dp_dflt_cursor_draw);
    return 0;
}
                                                                                                                
/*detach a cursor from a viewport. */
void view_port::detach_cursor()
{
    cursor->detach();
    cursor = 0;
}

/*set the foreground color id*/
void view_port::set_foreg_id(long new_id)
{
    foreg_id = new_id;
    (*(ftbl->ptr_2_tml_dp_set_foreg_id))(bitmap_handle, new_id);
}

/*set the background color id*/
void view_port::set_backg_id(long new_id)
{
    backg_id = new_id;
    (*(ftbl->ptr_2_tml_dp_set_backg_id))(bitmap_handle, new_id);
}

/*set_state--set a property override at the viewport level*/
long view_port::set_state(p_index prop_ind, void* value)
{
    return view_state->set_default(prop_ind, value);
}

/*unset_state--cancel property override at the viewport level*/
long view_port::unset_state(p_index prop_ind)
{
    return view_state->set_default(prop_ind, 0);
}

/**************************************************************************************
        Misc. routines--allocating and freeing view_positions

**************************************************************************************/
                                                      
/*get_pos--allocates a new view_pos from the available free store of those
    objects. If none are available, another batch is allocated and placed on
    the free store*/
view_pos* view_port::get_pos()
{
    register view_pos   *return_pos;
    register view_pos   *hold_pos;
    register long         i;

    if (!tf_vp) {
        tf_vp = new view_pos;
        tf_vp->in_viewport = this;
        for (i = 0; i < tml_view_pos_alloc_size; i++){
            hold_pos = new view_pos;
            hold_pos->in_viewport = this;
            hold_pos->next = tf_vp;
            tf_vp = hold_pos;
            }
        }

    return_pos = tf_vp;
    tf_vp = tf_vp->next;
    return_pos->next = 0;
    return return_pos;
}

/*free_pos--returns a view_pos to the free heap for a viewport,
    initializing the contents--NOTE--when freeing a view_pos, all
    view_pos to the right of the freed view_pos are also freed*/
void view_port::free_pos(view_pos* to_free)
{
    to_free->detach();
    to_free->in_viewport = this;
    to_free->ulx =  to_free->uly = -1;
    to_free->this_line = 0;
    to_free->next = tf_vp;
    to_free->prior = 0;
    tf_vp = to_free;
}
                                                                                     
/*free_all_pos--frees all of the view positions in a viewport*/
void view_port::free_all_pos()
{
    register view_pos   *this_pos,
                        *next_pos;

    for(this_pos = top_line; this_pos; this_pos = next_pos){
        next_pos = this_pos->get_next();
        free_pos(this_pos);
        }
    top_line = bottom_line = 0;
}


/**************************************************************************************
        Navigation routines--getting from one coordinate space to
        view_pos, cline, or model_pos occupying that space.

**************************************************************************************/
                                                                                 
/* xy_2_vpos -- convert from x,y coordinates to the view_pos object */
/*   occupying that space in the viewport. */
/* assumes that view_pos objects (base class or any future derived extension) */
/*   are ordered with respect to increasing x within increasing y of their */
/*   upper left corners */
view_pos* view_port::xy_2_vpos(screen_xy where)
{
    register view_pos *candidate;
             view_pos *best_fit = NULL;

    for (candidate = top_line; candidate; candidate = candidate->get_next()) {
        if (where.at_y < (candidate->get_y() + candidate->get_cline()->get_hi())) {
            if ((candidate->get_cline()->get_lry() < candidate->get_cline()->get_phead()->get_tused()->get_uly())
              || (candidate->get_cline()->get_uly() >= candidate->get_cline()->get_phead()->get_bused()->get_lry()))
                /* skip over lines in header or footer */
                continue;
            if (where.at_x >= candidate->get_x()) {
                /* point desired appears right of upper left corner of candidate */
                if ((where.at_x < (candidate->get_x() + candidate->get_cline()->get_wide()))
                  || 
                  ((cl_text == candidate->get_cline()->get_type()) &&
                  (where.at_x < ((tline*) candidate->get_cline())->get_full_text_width()))) {
                    /* candidate spans where.at_x */
                    best_fit = candidate;
                    break;
                    }
                else if (NULL == best_fit)
                    /* acquires value of first view_pos that exceeds where.at_y and spans where.at_x */
                    /*   or remains equal to top_line if no such view_pos is found */
                    best_fit = candidate;
                }
            else if (NULL == best_fit)
                best_fit = candidate;
            }
        }
    return (best_fit) ? best_fit : bottom_line;
}

/*xy_2_cline--getting from x,y coordinate space to the cline occupying
    that space. If no cline occupies the space, get the one closest*/
cline* view_port::xy_2_cline(screen_xy where)
{
    register view_pos *in_pos = xy_2_vpos (where);

    return (in_pos) ? in_pos->get_cline() : 0;
}

/*xy_2_mpos--return the character position corresponding to the x,y supplied. If no
    character occupies the space, get the one closest*/
model_pos view_port::xy_2_mpos(screen_xy where)
{
    model_pos          return_pos;
    register view_pos *in_pos,
                            *hold_pos;
    register cline      *in_cline;
    xy_coord           cpage_xy;

    in_pos = xy_2_vpos(where);
    if (!in_pos){
        //NOTE--if this ever happens, all hell has broken loose...
        return_pos.in_node = 0;
        return_pos.char_index = 0;
        return return_pos;
        }

    /*calculate the conceptual page coordinate corresponding to the screen coordinate*/
    in_cline = in_pos->get_cline();
    for (hold_pos = in_pos->get_next(); !in_cline->get_start().in_node; hold_pos = hold_pos->get_next()){
        if (!hold_pos)
            break;
        in_cline = hold_pos->get_cline();
        }
    if (!in_cline->get_start().in_node)
        for (hold_pos = in_pos->get_prior(); !in_cline->get_start().in_node; hold_pos = hold_pos->get_prior()){
            if (!hold_pos)
                break;
            in_cline = hold_pos->get_cline();
            }

    cpage_xy.x_coord = where.at_x - h_offset;
    cpage_xy.y_coord = in_cline->get_uly() + (in_cline->get_hi() / 2);

    /*now, find what is there in the conceptual line*/
    return_pos = in_cline->xy_2_mpos(cpage_xy);
    return return_pos;
}

/*cline_2_vpos--find the first view_pos containing a given conceptual line.
    if the line is not on the screen, return 0*/
view_pos* view_port::cline_2_vpos(cline *find_line)
{
    view_pos   *return_pos;

    for (return_pos = top_line; return_pos; return_pos = return_pos->get_next())
        if (return_pos->get_cline() == find_line) break;
    return return_pos;
}

/*cline_2_xy--return the x,y coordinates of the beginning of a conceptual line if
    it appears in the viewport; otherwise return -1, -1*/
screen_xy view_port::cline_2_xy(cline *find_line)
{
    screen_xy  return_xy;
    view_pos   *in_pos;

    in_pos = cline_2_vpos(find_line);
    if (!in_pos){
        return_xy.at_x = -1;
        return_xy.at_y = -1;
        return return_xy;
        }

    return_xy.at_y = in_pos->get_y();
    return_xy.at_x = in_pos->get_x();
    return return_xy;
}
                               
/*mpos_2_vpos--returns the view_pos containing the supplied character if that
    character is on the screen, otherwise returns 0*/
view_pos* view_port::mpos_2_vpos(model_pos find_char)
{
    view_pos   *return_pos;
                                  
    for (return_pos = top_line; return_pos; return_pos = return_pos->get_next())
        if (return_pos->contains(find_char))
            return return_pos;

    return return_pos;
}

/*mpos_2_xy --returns the xy position of the origin of a character on the screen. If
    hint is supplied, the routine will first try to find the character in that view_pos. If it
    fails, or if hint is 0, mpos_2_vpos is called to find the correct view_position. Returns
    a screen xy of x=0, y = -1 if the character is not on the screen*/
screen_xy view_port::mpos_2_xy(model_pos find_char, view_pos *hint)
{
    register view_pos   *found_pos;
    screen_xy          where;
    xy_coord             cl_xy_loc;
                       
    if (!find_char.can_format()){
        where.at_x = 0;
        where.at_y = -1;
        return where;
        }

    if (!hint ||  !hint->contains(find_char)) {    //if there is a hint, see if it is good
        found_pos = mpos_2_vpos(find_char);     //hint not good, see if it is there
        if (!found_pos || !found_pos->contains(find_char)) { //not on the screen--return 0, -1
            where.at_x = 0;
            where.at_y = -1;
            return where;
            }
        }
    else
        found_pos = hint;
                                     
    cl_xy_loc = found_pos->get_cline()->mpos_2_xy(find_char);
    where.at_x = cl_xy_loc.x_coord + h_offset;
    where.at_y = cl_xy_loc.y_coord - found_pos->get_cline()->get_uly();
    where.at_y += found_pos->get_y();
    return where;
}
  
/*is_visible--returns 1 if the page indicated is currently visible in the viewport,
    0 otherwise*/
long view_port::is_visible(phead *page)
{
    register long     top_page_no,
                        bottom_page_no;

    if (!page) return 0;
    if (!top_line || !bottom_line) return 0;
  
    if (!top_line->get_cline() || !bottom_line->get_cline())
        return 0;
    if (!top_line->get_cline()->get_phead() ||
        !bottom_line->get_cline()->get_phead())
        return 0;

    top_page_no = top_line->get_cline()->get_phead()->get_page_no();
    bottom_page_no = bottom_line->get_cline()->get_phead()->get_page_no();

    if ((page->get_page_no() < top_page_no) || (page->get_page_no() > bottom_page_no))
        return 0;
    else
        return 1;
}

/**************************************************************************************
        Drawing routines--actually put some text on the screen

**************************************************************************************/
/*dummy used since default arguments cannot be initialized with aggregate values*/
void view_port::clear()
{
    clear(zero_zero, zero_zero);
}

/*dummy used since default arguments cannot be initialized with aggregate values*/
void view_port::clear(screen_xy ul)
{
    clear(ul, zero_zero);
}

/*clear the viewport-- if no arguments are given, the whole viewport is cleared;
    if no lower corner is given, the viewport is cleared to the bottom edge*/
void view_port::clear(screen_xy ul, screen_xy lr)
{
    register view_pos   *this_pos,
                            *next_pos;
  
    if (cursor)
        cursor->set_visible(0);
    erase_all_marks();
    disable_all_marks();
        
    /*update viewport size information*/
    dx = (*(ftbl->ptr_2_tml_dp_get_wide))(bitmap_handle);
    dy = (*(ftbl->ptr_2_tml_dp_get_hi))(bitmap_handle);

    if ((lr.at_x == zero_zero.at_x)&&(lr.at_y == zero_zero.at_y)) {
        lr.at_x = dx;
        lr.at_y = dy;
        if ((ul.at_x == zero_zero.at_x)&&(ul.at_y == zero_zero.at_y)) {    //special case--clear the whole screen
            (*(ftbl->ptr_2_tml_dp_clear))(bitmap_handle, ul.at_x, ul.at_y, lr.at_x, lr.at_y);
            free_all_pos();
            enable_all_marks();
            return;
            }
        }
             
    /*NOTE--in what follows, it is assumed that the area to be cleared contains the width of the
        screen--if this assumption changes, this will have to be rewritten*/
    (*(ftbl->ptr_2_tml_dp_clear))(bitmap_handle,  ul.at_x, ul.at_y, lr.at_x, lr.at_y);
    this_pos = ((ul.at_x == zero_zero.at_x)&&(ul.at_y == zero_zero.at_y)) ? top_line : xy_2_vpos(ul);
    for (next_pos = this_pos; next_pos && (next_pos->get_y() < ul.at_y); next_pos = next_pos->get_next())
        this_pos = next_pos;
    if (!this_pos){
        top_line = bottom_line = 0;
        enable_all_marks();
        draw_all_marks();
        return;
        }

    if (lr.at_y == dy)
        if (this_pos == top_line){
            if (this_pos->get_y() > ul.at_y)
                top_line = bottom_line = 0;
            else{
                bottom_line = top_line;
                }
            }
        else
            bottom_line = this_pos;
    for (this_pos = this_pos->get_next();
          this_pos && (this_pos->get_y() >= ul.at_y) &&
          ((this_pos->get_y() + this_pos->get_cline()->get_hi()) <= lr.at_y);
          this_pos = next_pos){
        next_pos = this_pos->get_next();
        free_pos(this_pos);
        }
    enable_all_marks();
    draw_all_marks();

}
                                                                                                          
/*draw full viewport starting with the given cline--basic draw routine*/
void view_port::draw(cline *first_line, screen_xy where)
{
    register cline     *top_cline,
                       *old_cline,
                       *new_cline,
                       *hold_line;
    register view_pos  *found_pos;
             screen_xy  found_xy;
    register long       i,
                        v_offset;
    phead              *new_phead;
    model_pos           first_pos;
    screen_xy           show_pos;

    /* check for incoming trash */
    if (NULL == first_line)
        return;

    /* find first formatted model position if any */
    for (old_cline = first_line;
      old_cline && (NULL == old_cline->get_start().in_node);
      old_cline = old_cline->get_down())
        ;
    if (old_cline) {
        first_pos = old_cline->get_start();
        show_pos.at_x = where.at_x + old_cline->mpos_2_xy(first_pos).x_coord;
        show_pos.at_y = where.at_y + old_cline->get_uly() - first_line->get_uly()
                        + ((cl_text == old_cline->get_type()) ? ((tline*) old_cline)->get_baseline() : 0);
        }

    /*update viewport size information*/
    dx = (*(ftbl->ptr_2_tml_dp_get_wide))(bitmap_handle);
    dy = (*(ftbl->ptr_2_tml_dp_get_hi))(bitmap_handle);

    /*do any damage repair*/
    (*(ftbl->ptr_2_tml_dp_damage_repair))(bitmap_handle);
     
    /*erase any visible marks*/
    erase_all_marks();
    disable_all_marks();
                            
    /*if the line is already on the page we can just  call adjust*/
    if (found_pos = cline_2_vpos(first_line)){
        found_xy.at_x = found_pos->get_x();
        found_xy.at_y = found_pos->get_y();
        v_offset = first_line->get_uly() +
                       (first_line->get_phead()->get_hi() * (first_line->get_phead()->get_page_no() - 1));
        if (v_offset < where.at_y)
            where.at_y = v_offset;
        adjust(where.at_x - found_xy.at_x, where.at_y - found_xy.at_y);
        enable_all_marks();
        draw_all_marks();
        return;
        }

    /*find out what the new top line will be*/
    i = where.at_y;
    /*if i = 0, first_line will be the top line*/
    if (!i)
        top_cline = first_line;
    else {
        for (top_cline = first_line, hold_line = first_line->get_up(); 
          hold_line && (i > 0);
          hold_line = hold_line->get_up()) {
            if (hold_line->can_draw()) {
                i -= hold_line->get_hi();
                top_cline = hold_line;
                }
            }

        /*if the top line is on the page, we can just adjust*/
        if (found_pos = cline_2_vpos(top_cline)){
            found_xy.at_x = found_pos->get_x();
            found_xy.at_y = found_pos->get_y();
            v_offset = top_cline->get_uly() +
                       (top_cline->get_phead()->get_hi() * (top_cline->get_phead()->get_page_no() - 1));
            if (v_offset < where.at_y)
                where.at_y = v_offset;
            adjust(where.at_x - found_xy.at_x, 0 - found_pos->get_y());
            enable_all_marks();
            draw_all_marks();
            return;
            }
        }
     
    /* top line not on page; handle outdated page or line */
    if (old_cline) {
        new_phead = this_format->mpos_2_phead(first_pos);
        new_cline = (new_phead) ? new_phead->mpos_2_hcline(first_pos) : NULL;
        if (new_cline && (new_cline != first_line)) {
            /* new_phead and new_cline are NULL when old_cline is header */
            /* prior reformatting changed clines or pheads; redraw  */
            draw(first_pos, show_pos);
            return;
            }
        }

    /* top line not on page; pages/lines are OK; redraw */
    clear();      //clear the whole viewport

    /*now draw_to_bottom, using the top line*/
    where.at_y = 0;
    draw_to_bottom(where, top_cline);

    enable_all_marks();
    draw_all_marks();
    if (cursor)
        cursor->set(cursor->get_model_pos());
}

/*draw the screen, putting the specified character at (or as close as possible to) the screen
    coordinate supplied--the coordinate refers to the origin of the character*/
void view_port::draw(model_pos draw_char, screen_xy where)
{
    xy_coord  page_xy;
    phead        *on_phead;
    cline      *on_cline;
    font_tbl    *use_font;
                     
    /*update viewport size information*/
    dx = (*(ftbl->ptr_2_tml_dp_get_wide))(bitmap_handle);
    dy = (*(ftbl->ptr_2_tml_dp_get_hi))(bitmap_handle);

    if (!draw_char.in_node)
        return;
              
    on_phead = this_format->mpos_2_phead(draw_char);
    on_cline = on_phead ? (on_phead->mpos_2_hcline(draw_char)) : 0;
    if (!on_cline)
        return;                                                                                

    page_xy = on_cline->mpos_2_xy(draw_char);
    where.at_x -= page_xy.x_coord;

    use_font = (font_tbl *)this_format->inq_state(draw_char.in_node, tml_font_prop);
    if (use_font)
        where.at_y -= use_font->get_baseline();

    if (where.at_y < 0)
        where.at_y = 0;

    draw(on_cline, where);
}



/*adjust the viewport over the current contents by the amounts supplied. If delta_x and delta_y
    are greater than dx, dy, don't do anything*/
void view_port::adjust(long delta_x, long delta_y)
{
             long       s_ulx, s_uly,
                        s_lrx, s_lry,
                        dst_x, dst_y;
    register long       v_offset;
             screen_xy  draw_xy,
                        clear_xy;
             cline     *draw_cline;
             view_pos  *draw_pos,
                       *current_pos;

    register phead     *current_phead;

    /*update viewport size information*/
    dx = (*(ftbl->ptr_2_tml_dp_get_wide))(bitmap_handle);
    dy = (*(ftbl->ptr_2_tml_dp_get_hi))(bitmap_handle);

    /*first, do any damage repair*/
    (*(ftbl->ptr_2_tml_dp_damage_repair))(bitmap_handle);

    if (!delta_x && !delta_y) return;

    if (delta_x < 0){
        s_ulx = (delta_x * -1);
        s_lrx = dx;
        dst_x = 0;
        }
    else {
        s_ulx = 0;
        s_lrx = dx - delta_x;
        dst_x = delta_x;
        }

    if (delta_y < 0){
        draw_xy.at_x = 0;
        draw_xy.at_y = (delta_y * -1);
        draw_pos = xy_2_vpos(draw_xy);
        if (!draw_pos)
            s_uly = (delta_y * -1);
        else {
            s_uly = draw_pos->get_y();
            if (s_uly > draw_xy.at_y)
                s_uly += draw_pos->get_cline()->get_hi();
            }
        delta_y = s_uly * -1;
        s_lry = dy;
        dst_y = 0;
        draw_cline = bottom_line->get_cline()->get_down();
        while(draw_cline){
            if (draw_cline->get_uly() != bottom_line->get_cline()->get_uly())
                break;
            else
                draw_cline = draw_cline->get_down();
            }
        if (draw_cline && draw_cline->get_phead()->next_page(1));
        }
    else if (delta_y > 0){
        draw_cline = top_line->get_cline()->get_up();
        while(draw_cline){
            if (draw_cline->get_uly() != top_line->get_cline()->get_uly())
                break;
            else
                draw_cline = draw_cline->get_up();
            }
        if (draw_cline && draw_cline->get_phead()->prior_page(1));
           
        if (draw_cline && draw_cline->get_hi() > delta_y)
            return;
        if (dy < (delta_y + top_line->get_y() + top_line->get_cline()->get_hi()))
            delta_y = (dy - (top_line->get_y() + top_line->get_cline()->get_hi()));
        s_uly = 0;
        draw_xy.at_x = 0;
        draw_xy.at_y = dy - delta_y;
        draw_pos = xy_2_vpos(draw_xy);
        if (draw_pos)
            s_lry = draw_pos->get_y() + draw_pos->get_cline()->get_hi();
        else
            s_lry = delta_y;
        dst_y = delta_y;
        }
    else {
        s_uly = dst_y = 0;
        s_lry = dy;
        }
                               
    if (cursor)
        cursor->erase();
    erase_all_marks();
    disable_all_marks();

    (*(ftbl->ptr_2_tml_dp_copy))(bitmap_handle, s_ulx, s_uly, s_lrx, s_lry, dst_x, dst_y);

    top_line->adjust(delta_x, delta_y);
    if (delta_x){
        h_offset += delta_x;
        if (delta_x < 0){
            (*(ftbl->ptr_2_tml_dp_set_clip))(bitmap_handle, dx + delta_x, 0, dx, dy);
            (*(ftbl->ptr_2_tml_dp_clear))(bitmap_handle, dx + delta_x, 0, dx, dy);
            }
        else {
            (*(ftbl->ptr_2_tml_dp_set_clip))(bitmap_handle, 0, 0, delta_x, dy);
            (*(ftbl->ptr_2_tml_dp_clear))(bitmap_handle, 0, 0, delta_x, dy);
            }
        for (current_pos = top_line; current_pos; current_pos = current_pos->get_next()){
            draw_xy.at_x = current_pos->get_x();
            draw_xy.at_y = current_pos->get_y();
            draw_pos = current_pos->get_cline()->draw_line(this, draw_xy, current_pos->get_start().char_index);
            delete draw_pos;
            }
        (*(ftbl->ptr_2_tml_dp_clear_clip))(bitmap_handle);
        }

    if (delta_y < 0){
        if (!draw_cline){
            enable_all_marks();
            draw_all_marks();
            return;
            }
        draw_xy.at_x = h_offset;
        if (bottom_line)
            draw_xy.at_y = bottom_line->get_y() + bottom_line->get_cline()->get_hi();
        else
            draw_xy.at_y = dy + delta_y;
        clear_xy.at_x = dx;
        clear_xy.at_y = dy;
        clear(draw_xy, clear_xy);
        while(draw_cline){
            if (draw_cline->can_draw())
                break;
            draw_cline = draw_cline->cline::get_next();
            }
        if (draw_cline)
            draw_to_bottom(draw_xy, draw_cline);
        }
    else if (delta_y > 0){
        if (top_line) {
            draw_xy.at_y = top_line->get_y();
            clear_xy.at_y = draw_xy.at_y;
            }
        else
            clear_xy.at_y = delta_y;
        clear_xy.at_x = dx;
        clear(zero_zero, clear_xy);
        draw_xy.at_x = h_offset;
        if (top_line){
            v_offset = draw_xy.at_y - top_line->get_cline()->get_uly();
            current_phead = top_line->get_cline()->get_phead();
            }
        else{
            v_offset = draw_xy.at_y;
            current_phead = draw_cline->get_phead();
            }
        for ( ; draw_cline; draw_cline = draw_cline->get_prior()){
            if (draw_cline->get_phead() != current_phead){
                current_phead = draw_cline->get_phead();
                v_offset = draw_xy.at_y - draw_cline->get_lry();
                }
            if (!draw_cline->can_draw()){
                v_offset += draw_cline->get_hi();
                continue;
                }
            if ((draw_cline->get_lry() + v_offset) < 0)
                break;
            if ((draw_cline->get_uly() + v_offset) < 0)
                continue;
            draw_xy.at_y = draw_cline->get_uly() + v_offset;
            draw_xy.at_x = draw_cline->get_ulx() + h_offset;
            draw_pos = draw_cline->draw_line(this,draw_xy);
            if (draw_pos){
                draw_pos->attach(top_line, vp_up);
                top_line = draw_pos;
                }
            else
                draw_xy.at_y += draw_cline->get_hi();
            }
        draw_xy.at_x = 0;
        draw_xy.at_y = bottom_line->get_y();
        if (bottom_line->get_cline())
            draw_xy.at_y += bottom_line->get_cline()->get_hi();
        clear_xy.at_x = dx;
        clear_xy.at_y = dy;
        clear(draw_xy, clear_xy);
        }
    enable_all_marks();
    draw_all_marks();
    if (cursor)
        cursor->set(cursor->get_model_pos());
}

/*redraw the viewport from scratch, reformatting all of the pages needed*/
void view_port::redraw()
{
    register cline     *start_line;
    register node      *trace_node;
    register cline     *hold_line;
             model_pos  start_pos;
             phead     *first_page,
                       *this_page,
                       *last_page;
             screen_xy  where;
             long       old_vp_size[2];
             long      *format_value;


    old_vp_size[0] = dx;
    old_vp_size[1] = dy;

    /* update viewport size information */
    dx = (*(ftbl->ptr_2_tml_dp_get_wide))(bitmap_handle);
    dy = (*(ftbl->ptr_2_tml_dp_get_hi))(bitmap_handle);

    /* do any damage control */
    (*(ftbl->ptr_2_tml_dp_damage_repair))(bitmap_handle);
            
    /* erase any visible marks */
    erase_all_marks();
    disable_all_marks();

    /* if nothing in viewport, redraw marks and return */
    if (!top_line || !bottom_line){
        enable_all_marks();
        draw_all_marks();
        return;
        }

    /* find first and last pheads on the page */
    first_page = top_line->get_cline()->get_phead();
    last_page = bottom_line->get_cline()->get_phead();
    start_pos = xy_2_mpos(zero_zero);
    if (!start_pos.in_node){
        for (hold_line = top_line->get_cline()->get_next(); hold_line; hold_line = hold_line->get_next()){
            start_pos = hold_line->get_start();
            if (start_pos.in_node)
                break;
            }
        }

    if (!start_pos.in_node){
        enable_all_marks();
        draw_all_marks();
        return;
        }
    
    /* verify that start_pos is still ok */
    for (trace_node = start_pos.in_node; trace_node->get_node(n_up); trace_node = trace_node->get_node(n_up))
        ;
    if (trace_node != first_page->get_format()->get_model()->get_root()){
        start_pos.in_node = 0;
        for(start_line = top_line->get_cline()->get_next(); start_line; start_line = start_line->get_next()){
            for (trace_node = start_line->get_start().in_node;
                  trace_node && trace_node->get_node(n_up);
                  trace_node = trace_node->get_node(n_up))
                  ;
            if (trace_node == first_page->get_format()->get_model()->get_root()){
                start_pos = start_line->get_start();
                break;
                }
            }
        }
    if (!start_pos.in_node){
        enable_all_marks();
        draw_all_marks();
        return;
        }
               
    /* mark first through last visible pages for reformatting */
    for (this_page = first_page; this_page && (this_page != last_page); this_page = this_page->get_next())
        this_page->mark_reformat();
    last_page->mark_reformat();

    /* if viewport size has changed, mark for reformatting */
    /*   all remaining pages in format that are not */
    /*   dominated by page size properties and for which */
    /*   char_wrap or word_wrap is enabled */
    /* if there is no wrap property enabled on the node containing */
    /*   page start for any of these pages, or if the property is */
    /*   no_wrap, examine individual lines on the page for presence */
    /*   of a char_wrap or word_wrap property at the node containing */
    /*   the start position of the line */
    /* NOTE: A null format_value means that no_wrap is enabled */
    /*   by default and therefore reformatting has no effect on */
    /*   line widths */
    if ((dx != old_vp_size[0])
      || (dy != old_vp_size[1])) {
        for (this_page = first_page->get_prior();
          this_page; 
          this_page = this_page->get_prior()) {
            if (NULL == this_page->get_format()->inq_state(this_page->get_start().in_node, tml_pgsize_prop)) {
                if ((NULL != (format_value = (long*) this_page->get_format()->inq_state(this_page->get_start().in_node, tml_wrap_prop)))
                  && (no_wrap != *format_value))
                    this_page->mark_reformat();
                if (this_page->get_format_ok()) {
                    for (hold_line = this_page->get_tline(); hold_line; hold_line = hold_line->inq_next()) {
                        if (cl_text != hold_line->get_type())
                            continue;
                        if ((NULL != (format_value = (long*) hold_line->get_format()->inq_state(hold_line->get_start().in_node, tml_wrap_prop)))
                          && (no_wrap != *format_value))
                            this_page->mark_reformat();
                            break;
                        }
                    }
                }
            }
        for (this_page = last_page->get_next();
          this_page; 
          this_page = this_page->get_next()) {
            if (NULL == this_page->get_format()->inq_state(this_page->get_start().in_node, tml_pgsize_prop)) {
                if ((NULL != (format_value = (long*) this_page->get_format()->inq_state(this_page->get_start().in_node, tml_wrap_prop)))
                  && (no_wrap != *format_value))
                    this_page->mark_reformat();
                if (this_page->get_format_ok()) {
                    for (hold_line = this_page->get_tline(); hold_line; hold_line = hold_line->inq_next()) {
                        if (cl_text != hold_line->get_type())
                            continue;
                        if ((NULL != (format_value = (long*) hold_line->get_format()->inq_state(hold_line->get_start().in_node, tml_wrap_prop)))
                          && (no_wrap != *format_value))
                            this_page->mark_reformat();
                            break;
                        }
                    }
                }
            }
        }

    /* clear viewport */
    clear();

    /* find cline on which start_pos now resides */
    if (!(start_line = first_page->mpos_2_hcline(start_pos))){
        first_page = first_page->get_format()->mpos_2_phead(start_pos);
        if (!first_page){
            enable_all_marks();
            draw_all_marks();
            return;
            }
        start_line = first_page->mpos_2_hcline(start_pos);
        if (!start_line) return;
        }

    /* draw */
    where.at_y = 0;
    where.at_x = start_line->get_ulx() + h_offset;
    draw_to_bottom(where, start_line);

    enable_all_marks();
    draw_all_marks();
    if (cursor)
        cursor->set(cursor->get_model_pos());
}

/* calculate amount by which the text for a line overlaps the left edge of its bounding box */
long calc_left_overhang(tline* for_line)
{
             char_set *in_char_set;
    unsigned long      char_size;
             char     *left_char;
             long      byte_count, i, tab_cnt, left_corr;
             lrud_dir  text_dir;

    in_char_set = (char_set*) for_line->get_start().in_node->get_model()->inq_state(for_line->get_start().in_node, tml_char_set_prop);
    char_size = in_char_set->get_char_size();
    left_char = ((char*) for_line->get_start().in_node->get_contents()) + for_line->get_start().char_index;
    byte_count = for_line->get_count();
    if (::right == (text_dir = in_char_set->get_dir_to_char((unsigned char*)left_char))) {
        /* text runs right -- count tabs */
        for (i = 0, tab_cnt = 0; i < byte_count; i += char_size, tab_cnt++) {
            if (0 == in_char_set->is_htab((unsigned char*)left_char))
                break;
            in_char_set->advance((unsigned char*&)left_char);
            }
        if (i == byte_count)
            /* line is all tabs */
            left_corr = 0;
        else {
            /* nontab character found in line */
            /* get its extent left of width bounding box for line */
            left_corr = for_line->get_font()->get_lbearing(in_char_set->as_ulong((unsigned char*)left_char));
            if (tab_cnt && (0 <= (left_corr += for_line->get_start_x_off())))
                /* tabs consume left correction */
                left_corr = 0;
            }
        }
    else if (::left == text_dir) {
        /* text runs left -- find overhang */
        left_char += (byte_count - char_size);
        if (0 == in_char_set->is_htab((unsigned char*)left_char)) {
            /* nontab character on left */
            left_corr = for_line->get_font()->get_rbearing(in_char_set->as_ulong((unsigned char*)left_char))
                        - for_line->get_font()->get_width(in_char_set->as_ulong((unsigned char*)left_char))
                        - for_line->get_font()->get_icspace();
            }
        else
            left_corr = 0;
        }
    return left_corr;
}

/* incremental redraw */
/* redraw the viewport, reformatting as required, from the supplied view_pos */
/*   until no more is needed. Argument "ref_start_mpos" specifies the last */
/*   valid model position within the current contents of start_vpos. Argument */
/*   "rend_start_mpos" specifies the model position to be reformatted at the */
/*   viewport position of the invalid character. Argument "start_vpos" specifies */
/*   the invalid view position within which ref_start_mpos is found. */
/* The function first reformats the line contained in start_vpos by calling */
/*   cline::reformat() on that line. When this call resolves to tline::reformat(), */
/*   and rend_start_mpos refers to textual data, tline::reformat() alters the */
/*   starting position of the line to be reformatted appropriately if its force_start */
/*   argument is set or if the new start position for reformatting precedes (with */
/*   respect to model position arithmetic) the model position that begins the line that */
/*   is to be reformatted. The force_start argument of that function is set when */
/*   this function detects a rend_start_mpos less than or equal to ref_start_mpos. */
/* In this way all cases of incremental redraw can be covered by one function. */
/*   Cases where reformatting causes overflow on the conceptual page are handled */
/*   by tline::reformat() on the formatting side, and by this function on the */
/*   rendering side after tline::reformat() returns. */
/* Note that page (and therefore viewport) underflow is automatically handled by */
/*   the recursive call. Underflow occurs when this call follows a deletion */
/*   operation that removed a line-ending character from the model. */
void view_port::redraw(model_pos ref_start_mpos, model_pos rend_start_mpos, view_pos *start_vpos)
{       
    register cline     *old_contents,
                       *new_contents,
                       *next_old_contents,
                       *last_line,
                       *new_line;
             long       next_ok,
                        old_hi,
                        left_corr;
             xy_coord   cpage_xy;
             screen_xy  redraw_xy,
                        clear_xy;
             view_pos  *junk;
             view_pos  *more_junk;
             model_pos  new_pos;
             long       old_width;
             long       force_new_start;

    /* detect incoming trash */
    if ( ! start_vpos)
        return;

    /* update viewport size information */
    dx = (*(ftbl->ptr_2_tml_dp_get_wide))(bitmap_handle);
    dy = (*(ftbl->ptr_2_tml_dp_get_hi))(bitmap_handle);

    /* change that provoked redraw may have left garbage on screen */
    if ( ! rend_start_mpos.in_node) {
        clear_xy.at_x = start_vpos->get_x();
        clear_xy.at_y = start_vpos->get_y();
        clear(clear_xy);
        return;
        }

    /* reformat the line */
    old_contents = start_vpos->get_cline();
    if (!old_contents)
        return;
    else {
        old_hi = old_contents->get_hi();
        old_width = old_contents->get_contents_width();
        }
    force_new_start = (rend_start_mpos <= ref_start_mpos) ? 1 : 0;
    new_contents = old_contents->reformat(rend_start_mpos, next_ok, force_new_start);

    /* erase any currently visible marks */
    erase_all_marks();
    disable_all_marks();

    if (new_contents != old_contents) {
        if (new_contents) {
            /* reformatted line inserted before old line on conceptual page */
            /* adjust current viewport contents from old line downward */
            (*(ftbl->ptr_2_tml_dp_copy))(bitmap_handle, start_vpos->get_x(), start_vpos->get_y(),
                                dx, dy,
                                start_vpos->get_x(), start_vpos->get_y() + new_contents->get_hi());
            redraw_xy.at_y = start_vpos->get_y();
            redraw_xy.at_x = new_contents->get_ulx() + h_offset;
            if (cl_text == new_contents->get_type())
                redraw_xy.at_x += (left_corr = calc_left_overhang((tline*) new_contents));
            (*(ftbl->ptr_2_tml_dp_clear))(bitmap_handle, redraw_xy.at_x, redraw_xy.at_y, dx, redraw_xy.at_y + new_contents->get_hi());
            if (cl_text == new_contents->get_type())
                redraw_xy.at_x -= left_corr;
            if ((start_vpos != bottom_line) || 
              (start_vpos->get_y() + start_vpos->get_cline()->get_hi() + new_contents->get_hi()) < dy) {
                /* above bottom line or new and old lines will both fit */
                /* make room for new line */
                /* adjust view_pos objects from view_pos for old line downward */
                start_vpos->adjust(0, new_contents->get_hi());
                /* draw new line and attach view_pos returned above view_pos containing old line */
                junk = new_contents->draw_line(this, redraw_xy, new_contents->get_start().char_index);
                junk->attach(start_vpos, vp_up);
                }
            else {
                /* at bottom line and/or new and old lines will not both fit */
                junk = new_contents->draw_line(this, redraw_xy, new_contents->get_start().char_index);
                if (start_vpos == bottom_line) {
                    /* at bottom line */
                    /* reuse view_pos of old contents for new contents */
                    /* old_contents are now below screen */
                    start_vpos->set_line(new_contents);
                    free_pos(junk);
                    }
                else {
                    /* not at bottom line and new and old lines will not both fit */
                    /* insert new view_pos above old */
                    junk->attach(start_vpos, vp_up);
                    /* adjust view_pos objects from view_pos for old line downward */
                    start_vpos->adjust(0, new_contents->get_hi());
                    }
                }
            /* clear from below bottom line to end of viewport */
            (*(ftbl->ptr_2_tml_dp_clear))(bitmap_handle, 
                                          bottom_line->get_x(), 
                                          bottom_line->get_y() + bottom_line->get_cline()->get_hi(),
                                          dx, dy);
            }  
        }
    else {
        /* end of reformatted line does not fall immediately before start */
        /*   of old line */
        /* view_pos for old line reused for new line */
        if (old_hi != new_contents->get_hi()) {
            /* new line not same height as old */
            /* adjust material from old line downward */
            (*(ftbl->ptr_2_tml_dp_copy))(bitmap_handle, start_vpos->get_x(), start_vpos->get_y() +old_hi,
                               dx, dy,
                               start_vpos->get_x(), start_vpos->get_y() + new_contents->get_hi());
            if (start_vpos->get_down())
                start_vpos->get_down()->adjust(0, new_contents->get_hi() - old_hi);
            next_ok = 0;
            /* indicate line boundary changed */
            force_new_start = 1;
            /* start over to get correct height this time */
            rend_start_mpos = new_contents->get_start();
            }
        /* see if beginning of material to be rerendered falls in new line */
        /* ("no" answer handled by first "if" can occur only when line height unchanged) */
        cpage_xy = new_contents->mpos_2_xy(rend_start_mpos);
        if (cpage_xy.y_coord < 0) {
            /* beginning not in new line */
            next_ok = 0;
            if (old_width != new_contents->get_contents_width()) {
                /* old and new line widths differ */
                /* start over to get correct width this time */
                rend_start_mpos = new_contents->get_start();
                cpage_xy = new_contents->mpos_2_xy(rend_start_mpos);
                force_new_start = 1;
                }
            }
        if ((cpage_xy.y_coord > 0) && (NULL != new_contents->get_phead())) {
            /* a location for the character and a page on which to draw it */
            /*   (page is NULL if new_contents wrapped a page boundary) */
            next_old_contents = new_contents->get_phead()->mpos_2_cline(rend_start_mpos);
            redraw_xy.at_y = start_vpos->get_y();
            redraw_xy.at_x = h_offset + cpage_xy.x_coord;
            /* clear space to be occupied by new line */
            if (cl_text == next_old_contents->get_type())
                redraw_xy.at_x += (left_corr = calc_left_overhang((tline*) next_old_contents));
            (*(ftbl->ptr_2_tml_dp_clear))(bitmap_handle, redraw_xy.at_x, redraw_xy.at_y, dx, redraw_xy.at_y + next_old_contents->get_hi());
            if (cl_text == next_old_contents->get_type())
                redraw_xy.at_x -= left_corr;
            junk = next_old_contents->draw_line(this, redraw_xy, rend_start_mpos.char_index);
            /* reuse view_pos of old_line for new line */
            start_vpos->set_line(new_contents);
            free_pos(junk);
            }
        }

    /* next_ok set to zero in tline::reformat when no line returned */
    /*   or when invalid view_pos objects result from page overflow */
    /*   or when end of reformatted line changes */
    /* also set to zero above when new line does not end immediately */
    /*   before beginning of old and beginning of material to be */
    /*   rerendered does not appear in new line */
    if (next_ok) {
        enable_all_marks();
        draw_all_marks();
        return;
        }

    /* reformatting line may have caused page overflow */
    /* all view_pos objects from overflow down are invalid */
    /* find overflow start by searching through view_pos from */
    /*   bottom up for view_pos associated with last line */
    /*   on updated conceptual page */
    junk = NULL;
    if (new_contents && (new_contents->get_phead())) {
        last_line = new_contents->get_phead()->get_bline();
        for (junk = bottom_line;
          junk && (last_line != junk->get_cline());
          junk = junk->get_prior())
            ;
        }

    /* on overflow, reset viewport bottom line to last valid view_pos */
    /*   or NULL if no valid view_pos objects remain */
    if (junk) {
        bottom_line = junk;
        junk = junk->get_next();
        }
    /* delete invalid view_pos objects */
    for (more_junk = junk; junk; junk = more_junk) {
        more_junk = junk->get_next();
        if (start_vpos == junk)
            start_vpos = NULL;
        free_pos(junk);
        }

    if (new_contents == old_contents) {
        /* no need to redraw succeeding view positions when simple */
        /*   insertion took place, as they were copied downward  */

        /* this code will also redraw tail of viewport when invalid */
        /*   view_pos objects have been deleted */
        
        /* get next view_pos object that dominates text */
        if (start_vpos) {
            for (start_vpos = start_vpos->get_next(); 
              start_vpos && (0 == start_vpos->get_start().in_node); 
              start_vpos = start_vpos->get_next())
                /* find first remaining valid view_pos that dominates model data */
                ;
            }
        
        if (start_vpos) {
            /* a view_pos remains that dominates model data */
            if (start_vpos->get_cline()->get_phead() && 
              new_contents->get_end().get_next().in_node)
                /* the view_pos contains a valid cline (cline not freed due to overflow) */
                /* more remains in the model */
                /* redraw incrementally from the view_pos downward as far as necessary */
                redraw(new_contents->get_end().get_next(), new_contents->get_end().get_next(),
                        start_vpos);
            else {
                /* nothing left in model and viewport */
                /* clear to bottom of viewport */
                redraw_xy.at_x = start_vpos->get_x();
                redraw_xy.at_y = start_vpos->get_y();
                (*(ftbl->ptr_2_tml_dp_clear))(bitmap_handle, redraw_xy.at_x, redraw_xy.at_y, dx, dy);
                free_pos(start_vpos);
                }
            }
        
        /* no valid view_pos objects remain but a line was returned -- underflow */
        else if ((new_contents) && (new_contents->get_phead())) {
            /* no view_pos exists that dominates model data */
            /* or page overflow was detected starting somewhere below reformatted line */
            /* get model position of last character formatted */
            new_line = NULL;
            new_pos = new_contents->get_end();
            if (get_hi() > (bottom_line->get_y() + bottom_line->get_cline()->get_hi())) {
                /* space remains in viewport */
                if (new_pos == new_contents->get_phead()->get_end())
                    /* last character formatted was last on page */
                    /* get first line of next page */
                    new_line = new_contents->get_phead()->get_bline()->get_down();
                else {
                    /* last character formatted was not last on page */
                    /* get reformatted line associated with last character */
                    new_contents->get_phead()->reformat_page(new_pos);
                    new_line = new_contents->get_phead()->mpos_2_hcline(new_pos.get_next());
                    }
                if (new_line) {
                    /* not at end of model: a line was found */
                    /* clear below bottom line */
                    /* redraw to bottom of viewport */
                    redraw_xy.at_x = h_offset;
                    redraw_xy.at_y = bottom_line->get_y() + bottom_line->get_cline()->get_hi();
                    (*(ftbl->ptr_2_tml_dp_clear))(bitmap_handle, redraw_xy.at_x, redraw_xy.at_y, dx, dy);
                    draw_to_bottom(redraw_xy, new_line);
                    }
                }
            }
        
        /* no line returned -- clear to bottom of viewport */
        else {
            redraw_xy.at_x = h_offset;
            redraw_xy.at_y = bottom_line->get_y() + bottom_line->get_cline()->get_hi();
            (*(ftbl->ptr_2_tml_dp_clear))(bitmap_handle, redraw_xy.at_x, redraw_xy.at_y, dx, dy);
            }
        }
    enable_all_marks();
    draw_all_marks();
}

/**************************************************************************************
        Editing routines--functions which alter the model, reformat and
        redisplay

**************************************************************************************/
                    
long view_port::append_string(char* in_string, long count, model_pos after, char_set* ins_char_set)
{
            model_pos  reformat_start;
                phead *on_phead;
    register view_pos *view_start_ref;

    if (!this_format || !this_format->get_model())
        return 0;
                            
    if (!after.in_node)
        return 0;

    (*(ftbl->ptr_2_tml_dp_damage_repair))(bitmap_handle);
     
    /*find out where it is in the viewport*/
    view_start_ref = mpos_2_vpos(after);
    /*if it isn't in the viewport, just insert into the model and return*/
    if (!view_start_ref){
        reformat_start = after.in_node->insert(after.char_index, in_string, count, ins_char_set);
        if (!reformat_start.in_node)
            return 0;
        else {
            // mark page where insertion took place as needing reformatting
            on_phead = this_format->mpos_2_phead(after);
            on_phead->mark_reformat();
            return 1;
            }
        }

    /*make the change in the model*/
    reformat_start = after.in_node->insert(after.char_index, in_string, count, ins_char_set);
    if (!reformat_start.in_node)
        return 0;

    /*get rid of the cursor, reformat and redraw, and then put the cursor back*/
    if (cursor && cursor->get_visible())
        cursor->erase();
    redraw(view_start_ref->get_cline()->get_start(), after.get_next(), view_start_ref);
    if (cursor)
        cursor->set(cursor->get_model_pos());
    return 1;
}

long view_port::prepend_string(char* in_string, long count, model_pos before, char_set* ins_char_set)
{
            model_pos  reformat_start,
                       target_node;
            phead*     on_phead;
    register view_pos *view_start_ref;

    if (!this_format || !this_format->get_model())
        return 0;
                            
    if (!before.in_node)
        return 0;

    (*(ftbl->ptr_2_tml_dp_damage_repair))(bitmap_handle);
     
    /*find out where it is in the viewport*/
    target_node = before.get_prior();
    view_start_ref = mpos_2_vpos(target_node);
    if (!view_start_ref)
        view_start_ref = mpos_2_vpos(before);
    /*if it isn't in the viewport, just insert into the model and return*/
    if (!view_start_ref){
        reformat_start = before.in_node->insert(before.char_index - 1, in_string, count, ins_char_set);
        if (!reformat_start.in_node)
            return 0;
        else {
            // mark page where insertion took place as needing reformatting
            on_phead = this_format->mpos_2_phead(before);
            on_phead->mark_reformat();
            return 1;
            }
        }

    /*make the change in the model*/
    reformat_start = before.in_node->insert(before.char_index - 1, in_string, count, ins_char_set);
    if (!reformat_start.in_node)
        return 0;

    /*get rid of the cursor, reformat and redraw, and then put the cursor back*/
    if (cursor && cursor->get_visible())
        cursor->erase();
    redraw(view_start_ref->get_cline()->get_start(), reformat_start - (count - 1), view_start_ref);
    if (cursor)
        cursor->set(cursor->get_model_pos());
    return 1;
}

void view_port::draw_to_bottom(screen_xy where, cline *start_line)
{
    register cline  *hold_line,
                        *last_line;
    register long     v_offset;
    register phead    *current_phead;
    register view_pos   *draw_pos;
                                                  
    if (!start_line) return;

    /*update viewport size information*/
    dx = (*(ftbl->ptr_2_tml_dp_get_wide))(bitmap_handle);
    dy = (*(ftbl->ptr_2_tml_dp_get_hi))(bitmap_handle);

    v_offset = where.at_y - start_line->get_uly();
    h_offset = where.at_x - start_line->get_ulx();
    current_phead = start_line->get_phead();
    last_line = start_line;

    for(hold_line = start_line; hold_line; hold_line = hold_line->get_down()){
        if (hold_line->get_phead() != current_phead){
            v_offset = bottom_line ? (bottom_line->get_y() + last_line->get_hi()) : 0;
            current_phead = hold_line->get_phead();
            }
        if ((hold_line->get_uly() + v_offset) >= dy)
            break;
        if ((hold_line->get_lry() + v_offset) >= dy)
            break;
        if (hold_line->can_draw()){
            where.at_y = hold_line->get_uly() + v_offset;
            where.at_x = hold_line->get_ulx() + h_offset;
            draw_pos = hold_line->draw_line(this, where);
            if (draw_pos){
                if (bottom_line){
                    draw_pos->attach(bottom_line, vp_down);
                    bottom_line = draw_pos;
                    }
                else
                    top_line = bottom_line = draw_pos;
                last_line = hold_line;
                }
            }
        }
}


/* insert a string before the current cursor position--if the current cursor position is */
/* invalid, insert before the first character dominated by root */
long view_port::ins_before_cursor(char* in_string, long count, char_set* ins_char_set)
{
    model_pos insert_pos;
        
    if (!this_format || !this_format->get_model())
        return 0;
    if (!cursor)
        return 0;

    insert_pos = cursor->get_model_pos();
    if (insert_pos.in_node)
        insert_pos = insert_pos.get_prior();
    if (!insert_pos.in_node){
        insert_pos.in_node = this_format->get_model()->get_root();
        insert_pos.char_index = 0;
        insert_pos = insert_pos.get_next();
        if (!insert_pos.in_node)
            return 0;
        return prepend_string(in_string, count, insert_pos, ins_char_set);
        }
    else if (!insert_pos.in_node->can_format()){
        insert_pos = insert_pos.get_next();
        if (!insert_pos.in_node){
            insert_pos = cursor->get_model_pos();
            insert_pos = insert_pos.get_prior();
            if (!insert_pos.in_node)
                return 0;
            else
                return append_string(in_string, count, insert_pos, ins_char_set);
            }
        }

    return append_string(in_string, count, insert_pos, ins_char_set);
}

/* insert a string after the current cursor position--if the current cursor position is */
/* invalid, insert before the first character dominated by root */
long view_port::ins_after_cursor(char* in_string, long count, char_set* ins_char_set)
{
    model_pos insert_pos,
                    prior_pos;
    long        ret_val;

    if (!this_format || !this_format->get_model())
        return 0;
    if (!cursor)
        return 0;

    insert_pos = cursor->get_model_pos();
    if (!insert_pos.in_node)
        return ins_before_cursor(in_string, count);

    if (!insert_pos.in_node->can_format()){
        insert_pos = insert_pos.get_next();
        if (!insert_pos.in_node)
            return ins_before_cursor(in_string, count);
        }
        
    /*get the prior model position, and if it makes a difference prepend the new characters
        before the beginning of the insert position*/
    prior_pos = insert_pos.get_prior();
    if (!prior_pos.in_node)
        return prepend_string(in_string, count, insert_pos, ins_char_set);
    
    if ((cursor->get_cline()) && (insert_pos == cursor->get_cline()->get_start()))
        return prepend_string(in_string, count, insert_pos, ins_char_set);

    if ((prior_pos.in_node != insert_pos.in_node)
       && ((prior_pos.in_node->get_node(n_right) != insert_pos.in_node) ||
            (prior_pos.in_node->get_props() != insert_pos.in_node->get_props())))
            return prepend_string(in_string, count, insert_pos, ins_char_set);

    /*it doesn't make a difference if we prepend or append--so do an append so we don't
        spawn any unnecessary gnodes*/
    ret_val =  append_string(in_string, count, prior_pos, ins_char_set);
    /*since we appended to the character before, we need to see if we are adding at the beginning
        of the viewport, and if so redraw from the first line*/
    if ((cursor->get_view_pos() == top_line) &&
            (cursor->get_cline()->get_start() == insert_pos)){
        if (cursor && cursor->get_visible())
            cursor->erase();
        redraw(prior_pos.get_next(), prior_pos.get_next(), top_line);
        if (cursor)
            cursor->set(cursor->get_model_pos());
        }

    return ret_val;
}

long view_port::delete_char(model_pos delete_pos)
{
             model_pos  reformat_start;
             model_pos  reformat_end;
    register tline     *end_line;
    register view_pos  *view_start_ref;
             phead     *on_phead;

    if (!this_format || !this_format->get_model())
        return 0;
                            
    if (!delete_pos.in_node)
        return 0;

    (*(ftbl->ptr_2_tml_dp_damage_repair))(bitmap_handle);
     
    /*find out where it is in the viewport*/
    view_start_ref = mpos_2_vpos(delete_pos);
    /*if it isn't in the viewport, just delete in the model and return*/
    if (!view_start_ref){
        reformat_start = delete_pos.in_node->del_cont(delete_pos.char_index);
        if (!reformat_start.in_node)
            return 0;
        else
            // mark page where deletion began as needing reformatting
            on_phead = this_format->mpos_2_phead(reformat_start);
            if (on_phead) {
                on_phead->mark_reformat();
                reformat_end = reformat_start.get_next();
                if (reformat_end.in_node) {
                    // mark page where deletion ended as needing reformatting
                    on_phead = this_format->mpos_2_phead(reformat_end);
                    on_phead->mark_reformat();
                    }
                }
            return 1;
        }
                                       
    /*erase any marks, and get rid of the cursor*/
    erase_all_marks();
    disable_all_marks();
    if (cursor && cursor->get_visible())
        cursor->erase();

    /*make the change in the model*/
    reformat_start = delete_pos.in_node->del_cont(delete_pos.char_index);
    if (!reformat_start.in_node){
        /*redraw the marks and the cursor*/
        enable_all_marks();
        draw_all_marks();
        if (cursor)
            cursor->set(cursor->get_model_pos());
        return 0;
        }
 
    /*if we are deleting the first character in a node, and that node ends the line,
        decrement the count on the end of the line by 1*/
    if (!delete_pos.char_index && (view_start_ref->get_cline()->get_end().in_node == delete_pos.in_node))
        if (view_start_ref->get_cline()->get_type() == cl_text){
            end_line = (tline*)view_start_ref->get_cline();
            for( ; end_line->get_right(); end_line = (tline*)end_line->get_right())
                ;
            end_line->set_count(end_line->get_count() - 1);
            }
 
    if ((view_start_ref->get_cline()->get_start() == delete_pos) && 
       (delete_pos.char_index > 0) &&
       (view_start_ref->get_cline()->get_type() == cl_text)){
        end_line = (tline*)view_start_ref->get_cline();
        end_line->set_start(reformat_start);
        }

    /* reformat and redraw */
    redraw(view_start_ref->get_cline()->get_start(), reformat_start, view_start_ref);
    
    /*redraw marks*/
    enable_all_marks();
    draw_all_marks();

    /* restore cursor */
    if (cursor)
        cursor->set(cursor->get_model_pos());
    return 1;
}

/*add a property to a node, and update the screen if necessary. Returns 1 if the insert
    is successful, 0 if an error occured*/
long view_port::insert_prop(node *to_node, property *add_prop)
{
    register property    *attach_prop;
    register prop_vec   *attach_pvec;
    model_range            prop_range;
    register view_pos   *first_disp_vpos,
                                *last_disp_vpos;
    model_pos              first_disp_mpos,
                                 last_disp_mpos;

    /*check for trash*/
    if (!this || !this_format || !add_prop || !to_node)
        return 0;

    /*if the node isn't part of the model displayed in this viewport, return an error*/
    if (to_node->get_model() != this_format->get_model())
        return 0;

    /*if the property isn't in the model workspace of the node, copy it over*/
    if (add_prop->get_model() != to_node->get_model())
        attach_prop = add_prop->copy(to_node->get_model());
    else
        attach_prop = add_prop;

    if (!attach_prop)
        return 0;

    /*allocate a property vector element, and attach the property to it*/
    attach_pvec = to_node->get_model()->make_pvec();
    if (!attach_pvec)
        return 0;
    attach_pvec->attach(attach_prop);
    
    /*attach the new property vector element to the end of the node's property vector*/
    if (!to_node->get_props())
        to_node->prop_attach(attach_pvec);
    else
        to_node->get_props()->append(attach_pvec);

    /*now, find out if we need to update the display*/
    /*begin by getting the range of the new property*/
    prop_range = to_node->get_span();
    
    /*now find the first and last characters displayed on the screen*/
    for (first_disp_vpos = top_line; first_disp_vpos; first_disp_vpos = first_disp_vpos->get_next()){
        first_disp_mpos = first_disp_vpos->get_cline()->get_start();
        if (first_disp_mpos.in_node)
            break;
        }

    for (last_disp_vpos = bottom_line; last_disp_vpos; last_disp_vpos = last_disp_vpos->get_prior()){
        last_disp_mpos = last_disp_vpos->get_cline()->get_end();
        if (last_disp_mpos.in_node)
            break;
        }
    
    /*if there is nothing on the screen, nothing needs to be updated, so return 1*/
    if (!first_disp_mpos.in_node || !last_disp_mpos.in_node)
        return 1;

    /*if the screen start position or end position are contained in the node's span, we need to update*/
    if (prop_range.contains(first_disp_mpos) || prop_range.contains(last_disp_mpos)){
        redraw();
        return 1;
        }

    /*if the range starts after the end of the viewport, or end before the beginning, we don't need to update*/
    if ((prop_range.get_start() > last_disp_mpos) || (prop_range.get_end() < first_disp_mpos))
        return 1;
    
    /*if we've gotten here, the entire range is visible, so redraw the page*/
    redraw();
    return 1;
}

/*insert a property to dominate all the text given by the indicated model positions, redisplaying the
    viewport if necessary. Returns 1 if successful, 0 if something went wrong.*/
long view_port::insert_prop(model_pos first_char, model_pos last_char,property *add_prop)
{
    register node      *subtree_node;
    model_range        prop_range;

    /*check for garbage*/
    if (!this || !add_prop || !first_char.in_node || !last_char.in_node)
        return 0;
                    
    /*make sure that the model positions given are in the same model*/
    if (first_char.in_node->get_model() != last_char.in_node->get_model())
        return 0;

    /*now, make a node which dominates all and only the text spanned by first_char, last_char*/
    prop_range.set_start(first_char);
    prop_range.set_end(last_char);
    subtree_node = prop_range.isolate();
    if (!subtree_node)
        return 0;

    return(insert_prop(subtree_node, add_prop));
}

/*delete a property from a node, and if anything dominated by that node is visible, refresh
    the viewport. Returns 1 if the property is successfully deleted, 0 if an error occured*/
long view_port::delete_prop(property *del_prop, node *from_node)
{
    register prop_vec   *delete_pvec,
                                *hold_pvec;
    register long              found_prop;
    model_range            prop_range;
    register view_pos   *first_disp_vpos,
                                *last_disp_vpos;
    model_pos              first_disp_mpos,
                                 last_disp_mpos;

    /*first, check for trash*/
    if (!this || !del_prop || !from_node || !this_format)
        return 0;

    /*make sure that the node is in the model associated with the viewport*/
    if (from_node->get_model() != this_format->get_model())
        return 0;
             
    found_prop = 0;
    for (delete_pvec = from_node->get_props(); delete_pvec; delete_pvec = hold_pvec){
        hold_pvec = delete_pvec->get_next();
        if (delete_pvec->get_prop() == del_prop){
            found_prop = 1;
            if (from_node->get_props() == delete_pvec)
                from_node->prop_attach(delete_pvec->detach(from_node->get_props()));
            else
                delete_pvec->detach(from_node->get_props());
            }
        }
    
    /*if nothing got deleted, the property wasn't on the node*/
    if (!found_prop)
        return 0;

    /*now, find out if we need to update the display*/
    /*begin by getting the range of the old property*/
    prop_range = from_node->get_span();
    
    /*now find the first and last characters displayed on the screen*/
    for (first_disp_vpos = top_line; first_disp_vpos; first_disp_vpos = first_disp_vpos->get_next()){
        first_disp_mpos = first_disp_vpos->get_cline()->get_start();
        if (first_disp_mpos.in_node)
            break;
        }

    for (last_disp_vpos = bottom_line; last_disp_vpos; last_disp_vpos = last_disp_vpos->get_prior()){
        last_disp_mpos = last_disp_vpos->get_cline()->get_end();
        if (last_disp_mpos.in_node)
            break;
        }
    
    /*if there is nothing on the screen, nothing needs to be updated, so return 1*/
    if (!first_disp_mpos.in_node || !last_disp_mpos.in_node)
        return 1;

    /*if the screen start position or end position are contained in the node's span, we need to update*/
    if (prop_range.contains(first_disp_mpos) || prop_range.contains(last_disp_mpos)){
        redraw();
        return 1;
        }

    /*if the range starts after the end of the viewport, or end before the beginning, we don't need to update*/
    if ((prop_range.get_start() > last_disp_mpos) || (prop_range.get_end() < first_disp_mpos))
        return 1;
    
    /*if we've gotten here, the entire range is visible, so redraw the page*/
    redraw();
    return 1;
}
  
/*add a draw_mark object to the list of draw_marks in this viewport*/
void view_port::add_mark(draw_mark *new_mark)
{
    register draw_mark *hold_mark;

    /*check for trash*/
    if (!new_mark || !new_mark->get_model_mark())
        return;
                       
    if (!this_format || (this_format->get_model() != new_mark->get_model_mark()->get_range()->get_start().in_node->get_model()))
        return;

    /*make sure that it isn't in the list already--if it is, just return*/
    for (hold_mark = first_mark; hold_mark; hold_mark = hold_mark->get_next())
        if (hold_mark->get_model_mark() == new_mark->get_model_mark())
            return;

    /*it wasn't found, so add it to the beginning of the list of marks*/
    new_mark->set_next(first_mark);
    if (first_mark)
        first_mark->set_prior(new_mark);
    new_mark->set_prior(0);
    first_mark = new_mark;
    first_mark->set_view_port(this);
    
    /*set core function*/
    if ( ! new_mark->get_core_func())
        new_mark->set_core_func(ftbl->ptr_2_tml_dp_blt);

    /*now, draw the thing*/
    first_mark->enable();
    first_mark->draw();
}
                           
/*detach a draw_mark from the list of marks in the viewport*/
void view_port::detach_mark(draw_mark *del_mark)
{
    register draw_mark *hold_mark;

    /*first, check for incoming trash*/
    if (!del_mark)
        return;

    /*make sure that it is in the list*/
    for (hold_mark = first_mark; hold_mark; hold_mark = hold_mark->get_next())
        if (hold_mark == del_mark)
            break;

    if (!hold_mark)
        return;

    /*it's in the list, so we need to delete it*/
    if (hold_mark->get_prior())
        hold_mark->get_prior()->set_next(hold_mark->get_next());
    else
        first_mark = hold_mark->get_next();
    if (hold_mark->get_next())
        hold_mark->get_next()->set_prior(hold_mark->get_prior());

    /*erase the mark*/
    hold_mark->erase();

    /*now, delete it*/
    delete hold_mark;
}

/*mark a region in the viewport. Make a range in the model, make a model_mark, make a draw_mark,
    and then add the mark to this viewport*/
draw_mark* view_port::mark_region(model_pos start_mark, model_pos end_mark)
{
    register draw_mark *new_mark;
    model_range    this_range;
    register mark_list        *this_list;
    

    /*check for trash*/
    if (!start_mark.in_node || !end_mark.in_node)
        return 0;
    if (!this_format ||
       (this_format->get_model() != start_mark.in_node->get_model()) ||
       (this_format->get_model() != end_mark.in_node->get_model()))
        return 0;

    /*fill a model_range for this region*/
    this_range.set_start(start_mark);
    this_range.set_end(end_mark);

    /*now, make a mark_list entry in the model for this range*/
    this_list = this_format->get_model()->make_mark(this_range);
    if (!this_list)
        return 0;

    /*now, make a draw_mark object, and fill in the mark_list object it refers to*/
    new_mark = new draw_mark;
    if (!new_mark){
        this_format->get_model()->delete_mark(this_list);
        return 0;
        }

    new_mark->set_model_mark(this_list);
    add_mark(new_mark);
    return new_mark;
}
       
/*private function which disables all marks*/
void view_port::disable_all_marks(long force)
{
    register draw_mark     *current_mark;

    for (current_mark = first_mark; current_mark; current_mark = current_mark->get_next())
        current_mark->disable(force);
}

/*private function which enables all marks*/
void view_port::enable_all_marks(long force)
{
    register draw_mark     *current_mark;
    for (current_mark = first_mark; current_mark; current_mark = current_mark->get_next())
        current_mark->enable(force);
}

/*private function which erases any visible marks*/
void view_port::erase_all_marks()
{
    register draw_mark     *current_mark;

    for (current_mark = first_mark; current_mark; current_mark = current_mark->get_next())
        current_mark->erase();
}

/*private function to draw all marks*/
void view_port::draw_all_marks()
{
    register draw_mark *current_mark;

    for (current_mark = first_mark; current_mark; current_mark = current_mark->get_next())
        current_mark->draw();
}
