/**************************************************************************************
    write_model.cxx
    write a tml model to disk

    (c) Copyright Apollo Computer Inc., 1987 (unpublished)
     --------------------------------------------------------------------
    |THE FOLLOWING PROGRAMS ARE THE SOLE PROPERTY OF APOLLO COMPUTER INC.|
    |      AND CONTAIN ITS PROPRIETARY AND CONFIDENTIAL INFORMATION.     |
     --------------------------------------------------------------------
**************************************************************************************/

#include <stdio.h>
#include <fcntl.h>
#include <tml/tml_cxx2.0_linkage_decls.hxx>
#include <tml/tml_objects.hxx>
#include <tml/tml_os_depends.hxx>

/* Save a model on disk to file indicated by pathname and pathname length */
long model::d_write_model(char *pathname, long name_len)
{
    void *fdp;
    long bytes_written;
    if (NULL == (fdp = tml_os_open(pathname, name_len, O_WRONLY | O_CREAT, 0))) {
        printf("model::d_write_model(char *pathname, long name_len): tml_os_open() failed\n");
        return(-1);
        }
    bytes_written = this->d_write_model(fdp);
    if (-1 == tml_os_close(fdp))
        printf("model::d_write_model(char *pathname, long name_len): tml_os_close() failed\n");
    return(bytes_written);
}

/* Save a model on disk to file indicated by file_handle obtained from call to */
/*   tml_os_open() */
long model::d_write_model(void *file_handle)
{
    d_model_parts *model_parts;
    cs_heap       *csm_heaps;           // character set manager heaps
    long           total_bytes_written; // total bytes written by tml_os_write();
                                        //   check on successful model save to disk
 
    gc_hash_tbl hash_tbl;

    // 1. Lock the model and non-model-resident heaps in multiprocessing 
    //    applications, suspending self if model already locked.
    //    Model is unlocked in d_model_parts::save()

    if (model_lock)
        lock_suspend();
    if (heaps_lock)
        lock_heaps_suspend();

    // 2. Allocate and initialize control structure

    model_parts = new d_model_parts(this);

    // 3. Record disk versions of character sets in char_set_heaps member of
    //    model_parts, hashing char_set and union_char_set pointers in hash_tbl.
    //    Record disk pointer versions of object addresses in the appropriate positions
    //    in the heaps and copy to objects of reduced size in new heaps, referencing
    //    char_set and union_char_set numbers in argument values of char_set_arg and
    //    union_char_set_arg objects. Reference argument csm_heaps in call to
    //    model_parts::save_pointers() returns heaps created by that function's call
    //    to union_char_set::prepare_ucs_binary(). The value returned in the reference
    //    argument must be sent on as the cs_heap* argument to model_parts::save() so
    //    that cs_heap objects may be saved.

    csm_heaps = NULL;
    model_parts->save_pointers(csm_heaps, &hash_tbl);

    // 4. Complete d_model_parts based on information gained during pointer save

    model_parts->complete_for_save();

    // 5. Write disk save version of current model

    total_bytes_written = model_parts->save(file_handle, csm_heaps, &hash_tbl);

    // 6. Delete control structure

    delete model_parts;

    // 7. Delete csm_heaps if any created by call to d_model_parts::save_pointers()

    if (csm_heaps)
        delete csm_heaps;

    // 8. Return bytes written if successful, -1 otherwise

    return(total_bytes_written);
}

